/*****************************************************************************/
/*
                                MapUrl.c

Functions supporting mapping of URLs to VMS file specifications and VMS 
specifications to URLs, uses the HTTPd rule mapping file.  The number of rules 
it can store is not fixed, but is limited by available memory and the 
practical considerations of processing large, linear rule databases. 

Comment lines may be provided by beginning a line with a '#' character.

Lines may be continued by ensuring the last character on the line is a '\'.

Virtual server syntax is the same as for authentication/authorization rules,
[[virtual-host]] (all ports of host) or [[virtual-host:virtual-port]] (
matching specified port of host) for a specific host and [[*]] to resume rules
for all virtual hosts.


Extended File Specification
---------------------------
Extended file specifications (required for ODS-5) are supported.  A rule is
used to flag such paths for ODS-5 tailored processing.  Things that are not
done to ODS-5 paths; case-changes and character substitution.  Things that are
done to ODS-5 paths; escaping to and from the extended file specification
canonical escape syntax (i.e. '^').

Paths located on ODS-5 volumes need to have the ODS-5 rule SET against them.

  set /extended/* ODS-5

It is possible to explicitly mark a path as ODS-2 also, although as this is the
default it is usually unnecessary.  One situation where this might be useful is
where all paths are ODS-5 except a small minority, in which case a rule such as
the following might be used.

  set /* ODS-5
  set /old-path/* ODS-2

Paths derived using the USER rule are specially processed.  The SYSUAF user
home directory's structure is determined using a sys$getdvi() to determined
it's ACPTYPE.  Hence no explicit rule is required for these.


Mapping Rules
-------------
[[virtual-host:optional-virtual-port]]
EXEC       template  result           [setting(s)] [conditional(s)]
EXEC       (runtime)template  result  [setting(s)] [conditional(s)]
EXEC+      template  result           [setting(s)] [conditional(s)]
FAIL       template                                [conditional(s)]
MAP        template  result           [setting(s)] [conditional(s)]
PASS       template                                [conditional(s)]
PASS       template  result           [setting(s)] [conditional(s)]
REDIRECT   template  result           [setting(s)] [conditional(s)]
SCRIPT     template  result           [setting(s)] [conditional(s)]
SCRIPT+    template  result           [setting(s)] [conditional(s)]
SET        template  setting(s)                    [conditional(s)]
USER       template  result           [setting(s)] [conditional(s)]
UXEC       template  result           [setting(s)] [conditional(s)]
UXEC       (runtime)template  result  [setting(s)] [conditional(s)]
UXEC+      template  result           [setting(s)] [conditional(s)]

(Somewhat awkwardly the square brackets both denote optional fields and to
literally delimit a conditional.  Please do not confuse the two!)

Note that as of V8.0 path settings (anything that can be set with a SET rule)
can also follow the result string.  If the rule applies and any conditional is
matched then the settings are applied to the path BEFORE the mapping is done.

PASS rules can map to a result comprising an HTTP status code (e.g. 403) with a
following textual message (including white-space).  Just delimit the result
string with double or single quotes, or within curly braces.  For example:

PASS /private/stuff/* "403 Can't go in there!" [!ho:my.host.name]
PASS /private/stuff/* '403 "/private/stuff/" is off-limits!' [!ho:my.host.name]

EXEC rules can be used to map both directories and file types.  The former only
requires a single, trailing wildcard (example 1).  The latter requires two
wildcards (examples 2 and 3).  The  irst to locate the script, the second to
map any path.  This format can be used to globally map a file type, or confine
that mapping to specified part of a tree.

EXEC /cgi-bin/*
EXEC *.cgi* *.cgi*
EXEC /specified/path/*.cgi* /specified/path/*.cgi*


Run-Time Environment (RTE) Mapping
----------------------------------
Each persistant run-time environment must have it's own mapping environment. 
This may, or may not, correspond to physically distinct directory areas.  It is
the mapping rules that identify run-time environments.  A parenthesized
interprester specification leads the result component of the rule.  This is
first extracted and then the EXEC rule behaves as normal.  Hence the following 
example shows two such environments each with it's own interpreter.

  exec /plbin/* (cgi-bin:[000000]perlrte.exe)/ht_root/perl_local/*
  exec /pybin/* (cgi-bin:[000000]javarte.exe)/ht_root/python_local/*

The "/plbin/*" identifies request paths beginning with this string as a
hypothetical, persistant Perl run-time interpreter, with the Perl source
scripts available to it from HT_ROOT:[PERL_LOCAL].  Similarly with a
hypothetical Java Environment.

If a request with the following URI is given to the server

  http://the.host.name/plbin/plsearch/web/doc/

the following components will be derived

  run-time interpreter ... CGI-BIN:[000000]PERLRT.EXE
  Perl source script ..... HT_ROOT:[PERL_LOCAL]PLSEARCH.PL
  request path-info ...... WEB:[DOC]


Mapping User Directories
------------------------
The USER rule effectively PASSes the user's SYSUAF default device and directory
as the first wildcard substitution in the rule.  Accounts that are disusered,
captive or have expired passwords are never mapped, they are always "denied
access by default".  In the same way privileged accounts (those possessing
SYSPRV) cannot be mapped this way (see below for a workaround).

An example; if the user name DANIEL has the default device USER$DISK: and
default directory [DANIEL] the following request path

  /~daniel/

would be mapped to the result and VMS file specification

  /user$disk/daniel/www/
  USER$DISK:[DANIEL.WWW]

using the following rule (note that in all these examples the '@' character has
been  substituted for the '*' because of constraints with C-language comment
structures).

  USER  /~@/@  /@/www/

Note the "/www" subdirectory component.  It is recommended that user
directories always be mapped to a subdirectory of the physcial area.  This
effectively "sandboxes" Web access to that subdirectory hierarchy, allowing the
user privacy elsewhere in the home area.

To accomodate request user paths that do not incorporate a trailing delimiter
after the username the following redirect may be used to cause the browser to
re-request with a more appropriate path.

  REDIRECT  /~@  ///@/www/

WASD also "reverse maps" VMS specifications into paths and so requires
additional rules to provides these mappings.  (Reverse mapping is required
during directory listings and error reporting.)  In the above case the
following rules would actually be required (and in the stated order).

  USER  /~@/@  /@/www/
  PASS  /~@/@  /user$disk/@/www/@

Where user home directories are spread over multiple devices (physical or
concealed logical) a reverse-mapping rule would be required for each. Consider
the following situation, where user directories are distributed across these
devices (concealed logicals)

  USER$GROUP1:
  USER$GROUP2:
  USER$GROUP2:
  USER$OTHER:

This would require the following mapping rules (in the stated order).

  USER  /~@/@  /@/www/
  PASS  /~@/@  /user$group1/@/www/@
  PASS  /~@/@  /user$group2/@/www/@
  PASS  /~@/@  /user$group3/@/www/@
  PASS  /~@/@  /user$other/@/www/@

Of course vanilla mapping rules may be used to provide for special cases.  For
instance, if there is requirement for a particular, privileged account to have
a user mapping that could be provided as in the following (rather exagerated)
example.

  PASS  /~system/@  /sys$common/sysmgr/www/@
  USER  /~@/@  /@/www/
  PASS  /~@/@  /user$disk/@/www/@


Path SETtings
-------------
The SET rule allows characteristics to be set against a particular path.  They
are then transfered to any request matching the path and applied to the
request.  This applies to FILES, not necessarily to scripts that supply a full
HTTP response.  These need to check and act on information provided by relevant
CGI variables.  Note that as of V8.0 path SETtings can be applied with any rule
that has a result string (i.e. the setting must follow both a template and a
result and preceded any conditional).

[NO]ACCEPT=LANG=([DEFAULT=<language>][,CHAR=<char>][,VARIANT=NAME|TYPE])
                path will attempt to resolve a language-specific document,
                with default language (optional), 'char' is a character
                (optional) separating language from file type (default '_'),
                if the path template has a file type wildcard included the
                file content-type does not need to be "text/.."

[NO]ALERT         indicate that a request with this path has been detected,
                  HTTPDMON and OPCOM alert, and the process log gets a record
ALERT=MAP         alert provided immediately after mapping,
ALERT=AUTH        or immediately after authorization,
ALERT=END         or provided at request rundown (the default)

[NO]AUTH=ONCE     authorize only the script portion of the request,
                  do not repeat authorization on any path component
[NO]AUTH=MAPPED   authorize against the mapped path

AUTH=REVALIDATE=<hh:mm:ss>  revalidate authentication for this path
                            at the specified interval

AUTH=SYSUAF=PWDEXPURL=<string>  redirection URL if s SYSUAF-authenticated
                                user's primary password has expired

[NO]CACHE       files in the path should be cached (if at all possible,
                is overridden by a NOCACHE on the same line but overrides
                an NOCACHE set on any previous applicable rule)

CGIPREFIX=<string>  instead of the default "WWW_" specify what is needed,
                    to have no prefix at all specify CGIPREFIX=

CHARSET=<string>  any "text/..." content-type files use the specified charset

CONTENT=        all files have the "Content-Type:" forced to that specified
                (obvious use is to provide plain-text or binary versions of
                files using perhaps "set /plain-text/@ content=text/plain",
                then mapping that generally back with "map /plain-text/@ /@")
                another: "set /bogus/@ "content=text/html; charset=BLAH"

DIR=ACCESS            directory listing enabled  (same as [DirAccess])
DIR=ACCESS=SELECTIVE  allowed if directory contains .WWW_BROWSABLE
DIR=NOACCESS          directory listing disabled

DIR=CHARSET=<string>  directory listing page is given this charset
                      (this could be called the Alex Ivanov rule ;^)

DIR=LAYOUT=     same as index=<string>

DIR=IMPLIEDWILDCARD    implied wildcard enabled
DIR=NOIMPLIEDWILDCARD  implied wildcard disabled

DIR=WILDCARD    wildcard-forced listing enabled   (same as [DirWildcard])
DIR=NOWILDCARD  wildcard-forced listing disabled

[NO]EXPIRED     files in the path are sent "pre-expired"

HTTP=ACCEPT-CHARSET=<string>   explicitly set the request's accepted charset
HTTP=ACCEPT-LANGUAGE=<string>  explicitly set the request's accepted language

INDEX=<string>  "Index of" format string (same as "?httpd=index&...")

[NO]LOG         requests to this path are not to be access-logged

[NO]MAP=EMPTY     traditionally a path must have contained at least one char
                  to be mapped, this allows "empty" paths to be mapped
[NO]MAP=ONCE      normally, when a script has been derived from a path, WASD
                  goes on to map the 'path' portion in a separate pass - this
                  SETting provides the derived path as-is
[NO]MAP=[NO]ELLIPSIS  allows mapping of the VMS '...' ellipsis wildcard that
                      provides for tree traversal

ODS=2           path marked as ODS-2 (path default)
ODS=5           path marked as ODS-5, with EFS encoded file names
ODS=ADS         path marked as having filenames Advanced Server (6) encoded
ODS=PWK         path marked as having filenames PATHWORKS (4 & 5) encoded
ODS=SMB         same as ODS=ADS (only for the comfort of Samba users)
ODS=SRI         path marked as having filenames SRI (MultiNet NFS) encoded
NOODS           reset the path's ODS file naming

[NO]PROFILE     VMS SYSUAF-authenticated security profile applies/does not
                apply to this path

PROXY=BIND=<string>    bind the outgoing proxy socket to this IP address
PROXY=CHAIN=<string>   chain to this proxy server

REPORT=BASIC     supply basic error information for requests on this path
REPORT=DETAILED  supply detailed error information for requests on this path
REPORT=400=<integer>   bad request reported as (e.g. "report=400=403")
REPORT=403=<integer>   forbidden reported as (e.g. "report=403=404")
REPORT=404=<integer>   not found reported as (e.g. "report=404=403")

RMSCHAR=<char>  provides the RMS-invalid substitution character, which unless
                otherwise specified is the dollar symbol. Any general
                RMS-valid character may be specified (e.g. alpha-numeric,
                '$', '-' or '_', although the latter three are probably the
                only REAL choices).  When an RMS-invalid character (e.g. '+')
                or syntax (e.g. multiple periods) is encountered this character
                is substituted.  NOTE: is not applied to ODS-5 volumes.

SCRIPT=AS=<user>  the script process will be using this username/account
                  (only applies if detached script processes are enabled)
SCRIPT=BIT-BUCKET=<hh:mm:ss>  per-path equivalent to config file setting
SCRIPT=CPU=<hh:mm:ss>  CPU limit for each *script's* processing
                       (not a per-process limit like CPULM would be)
SCRIPT=FIND     (default) search for the script file in any 'script' or 'exec'
SCRIPT=NOFIND   do not search for the script file (for use with RTEs)
SCRIPT=PARAMS=(name=value)  a list of additional (non-CGI) variables assigned
                            and passed to a script

SSI=PRIV        mark a path as allowed to contain SSI documents that can
SSI=NOPRIV      use privileged SSI statements (e.g. "#dcl")

SSLCGI=NONE     sets the style of SSL-related CGI variables 
SSLCGI=APACHE_MOD_SSL      
SSLCGI=PURVEYOR      
[NO]SSLCGI

[NO]STMLF       files in this path can be converted from variable to stream-LF

THROTTLE=FROM=             n1, requests after which queuing occurs
THROTTLE=TO=               n2, requests after which the queue is FIFOed
THROTTLE=RESUME=           n3, requests after which FIFOing stops
THROTTLE=BUSY=             n4, requests after which an immediate 503 "busy"
THROTTLE=TIMEOUT=QUEUE=    t/o1, hh:mm:ss before a queued request processes
THROTTLE=TIMEOUT=BUSY=     t/o2, hh:mm:ss before a queued request terminates
THROTTLE=n1[,n2,n3,n4,t/o1,t/o2]  n1 = 'from',
                                  n2 = 'to',
                                  n3 = 'resume',
                                  n4 = 'busy',
                                  t/o1 = 'timeout-queue',
                                  t/o2 = 'timeout-busy'

TIMEOUT=KEEPALIVE=   hh:mm:ss  (can be made "none" to disable)
TIMEOUT=NOPROGRESS=  hh:mm:ss
TIMEOUT=OUTPUT=      hh:mm:ss
TIMEOUT=n,n,n        hh:mm:ss; keepalive, noprogress, output


PERSONA-based User Scripting
----------------------------
For use with VMS V6.2 and greater (or 6.0 and 6.1 if compiled with the
PERSONA_MACRO build option), with the /PERSONA qualifier active.
The following rule set will allow user account based scripting.

  SET   /~@/www/cgi-bin/@  script=as=~
  UXEC  /~@/cgi-bin/@  /@/www/cgi-bin/@
  USER  /~@/@  /@/www/@
  REDIRECT  /~@  /~@/
  PASS  /~@/@  /dka0/users/@/@


DECnet-based User Scripting
---------------------------
If DECnet/OSU scripting is enabled then the following rule will activate a
script in the specified user's directory (all other things being equal).
Always map to some unique and specific subdirectory of the home directory.
(In these examples commercial-at symbols substitute for wildcard asterisks ..
we are in a C-language comment, after-all!!)

exec /~@/cgi-bin/@ /0""::/where/ever/@/cgi-bin/@

It's a little like the user-mapping rules:

pass /~@ /where/ever/@


Mapping Conditionals
--------------------

 ***************
 * OBSOLESCENT *
 ***************

 ALTHOUGH MAPPING RULE CONDITIONALS DESCRIBED BELOW WILL CONTINUE
 TO BE SUPPORTED IN THE FORESEEABLE FUTURE THE PREFERED POST-7.2
 MECHANISM FOR CONDITIONAL PROCESSING OF RULES IS PROVIDED BY THE
 META-CONFIG CONDITIONAL STRUCTURES AND DIRECTIVES.

Conditionals are new to v4.4 and allow mapping rules to be applied
conditionally! That is, the conditionals contain strings that are matched to
specific elements of the request and only if matched are the corresponding
rules applied. The conditionals may contain the '*' and '%' wildcards, and
optionally be negated by an '!' at the start of the conditional string.

Conditionals must follow the rule and are delimited by '[' and ']'. Multiple,
space-separated conditions may be included within one '[...]'. This behaves as
a logical OR (i.e. the condition is true if only one is matched). Multiple
'[...]' conditionals may be included against a rule. These act as a logical
AND (i.e. all must have at least one condition matched). The result of an
entire conditional may be optionally negated by prefixing the '[' with a '!'.

If a conditional is not met the line is completely ignored.  Conditional rules
are not used for 'backward' translation (i.e. from VMS to URL path) so another
rule must exist somewhere to do these mappings.  Conditional rules are only
used by the HTTPd server, they are completely ignored by any scripts using the
MapUrl.c module.

[!]ac:accept             ('Accept:')
[!]al:accept-language    ('Accept-Language:')
[!]as:accept-charset     ('Accept-Charset:')
[!]ck:cookie             ('Cookie:')
[!]ex:                   (extended file specification path)
[!]fo:host-name/address  ('Forwarded:', proxies/gateways)
[!]ho:host-name/address  (client's of course!)
[!]hm:host network mask  (client mask, see immediately below)
[!]me:http-method        (GET, POST, etc.)
[!]pi:path-info          (request path information)
[!]qs:query-string       (request query string)
[!]rf:refering-URL       ('Referer:')
[!]ru:string             (request URI - non-decoded path string)
[!]sc:scheme             (request scheme 'http:', 'http', 'https:' or 'https')
[!]sn:server-name        (server host name)
[!]sp:server-port        (server port)
[!]ua:user-agent         ('User-Agent:')
[!]ud:user-details       (authorization user details)
[!]vs:host-name/address  (virtual server, "Host:", destination)

The host-mask ('HM') directive is a dotted-decimal network address, a slash,
then a dotted-decimal mask.  For example "[HM:131.185.250.0/255.255.255.192]". 
This has a 6 bit subnet.  It operates by bitwise-ANDing the client host address
with the mask, bitwise-ANDing the network address supplied with the mask, then
comparing the two results for equality.  Using the above example the host
131.185.250.250 would be accepted, but 131.185.250.50 would be rejected.


VERSION HISTORY
---------------
24-SEP-2002  MGD  added 'PI' and 'RU' conditionals
16-SEP-2002  MGD  MapUrl_ElementsToVms() excise parent directory syntax,
                  SET path [no]map=ellipsis (off by default),
                  SET report=4nn=nnn for mapping HTTP status
14-SEP-2002  MGD  only use MapUrl_VmsUserName() path ODS if not already set,
                  SET dir=charset= directory listing charset mapping rule
10-SEP-2002  MGD  add en/decoding for Advanced Server (Samba) file names
15-AUG-2002  MGD  SET alert=map, alert=auth, alert=end variants for ALERT,
                  SET ods=2, ods=5, now ods=pwk and ods=sri for below,
                  rework ..UrlToVms() and ..VmsToUrl() to allow for mapping
                  to and from SRI encoding (NFS) and PATHWORKS encoding
                  (and VMS to URL functions no longer explicitly URL-encode)
07-AUG-2002  MGD  bugfix; 'script' and 'exec' MetaConParseReset() state
04-AUG-2002  MGD  allow for 'pass /* 400' (i.e. no trailing message),
                  bugfix; template/result wildcard checking for scripting rules
31-MAY-2001  MGD  SET dir=access, dir=noaccess, dir=access=selective,
                  dir=impliedwildcard, dir=noimpliedwildcard,
                  dir=wildcard, dir=nowildcard
13-MAY-2002  MGD  SET auth=SYSUAF=pwdexpURL=
07-APR-2002  MGD  exchange DEVICE:[DIRECTORY]FILE.TYPE for the hopefully
                  more informative NO:[REVERSE.MAPPING.FOR.THIS]FILE.PATH
14-MAR-2002  NGD  bugfix; throttle report (jf.pieronne@laposte.net)
14-JAN-2002  MGD  added '**' and '*!' wildcard match and discard to template,
                  eliminate device:[.directory] (multiple consecutive '/'),
                  bugfix; arrghhh - wildcard substitution in MapUrl__Map()
18-NOV-2001  MGD  path SETtings may now trail any rule with a result string,
                  and are applied to the matched path before mapping occurs,
                  SET http=accept-charset=, accept=lang
28-OCT-2001  MGD  bugfix; DECnet user script mapping
13-OCT-2001  MGD  ensure a conditional is not mistaken for a missing template,
                  script=params=(name=value[,name="value1 value2"]),
                  proxy=bind=IP-address, proxy=chain=host:port, alert,
                  auth=revalidate=hh:mm:ss,
                  bugfix; wildcard substitution in MapUrl__Map()
25-AUG-2001  MGD  meta-config (even workin' on my birthday, sigh)
04-AUG-2001  MGD  support module WATCHing
15-MAY-2001  MGD  bugfix; remove 'RU' conditional (mapping before auth!)
10-MAY-2001  MGD  modify throttle parameter meanings,
                  set rule SCRIPT=CPU= for limiting script process CPU
13-APR-2001  MGD  add queue length to THROTTLE=n[,n,n,hh:mm:ss,hh:mm:ss]
                  add SCRIPT=BIT-BUCKET=hh:mm:ss,
                  bugfix; conditions for NetThisVirtualService() call
05-APR-2001  MGD  bugfix; 'HM' network mask processing
28-MAR-2001  MGD  EXEC of file type
13-MAR-2001  MGD  set rules THROTTLE=n[,n,n,n] & TIMEOUT=n,n,n
                  "hh:mm:ss" for providing a more versatile period
28-FEB-2001  MGD  OdsLoadTextFile(), OdsParseTextFile(), [IncludeFile]
22-DEC-2000  MGD  use of backslash to escape non-wildcard reserved characters
                  in rule and conditional strings (i.e. \] for ])
19-NOV-2000  MGD  bugfix; mapping '/'
10-NOV-2000  MGD  bugfix; MapUrl_UrlToVms() final ';'
01-OCT-2000  MGD  set rule SCRIPT=AS=,
                  rework mapping code somewhat,
                  MapUrl_VmsUserNameCache() on reset free entries
31-AUG-2000  MGD  bugfix; MapUrl_UrlToVms() escaped version semi-colon
24-JUN-2000  MGD  add network-mask to conditionals,
                  persistant run-time environment in EXEC rule (...),
                  change 'HH' to 'VS' (virtual service) backward compatible
06-APR-2000  MGD  set rules SSLCGI= and MAP=
04-MAR-2000  MGD  use NetWriteFaol(), et.al.
02-JAN-2000  MGD  config file opened via ODS module,
                  detection of a path's on-disk-structure (ODS-2 or ODS-5),
                  modify URL->VMS and VMS->URL mapping for extended syntax,
                  avoid confusion; redirect indicators changed from '^' and
                  '@' for local and remote, to '<' and '>' respectively,
                  bugfix; trailing wildcard in set rule comparison
24-OCT-1999  MGD  groan; ensure storage does not overflow!!
                  (plus copious quantities of self-flagellation),
                  SYSUAF-based user directory mapping via a "USER" rule,
                  CGIprefix, [no]profile and [no]authonce SET rules,
                  happy seventh wedding anniversary
12-SEP-1999  MGD  virtual services modifications
10-MAY-1999  MGD  conditional mapping on HTTP cookie and referer,
                  increase the size of scratch storage,
                  mapping endless-loop detection
16-JAN-1999  MGD  allow for proxy service mappings by removing requirement
                  for template and result paths to be absolute ("/...") paths
                  (every time I have to deal with this module I groan)
10-JAN-1999  MGD  bugfix; pass rules not returning mapped path
17-OCT-1998  MGD  SET mapping rule,
                  virtual services via "[[virtual-host:virtual-port]]",
                  "qs:" query string conditional,
                  invalid RMS file name/character substitution configurable
12-AUG-1998  MGD  MapUrl_VmsToUrl() now only optionally absorbs MFD
19-JUL-1998  MGD  'E'xec DECnet-based user scripts
10-MAR-1998  MGD  allow for local redirects (using '^' detect character)
07-FEB-1998  MGD  added "sc:" scheme/protocol conditional for SSL support
20-DEC-1997  MGD  added DECnet support in _UrlToVms(),
                  result paths can now begin with an '*'
25-OCT-1997  MGD  added "as:", "fo:", and "hh:" conditionals,
                  bugfix; MsgFor() required request pointer to find langauge
15-SEP-1997  MGD  MsgFor() was not returning a leading-null message (does now),
                  HTTP status code rules (e.g. "pass /no/* 403 forbidden"),
                  removed two consecutive slashes inhibiting rule mapping
07-SEP-1997  MGD  added "sn:" and "sp:" conditionals
09-AUG-1997  MGD  v4.3, message database, mapping "conditionals",
                  more cart-wheels, hand-stands and contortions (see 01-OCT-96)
05-JUN-1997  MGD  v4.2, CGIplus "exec+" and "script+" rules
01-FEB-1997  MGD  HTTPd version 4, major changes for form-based config
01-OCT-1996  MGD  not proud of how I shoe-horned the reporting features in,
                  good example of slop-down or object-disoriented programming,
                  :^( will rework the whole thing one day, sigh!
                  bugfix; requiring use of 'TheFirstRuleWillBeTheNextRule'
06-APR-1996  MGD  modified conversion of non-VMS characters in MapUrl_UrlToVms()
15-FEB-1995  MGD  bugfix; redirect rule
01-DEC-1995  MGD  v3.0
24-MAR-1995  MGD  bugfix; end-of-string sometimes not detected when matching
20-DEC-1994  MGD  revised for multi-threaded HTTP daemon
20-JUN-1994  MGD  single-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <devdef.h>
#include <dvidef.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <uaidef.h>

#include "wasd.h"

#define WASD_MODULE "MAPURL"

/**********/
/* macros */
/**********/

#define MAPURL_RMS_SUBSTITUTION_DEFAULT '$'

/* this allows indices derived from the digits '1'..'9', plus some */
#define MAPURL_MAX_WILDCARDS 20

/******************/
/* global storage */
/******************/

BOOL  MapUrlPathOds5;

MAPPING_META  MappingMeta;
MAPPING_META  *MappingMetaPtr;

#define MAPURL_DEFAULT_USER_CACHE_SIZE  32

LIST_HEAD  MapUrlUserNameCacheList;
int  MapUrlUserNameCacheCount,
     MapUrlUserNameCacheEntries;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern BOOL Debug;
#else
#define Debug 0 
#endif

extern BOOL  AuthPolicySysUafRelaxed,
             CliOdsExtendedDisabled,
             CliOdsExtendedEnabled,
             OdsExtended;

extern int  EfnWait,
            ServerPort,
            ServerHostNameLength,
            ServiceCount;

extern unsigned long  SysPrvMask[];

extern char  ErrorWatchSysFao[],
             ServerHostName[],
             ServerHostPort[],
             SoftwareID[],
             Utility[];

extern CONFIG_STRUCT  Config;
extern META_CONFIG  *MetaGlobalMappingPtr;
extern MSG_STRUCT  Msgs;
extern SYS_INFO  SysInfo;
extern WATCH_STRUCT  Watch;

/*****************************************************************************/
/*
This overly-long function (sorry) is not AST reentrant, but when executed 
within the context of the HTTPd, is executed either without the possibility of 
AST interruption, or during AST processing and so provides atomic 
functionality.  Pointers to any strings returned must be used immediately 
(before leaving AST context), or copied into storage maintained by the calling 
routine. 

Maps from URL format to VMS file specification, and REVERSE-MAPS from VMS file 
specification to URL format.  Maps scripts.

Always returns a pointer to char.  An error message is detected by being  
returned as a pointer to a string beginning with a null-character!  The 
message begins from character one. 

Call with 'PathPtr' pointing at a string containing the URL path and 'VmsPtr' 
pointing to the storage to contain the mapped VMS equivalent.

Call with 'VmsPtr' pointing to a VMS file specification (dev:[dir]file.ext) 
and 'PathPtr' pointing to an empty string, used as storage to contain the 
REVERSE-MAPPED file specification.

Call with 'PathPtr' pointing at a string containing the URL path, 'VmsPtr' 
pointing to storage to contain the mapped VMS equivalent, 'ScriptPtr' pointing 
to storage for the URL format script path and 'ScriptVmsPtr' pointing to 
storage to contain the VMS specification of the script procedure/image.

The MapUrl_Map() just wraps MapUrl__Map() so that it's function can be WATCHed.

See alter-egos defined in MapUrl.h
*/ 

char* MapUrl_Map
(
char *PathPtr,
int SizeOfPathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
char *ScriptPtr,
int SizeOfScriptPtr,
char *ScriptVmsPtr,
int SizeOfScriptVmsPtr,
char *RteVmsPtr,
int SizeOfRteVmsPtr,
int *PathOdsPtr,
REQUEST_STRUCT *rqptr
)
{
   BOOL  VmsToPath;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_Map()\n");

   if (!rqptr || !WATCHING(rqptr))
     return (MapUrl__Map (PathPtr, SizeOfPathPtr,
                          VmsPtr, SizeOfVmsPtr,
                          ScriptPtr, SizeOfScriptPtr,
                          ScriptVmsPtr, SizeOfScriptVmsPtr,
                          RteVmsPtr, SizeOfRteVmsPtr,
                          PathOdsPtr, rqptr));

   if (!WATCH_CATEGORY(WATCH_MAPPING))
     return (MapUrl__Map (PathPtr, SizeOfPathPtr,
                          VmsPtr, SizeOfVmsPtr,
                          ScriptPtr, SizeOfScriptPtr,
                          ScriptVmsPtr, SizeOfScriptVmsPtr,
                          RteVmsPtr, SizeOfRteVmsPtr,
                          PathOdsPtr, rqptr));

#if WATCH_CAT

   VmsToPath = false;
   if (PathPtr && PathPtr[0])
      WatchThis (rqptr, FI_LI, WATCH_MAPPING, "PATH !AZ", PathPtr); 
   else
   if (VmsPtr && VmsPtr[0])
   {
      VmsToPath = true;
      WatchThis (rqptr, FI_LI, WATCH_MAPPING, "VMS !AZ", VmsPtr); 
   }

   cptr = MapUrl__Map (PathPtr, SizeOfPathPtr,
                       VmsPtr, SizeOfVmsPtr, 
                       ScriptPtr, SizeOfScriptPtr,
                       ScriptVmsPtr, SizeOfScriptVmsPtr,
                       RteVmsPtr, SizeOfRteVmsPtr,
                       PathOdsPtr, rqptr);

   if (cptr[0] || cptr[1])
   {
      if (VmsToPath)
         WatchThis (rqptr, FI_LI, WATCH_MAPPING, "RESULT !AZ", cptr);
      else
      {
         if (!PathPtr) PathPtr = "";
         if (!VmsPtr) VmsPtr = "";
         if (!ScriptPtr) ScriptPtr = "";
         if (!ScriptVmsPtr) ScriptVmsPtr = "";
         if (!RteVmsPtr) RteVmsPtr = "";

         WatchThis (rqptr, FI_LI, WATCH_MAPPING, "RESULT");
         WatchDataFormatted (
"!AZ!AZ!AZ !AZ\n\
Translated: !AZ\n\
    Script: !AZ\n\
  Location: !AZ\n\
       RTE: !AZ\n",
            *cptr == '\0' ? "     Error:" : "",
            *cptr == '\1' ? "  Redirect:" : "",
            *cptr > '\1'  ? "    Mapped:" : "",
            *cptr > '\1'  ?          cptr : cptr+1,
            VmsPtr, ScriptPtr, ScriptVmsPtr, RteVmsPtr);
      }
   }
   else
      WatchThis (rqptr, FI_LI, WATCH_MAPPING, "RESULT \"!AZ\"", cptr+1);

   return (cptr);

#endif /* WATCH_CAT */
}

char* MapUrl__Map
(
char *PathPtr,
int SizeOfPathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
char *ScriptPtr,
int SizeOfScriptPtr,
char *ScriptVmsPtr,
int SizeOfScriptVmsPtr,
char *RteVmsPtr,
int SizeOfRteVmsPtr,
int *PathOdsPtr,
REQUEST_STRUCT *rqptr
)
{
   static char  DerivedPathBuffer [1024+128],
                PathBuffer [1024+128],
                VmsBuffer [1024+128];

   BOOL  ExecFileType,
         LoadingRules,
         MapEllipsis,
         MapPathToVms,
         PathOdsExtended,
         StatusCodeMapping,
         SubstituteForUserName,
         WatchThisOne;
   int  idx, status,
        MapPassCount,
        PathOds,
        TotalLength,
        WildBufferIndex;
   unsigned short  Length;
   char  ch;
   char  *cptr, *pptr, *sptr, *tcptr, *tpptr, *uptr, *zptr;
   /* this allows indices derived from the digits '1'..'9' */
   char  *WildBufferArray [MAPURL_MAX_WILDCARDS];
   char  RmsSubChar;
   char  Scratch [1024+128],
         UserDefault [256],
         VmsUserMappedBuffer [256],
         WildBuffer [1024+128];
   MAP_RULE_META  *mrptr;
   METACON_LINE  *mclptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_Map() !UL !&Z !UL !&Z !UL !&Z !UL",
                 SizeOfPathPtr, PathPtr, SizeOfVmsPtr, VmsPtr,
                 SizeOfScriptPtr, ScriptPtr, SizeOfScriptVmsPtr);

   /* if just loading the rules then free any previous set */
   if (LoadingRules = (!PathPtr && !VmsPtr && !ScriptPtr))
      MetaConUnload (&MetaGlobalMappingPtr, NULL);

   if (!MetaGlobalMappingPtr)
   {
      /*********************/
      /* read mapping file */
      /*********************/

      MapUrl_ConfigLoad (&MetaGlobalMappingPtr);
   }

   /* if just (re)loading rules then return now */
   if (LoadingRules) return ("\0\0");

   /*****************************************/
   /* map from the URL or VMS to VMS or URL */
   /*****************************************/

   if (!PathPtr)
   {
      *(unsigned short*)PathBuffer = '\0\0';
      PathPtr = PathBuffer + 1;
      SizeOfPathPtr = sizeof(PathBuffer);
   }
   if (!VmsPtr)
   {
      *(VmsPtr = VmsBuffer) = '\0';
      SizeOfVmsPtr = sizeof(VmsBuffer);
   }
   /* making 'ScriptPtr' non-empty means "exec" and "script" are ignored */
   if (!ScriptPtr) ScriptPtr = "?";

   if (PathPtr[0])
   {
      /* if the URL is not an empty string then force the conversion to VMS */
      MapPathToVms = true;
      *VmsPtr = '\0';
   }
   else
   {
      /* URL was an empty string, convert from VMS to URL */
      MapPathToVms = false;
      /* generate a URL-style version of the VMS specification */
      MapUrl_VmsToUrl (PathPtr, VmsPtr, SizeOfPathPtr, true,
                       rqptr->rqPathSet.OdsNaming);
   }

   /***********************/
   /* loop thru the rules */
   /***********************/

   if (!MetaGlobalMappingPtr)
      return (MsgFor(rqptr,MSG_MAPPING_DENIED_NO_RULES)-1);

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_MAPPING))
      WatchThisOne = true;
   else
      WatchThisOne = false;

   MapPassCount = PathOds = PathOdsExtended = 0;
   RmsSubChar = MAPURL_RMS_SUBSTITUTION_DEFAULT;
   MapEllipsis = false;

   MetaConParseReset (MetaGlobalMappingPtr, true);
   for (;;)
   {
      mclptr = MetaGlobalMappingPtr->ParsePtr =
         MetaGlobalMappingPtr->ParseNextPtr;

      if (WATCH_MODULE(WATCH_MOD_METACON) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!&X !UL !UL !UL !UL !&X !&Z\n",
            mclptr, mclptr->Size, mclptr->Token, mclptr->Number,
            mclptr->Length, mclptr->LineDataPtr, mclptr->TextPtr);

      /* if terminating empty "line" */
      if (!mclptr->Size) break;
      if (mclptr->Token == METACON_TOKEN_TEXT)
      {
         /* get required data, adjust the parse context to the next "line" */
         cptr = mclptr->TextPtr;
         mrptr = mclptr->LineDataPtr;
         MetaGlobalMappingPtr->ParseNextPtr =
            (METACON_LINE*)((char*)mclptr + mclptr->Size);
      }
      else
      {
         /* not a simple text line, have meta-config parse this one for us */
         cptr = MetaConParse (rqptr, MetaGlobalMappingPtr, &mclptr,
                              WatchThisOne);
         /* if end of rules */
         if (!cptr) break;
         /* if error string */
         if (!*cptr && *(cptr+1)) return (cptr);
         /* if inline rule and expression was false */
         if (*(unsigned long*)cptr == '\0\0\0\1') continue;
         mrptr = mclptr->LineDataPtr;
      }

      /* there was a problem in initializing the rule */
      if (!mrptr)
      {
         if (WatchThisOne) WatchDataFormatted ("?!AZ\n", cptr);
         continue;
      }

      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!&X !UL !&Z !&Z !&Z\n", 
            mrptr, mrptr->RuleType, mrptr->TemplatePtr,
            mrptr->ResultPtr, mrptr->ConditionalPtr);

      /* hmmm, endless loop!! */
      if (MapPassCount > 2)
         return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);

      /* when mapping from VMS to URL then ignore all BUT "pass" rules */
      if (!MapPathToVms && mrptr->RuleType != MAPURL_RULE_PASS) continue;

      /* if 'ScriptPtr' is not to be used then ignore "exec" and "script" */
      if (ScriptPtr[0] &&
          (mrptr->RuleType == MAPURL_RULE_EXEC ||
           mrptr->RuleType == MAPURL_RULE_SCRIPT ||
           mrptr->RuleType == MAPURL_RULE_UXEC))
         continue;

      /*****************************************************************/
      /* compare the URL with the template/result in the mapping entry */
      /*****************************************************************/

      if (mrptr->RuleType == MAPURL_RULE_SET)
      {
         /************/
         /* SET rule */
         /************/

         if (!StringMatch (PathPtr, mrptr->TemplatePtr, false))
         {
            if (!WatchThisOne) continue;
            MapUrl_WatchRule (rqptr, mrptr, PathPtr, PathOds,
                              MAPURL_REPORT_MATCH_NOT);
            continue;
         }
      }
      else
      {
         /******************************/
         /* other rules, build buffers */
         /******************************/

         /* if reverse-mapping VMS to URL use the result if available */
         if (!MapPathToVms && mrptr->ResultPtr[0])
            cptr = mrptr->ResultPtr;
         else
            cptr = mrptr->TemplatePtr;
         pptr = PathPtr;

/**
         if (WATCH_MODULE(WATCH_MOD_MAPURL))
            WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "!&Z !&Z", pptr, cptr);
**/

         /* 'WildBuffer' contains null-separated wildcard-matched portions */
         zptr = (sptr = WildBuffer) + sizeof(WildBuffer)-32;

         /* reset the array of pointers into 'WildBuffer' */
         for (idx = 0; idx < MAPURL_MAX_WILDCARDS; idx++)
            WildBufferArray[idx] = NULL;
         WildBufferIndex = 0;

         /* extract the wildcarded sections */
         while (*cptr && *pptr)
         {
            while (*cptr && *pptr &&
                   tolower(*pptr) == tolower(*cptr) &&
                   *cptr != '*')
            {
               cptr++;
               pptr++;
            }
            if (*cptr != '*') break;

            if (WildBufferIndex < MAPURL_MAX_WILDCARDS)
               WildBufferArray[WildBufferIndex++] = sptr;
            else
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);

            /* wildcard in template, copy matched portion to wildcard buffer */
            if (*(unsigned short*)cptr == '**')
            {
               /*****************************/
               /* two consecutive asterisks */
               /*****************************/

               while (*cptr == '*') cptr++;
               if (!*cptr)
               {
                  /* trailing wildcard, just copy rest of path */
                  while (*pptr && sptr < zptr) *sptr++ = *pptr++;
               }
               else
               {
                  /* copy to first string matching following the wildcard */
                  ch = tolower(*cptr);
                  while (*pptr && sptr < zptr)
                  {
                     while (*pptr && tolower(*pptr) != ch && sptr < zptr)
                        *sptr++ = *pptr++;
                     if (!*pptr || sptr >= zptr) break;
                     /* check trailing string (or up to a wildcard) matches */
                     tcptr = cptr;
                     tpptr = pptr;
                     while (*cptr && *pptr && *cptr != '*' &&
                            tolower(*cptr) == tolower(*pptr))
                     {
                        cptr++;
                        pptr++;
                     }
                     if ((!*cptr && !*pptr) || *cptr == '*') break;
                     /* mismatch of trailing string, continue copy wildcard */
                     cptr = tcptr;
                     pptr = tpptr;
                     if (sptr < zptr) *sptr++ = *pptr++;
                  }
               }
            }
            else
            {
               /****************/
               /* one asterisk */
               /****************/

               while (*cptr == '*') cptr++;
               if (!*cptr)
               {
                  /* trailing wildcard, just copy rest of path */
                  while (*pptr && sptr < zptr) *sptr++ = *pptr++;
               }
               else
               {
                  /* copy to first char matching the one following wildcard */
                  ch = tolower(*cptr);
                  while (*pptr && tolower(*pptr) != ch && sptr < zptr)
                        *sptr++ = *pptr++;
                  if (!*pptr || sptr >= zptr) break;
                  /* check the trailing string (or up to a wildcard) matches */
                  while (*cptr && *pptr && *cptr != '*' &&
                         tolower(*cptr) == tolower(*pptr))
                  {
                     cptr++;
                     pptr++;
                  }
               }
            }

            if (sptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            memset (sptr, 0, 32);
            sptr++;
         }

         /* if not matched then straight to next rule */
         if ((*cptr && *(unsigned short*)cptr != '*\0') || *pptr)
         {
            if (!WatchThisOne) continue;
            MapUrl_WatchRule (rqptr, mrptr, PathPtr, PathOds,
                              MAPURL_REPORT_MATCH_NOT);
            continue;
         }
      }

      /************/
      /* matched! */
      /************/

      if (WATCH_MODULE(WATCH_MOD_MAPURL))
      {
         WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "MATCHED");
         for (idx = 0; idx < WildBufferIndex; idx++)
             WatchDataFormatted ("!UL !&Z\n", idx+1, WildBufferArray[idx]);
      }

      if (mrptr->ConditionalPtr[0])
      {
         /* when not mapping using a request then ignore conditionals */
         if (rqptr)
         {
            /* if condition not met then continue */
            if (!MapUrl_Conditional ((REQUEST_STRUCT*)rqptr, mrptr))
            {
               if (WatchThisOne)
                  MapUrl_WatchRule (rqptr, mrptr, PathPtr, PathOds,
                                    MAPURL_REPORT_MATCH_RULE);
               continue;
            }

            /* report rule and conditional matched */
            if (WatchThisOne)
               MapUrl_WatchRule (rqptr, mrptr, PathPtr, PathOds,
                                 MAPURL_REPORT_MATCH_RULECOND);
         }

         if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
            WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "CONDITIONAL MATCHED");
      }
      else
      {
         /* report rule, there was no conditional */
         if (WatchThisOne)
            MapUrl_WatchRule (rqptr, mrptr, PathPtr, PathOds,
                              MAPURL_REPORT_MATCH_RULENOCOND);
      }

      if (mrptr->PathSet)
      {
         /*****************/
         /* path SETtings */
         /*****************/

         if (mrptr->mpPathSet.NoAcceptLang)
         {
            rqptr->rqPathSet.AcceptLangChar = '\0';
            rqptr->rqPathSet.AcceptLangWildcard =
               rqptr->rqPathSet.AcceptLangTypeVariant = false;
            rqptr->rqPathSet.AcceptLangPtr = NULL;
         }
         else
         if (mrptr->mpPathSet.AcceptLangChar)
         {
            rqptr->rqPathSet.AcceptLangChar = mrptr->mpPathSet.AcceptLangChar;
            rqptr->rqPathSet.AcceptLangWildcard =
               mrptr->mpPathSet.AcceptLangWildcard;
            rqptr->rqPathSet.AcceptLangTypeVariant =
               mrptr->mpPathSet.AcceptLangTypeVariant;
            if (mrptr->mpPathSet.AcceptLangLength)
            {
               rqptr->rqPathSet.AcceptLangPtr =
                  VmGetHeap (rqptr, mrptr->mpPathSet.AcceptLangLength+1);
               memcpy (rqptr->rqPathSet.AcceptLangPtr,
                       mrptr->mpPathSet.AcceptLangPtr,
                       mrptr->mpPathSet.AcceptLangLength+1);
            }
         }

         if (mrptr->mpPathSet.NoAlert)
            rqptr->rqPathSet.Alert = false;
         else
         if (mrptr->mpPathSet.Alert)
            rqptr->rqPathSet.Alert = mrptr->mpPathSet.Alert;

         if (mrptr->mpPathSet.NoAuthMapped)
            rqptr->rqPathSet.AuthorizeMapped = false;
         else
         if (mrptr->mpPathSet.AuthMapped)
            rqptr->rqPathSet.AuthorizeMapped = true;
   
         if (mrptr->mpPathSet.NoAuthOnce)
            rqptr->rqPathSet.AuthorizeOnce = false;
         else
         if (mrptr->mpPathSet.AuthOnce)
            rqptr->rqPathSet.AuthorizeOnce = true;
   
         /* only if they haven't been set in any previous pass */
         if (!rqptr->rqPathSet.AuthRevalidateTimeout)
            rqptr->rqPathSet.AuthRevalidateTimeout =
               mrptr->mpPathSet.AuthRevalidateTimeout;

         if (mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr &&
             mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr[0])
         {
            rqptr->rqPathSet.AuthSysUafPwdExpUrlPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.AuthSysUafPwdExpUrlLength+1);
            memcpy (rqptr->rqPathSet.AuthSysUafPwdExpUrlPtr,
                    mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr,
                    mrptr->mpPathSet.AuthSysUafPwdExpUrlLength+1);
         }
   
         if (mrptr->mpPathSet.NoCache)
            rqptr->rqPathSet.NoCache = true;
         else
         if (mrptr->mpPathSet.Cache)
            rqptr->rqPathSet.NoCache = false;
   
         if (mrptr->mpPathSet.DirNoAccess)
         {
            rqptr->rqPathSet.DirNoAccess = true;
            rqptr->rqPathSet.DirAccess =
               rqptr->rqPathSet.DirAccessSelective = false;
         }
         else
         if (mrptr->mpPathSet.DirAccess)
         {
            rqptr->rqPathSet.DirAccess = true;
            rqptr->rqPathSet.DirNoAccess =
               rqptr->rqPathSet.DirAccessSelective = false;
         }
         else
         if (mrptr->mpPathSet.DirAccessSelective)
         {
            rqptr->rqPathSet.DirAccessSelective = true;
            rqptr->rqPathSet.DirAccess =
               rqptr->rqPathSet.DirNoAccess = false;
         }

         if (mrptr->mpPathSet.DirNoImpliedWildcard)
            rqptr->rqPathSet.DirNoImpliedWildcard = true;
         else
         if (mrptr->mpPathSet.DirImpliedWildcard)
            rqptr->rqPathSet.DirNoImpliedWildcard = false;

         if (mrptr->mpPathSet.DirNoWildcard)
         {
            rqptr->rqPathSet.DirNoWildcard = true;
            rqptr->rqPathSet.DirWildcard = false;
         }
         else
         if (mrptr->mpPathSet.DirWildcard)
         {
            rqptr->rqPathSet.DirNoWildcard = false;
            rqptr->rqPathSet.DirWildcard = true;
         }

         if (mrptr->mpPathSet.NoExpired)
            rqptr->rqPathSet.Expired = false;
         else
         if (mrptr->mpPathSet.Expired)
            rqptr->rqPathSet.Expired = true;

         if (mrptr->mpPathSet.NoLog)
            rqptr->rqPathSet.NoLog = true;
         else
         if (mrptr->mpPathSet.Log)
            rqptr->rqPathSet.NoLog = false;

         if (mrptr->mpPathSet.NoMapEllipsis)
            rqptr->rqPathSet.MapEllipsis = MapEllipsis = false;
         else
         if (mrptr->mpPathSet.MapEllipsis)
            rqptr->rqPathSet.MapEllipsis = MapEllipsis = true;

         if (mrptr->mpPathSet.MapNonEmpty)
            rqptr->rqPathSet.MapEmpty = false;
         else
         if (mrptr->mpPathSet.MapEmpty)
            rqptr->rqPathSet.MapEmpty = true;

         if (mrptr->mpPathSet.MapOnce)
            rqptr->rqPathSet.MapOnce = true;
         else
         if (mrptr->mpPathSet.MapOnce)
            rqptr->rqPathSet.MapOnce = false;

         if (mrptr->mpPathSet.PrivSsi)
            rqptr->rqPathSet.PrivSsi = true;
         else
         if (mrptr->mpPathSet.PrivSsi)
            rqptr->rqPathSet.PrivSsi = false;
   
         if (mrptr->mpPathSet.NoProfile)
            rqptr->rqPathSet.NoProfile = true;
         else
         if (mrptr->mpPathSet.Profile)
            rqptr->rqPathSet.NoProfile = false;

         /* BASIC overrides DETAILED */
         if (mrptr->mpPathSet.ReportBasic)
            rqptr->rqPathSet.ReportBasic = true;
         else
         if (mrptr->mpPathSet.ReportDetailed)
            rqptr->rqPathSet.ReportBasic = false;

         if (mrptr->mpPathSet.Report400as)
            rqptr->rqPathSet.Report400as = mrptr->mpPathSet.Report400as;
         if (mrptr->mpPathSet.Report403as)
            rqptr->rqPathSet.Report403as = mrptr->mpPathSet.Report403as;
         if (mrptr->mpPathSet.Report404as)
            rqptr->rqPathSet.Report404as = mrptr->mpPathSet.Report404as;

         /* careful! this one's a little back-to-front */
         if (mrptr->mpPathSet.ScriptFind)
            rqptr->rqPathSet.ScriptNoFind = false;
         else
         if (mrptr->mpPathSet.ScriptNoFind)
            rqptr->rqPathSet.ScriptNoFind = true;

         if (!rqptr->rqPathSet.ThrottleFrom &&
              mrptr->mpPathSet.ThrottleFrom)
         {
            /* only if it hasn't been set in any previous pass */
            rqptr->rqPathSet.ThrottleBusy =
               mrptr->mpPathSet.ThrottleBusy;
            rqptr->rqPathSet.ThrottleFrom =
               mrptr->mpPathSet.ThrottleFrom;
            rqptr->rqPathSet.ThrottleIndex =
               mrptr->mpPathSet.ThrottleIndex;
            rqptr->rqPathSet.ThrottleResume =
               mrptr->mpPathSet.ThrottleResume;
            rqptr->rqPathSet.ThrottleTo =
               mrptr->mpPathSet.ThrottleTo;
            rqptr->rqPathSet.ThrottleTimeoutBusy =
               mrptr->mpPathSet.ThrottleTimeoutBusy;
            rqptr->rqPathSet.ThrottleTimeoutQueue =
               mrptr->mpPathSet.ThrottleTimeoutQueue;
         }

         /* only if they haven't been set in any previous pass */
         if (!rqptr->rqPathSet.ScriptBitBucketTimeout)
            rqptr->rqPathSet.ScriptBitBucketTimeout =
               mrptr->mpPathSet.ScriptBitBucketTimeout;

         if (!rqptr->rqPathSet.ScriptCpuMax)
            rqptr->rqPathSet.ScriptCpuMax =
               mrptr->mpPathSet.ScriptCpuMax;

         if (!rqptr->rqPathSet.TimeoutKeepAlive)
         {
            /* hmmm, bit of a kludge (but easy) */
            if (mrptr->mpPathSet.TimeoutKeepAlive < 0)
               rqptr->KeepAliveRequest = false;
            else
               rqptr->rqPathSet.TimeoutKeepAlive =
                  mrptr->mpPathSet.TimeoutKeepAlive;
         }

         if (!rqptr->rqPathSet.TimeoutNoProgress)
            rqptr->rqPathSet.TimeoutNoProgress =
               mrptr->mpPathSet.TimeoutNoProgress;
   
         if (!rqptr->rqPathSet.TimeoutOutput)
            rqptr->rqPathSet.TimeoutOutput =
                mrptr->mpPathSet.TimeoutOutput;
   
         if (mrptr->mpPathSet.NoStmLF)
            rqptr->rqPathSet.StmLF = false;
         else
         if (mrptr->mpPathSet.StmLF)
            rqptr->rqPathSet.StmLF = true;
   
         if (mrptr->mpPathSet.SSLCGIvar)
         {
            if (rqptr->rqPathSet.SSLCGIvar == SESOLA_CGI_VAR_NONE)
               rqptr->rqPathSet.SSLCGIvar = 0;
            else
               rqptr->rqPathSet.SSLCGIvar = mrptr->mpPathSet.SSLCGIvar;
         }

         if (mrptr->mpPathSet.OdsNaming)
         {
            if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_0)
               PathOds = rqptr->rqPathSet.OdsNaming = 0;
            else
               PathOds = rqptr->rqPathSet.OdsNaming =
                         mrptr->mpPathSet.OdsNaming;
            if (PathOdsPtr) *PathOdsPtr = PathOds;
#ifdef ODS_EXTENDED
            PathOdsExtended = (OdsExtended && PathOds == MAPURL_PATH_ODS_5);
#endif /* ODS_EXTENDED */
         }

         /* CGI prefix is allowed to be an empty string */
         if (mrptr->mpPathSet.CgiPrefixPtr)
         {
            rqptr->rqPathSet.CgiPrefixPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.CgiPrefixLength+1);
            memcpy (rqptr->rqPathSet.CgiPrefixPtr,
                    mrptr->mpPathSet.CgiPrefixPtr,
                    mrptr->mpPathSet.CgiPrefixLength+1);
         }

         if (mrptr->mpPathSet.CharsetPtr &&
             mrptr->mpPathSet.CharsetPtr[0])
         {
            rqptr->rqPathSet.CharsetPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.CharsetLength+1);
            memcpy (rqptr->rqPathSet.CharsetPtr,
                    mrptr->mpPathSet.CharsetPtr,
                    mrptr->mpPathSet.CharsetLength+1);
         }
   
         if (mrptr->mpPathSet.ContentTypePtr &&
             mrptr->mpPathSet.ContentTypePtr[0])
         {
            rqptr->rqPathSet.ContentTypePtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.ContentTypeLength+1);
            memcpy (rqptr->rqPathSet.ContentTypePtr,
                    mrptr->mpPathSet.ContentTypePtr,
                    mrptr->mpPathSet.ContentTypeLength+1);
         }

         if (mrptr->mpPathSet.DirCharsetPtr &&
             mrptr->mpPathSet.DirCharsetPtr[0])
         {
            rqptr->rqPathSet.DirCharsetPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.DirCharsetLength+1);
            memcpy (rqptr->rqPathSet.DirCharsetPtr,
                    mrptr->mpPathSet.DirCharsetPtr,
                    mrptr->mpPathSet.DirCharsetLength+1);
         }

         if (mrptr->mpPathSet.HttpAcceptCharsetPtr)
         {
            if (mrptr->mpPathSet.HttpAcceptCharsetPtr[0])
            {
               rqptr->rqHeader.AcceptCharsetPtr =
                  VmGetHeap (rqptr, mrptr->mpPathSet.HttpAcceptCharsetLength+1);
               memcpy (rqptr->rqHeader.AcceptCharsetPtr,
                       mrptr->mpPathSet.HttpAcceptCharsetPtr,
                       mrptr->mpPathSet.HttpAcceptCharsetLength+1);
            }
            else
            {
               /* empty, as if it had never been part of the request header */
               rqptr->rqHeader.AcceptCharsetPtr = NULL;
               rqptr->rqHeader.AcceptCharsetLength = 0;
            }
         }
   
         if (mrptr->mpPathSet.HttpAcceptLangPtr)
         {
            if (mrptr->mpPathSet.HttpAcceptLangPtr[0])
            {
               rqptr->rqHeader.AcceptLangPtr =
                  VmGetHeap (rqptr, mrptr->mpPathSet.HttpAcceptLangLength+1);
               memcpy (rqptr->rqHeader.AcceptLangPtr,
                       mrptr->mpPathSet.HttpAcceptLangPtr,
                       mrptr->mpPathSet.HttpAcceptLangLength+1);
            }
            else
            {
               /* empty, as if it had never been part of the request header */
               rqptr->rqHeader.AcceptLangPtr = NULL;
               rqptr->rqHeader.AcceptLangLength = 0;
            }
         }
   
         if (mrptr->mpPathSet.IndexPtr &&
             mrptr->mpPathSet.IndexPtr[0])
         {
            rqptr->rqPathSet.IndexPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.IndexLength+1);
            memcpy (rqptr->rqPathSet.IndexPtr,
                    mrptr->mpPathSet.IndexPtr,
                    mrptr->mpPathSet.IndexLength+1);
         }
   
         if (mrptr->mpPathSet.ScriptAsPtr &&
             mrptr->mpPathSet.ScriptAsPtr[0])
         {
            rqptr->rqPathSet.ScriptAsPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.ScriptAsLength+1);
            memcpy (rqptr->rqPathSet.ScriptAsPtr,
                    mrptr->mpPathSet.ScriptAsPtr,
                    mrptr->mpPathSet.ScriptAsLength+1);
         }

         if (mrptr->mpPathSet.ScriptParamsPtr &&
             mrptr->mpPathSet.ScriptParamsPtr[0])
         {
            rqptr->rqPathSet.ScriptParamsPtr =
               VmGetHeap (rqptr, mrptr->mpPathSet.ScriptParamsLength+1);
            memcpy (rqptr->rqPathSet.ScriptParamsPtr,
                    mrptr->mpPathSet.ScriptParamsPtr,
                    mrptr->mpPathSet.ScriptParamsLength+1);
         }

         if (mrptr->mpPathSet.RmsSubChar)
            RmsSubChar = mrptr->mpPathSet.RmsSubChar;

         if (!rqptr->rqPathSet.ProxyBindIpAddress &&
             mrptr->mpPathSet.ProxyBindIpAddress)
            rqptr->rqPathSet.ProxyBindIpAddress =
               mrptr->mpPathSet.ProxyBindIpAddress;

         if (!rqptr->rqPathSet.ProxyChainIpAddress &&
             mrptr->mpPathSet.ProxyChainIpAddress)
         {
            rqptr->rqPathSet.ProxyChainIpAddress =
               mrptr->mpPathSet.ProxyChainIpAddress;
            rqptr->rqPathSet.ProxyChainPort =
               mrptr->mpPathSet.ProxyChainPort;
            if (mrptr->mpPathSet.ProxyChainHostPortPtr &&
                mrptr->mpPathSet.ProxyChainHostPortPtr[0])
            {
               rqptr->rqPathSet.ProxyChainHostPortPtr =
                  VmGetHeap (rqptr,
                             mrptr->mpPathSet.ProxyChainHostPortLength+1);
               memcpy (rqptr->rqPathSet.ProxyChainHostPortPtr,
                       mrptr->mpPathSet.ProxyChainHostPortPtr,
                       mrptr->mpPathSet.ProxyChainHostPortLength+1);
            }
         }
      }

      if (MapPathToVms)
      {
         /***************************/
         /* mapping from URL to VMS */
         /***************************/

         switch (mrptr->RuleType)
         {
         case MAPURL_RULE_EXEC :
         case MAPURL_RULE_UXEC :

            /*******************/
            /* EXEC/UXEC rule */
            /*******************/

            /* first get the script component from the request */
            cptr = WildBuffer;
            sptr = mrptr->TemplatePtr;
            zptr = (pptr = ScriptPtr) + SizeOfScriptPtr;

            if (mrptr->RuleType == MAPURL_RULE_UXEC)
            {
               if (!rqptr->rqPathSet.ScriptAsPtr ||
                   rqptr->rqPathSet.ScriptAsPtr[0] != '~')
               {
                  /* hmmm, not enabled with a mapping rule!! */
                  return (MsgFor(rqptr,MSG_GENERAL_DISABLED)-1);
               }

               /* first insert the /~username component */
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
               if (*sptr) sptr++;
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
            }
            else
            if (*(unsigned short*)sptr == '/~')
            {
               /* mapping a DECnet username based script */
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
               while (*sptr == '*') sptr++;
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
            }

            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            if (*(unsigned short*)sptr == '*.')
            {
               /* EXECing a file type here */
               ExecFileType = true;
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
               /* now get the script file type (e.g. ".cgi") */
               sptr++;
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            }
            else
            {
               /* EXECing a directory here */
               ExecFileType = false;
               while (*cptr && *cptr != '/' && pptr < zptr) *pptr++ = *cptr++;
            }

            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr = '\0';

            if (mrptr->RuleType == MAPURL_RULE_UXEC)
            {
               /* "append" the username to the circumflex */
               rqptr->rqPathSet.ScriptAsPtr = sptr =
                   VmGetHeap (rqptr, strlen(WildBuffer)+2);
               *sptr++ = '~';
               for (cptr = WildBuffer; *cptr; *sptr++ = toupper(*cptr++));
               *sptr = '\0';
            }

            /* now get what it's mapped into (the more complex bit ;^) */
            cptr = WildBuffer;
            sptr = mrptr->ResultPtr;

            if (*sptr == '(')
            {
               /* run-time, e.g. exec /plbin/@ (ht_exe:pl.exe)/plbin/@ */
               sptr++;
               zptr = (pptr = RteVmsPtr) + SizeOfRteVmsPtr;
               while (*sptr && *sptr != ')' && pptr < zptr) *pptr++ = *sptr++;
               if (pptr >= zptr)
                  return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
               *pptr = '\0';
               if (*sptr) sptr++;
            }

            zptr = (pptr = Scratch) + sizeof(Scratch);

            if (mrptr->RuleType == MAPURL_RULE_UXEC)
            {
               /* begin with the mapped user instead of the /~username */
               if (PathOds)
                  uptr = MapUrl_VmsUserName (rqptr, WildBuffer, NULL);
               else
               {
                  uptr = MapUrl_VmsUserName (rqptr, WildBuffer, &PathOds);
                  rqptr->rqPathSet.OdsNaming = PathOds;
                  if (PathOdsPtr) *PathOdsPtr = PathOds;
               }
               if (!uptr[0]) return (uptr);
#ifdef ODS_EXTENDED
               PathOdsExtended = (OdsExtended && PathOds == MAPURL_PATH_ODS_5);
#endif /* ODS_EXTENDED */
               /* copy in the-now mapped username */
               while (*uptr && pptr < zptr) *pptr++ = *uptr++;
               /* skip over the equivalent in the result (i.e. "/~*") */
               while (*sptr && *sptr != '*') sptr++;
               if (*sptr) sptr++;
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "!&Z", cptr);
               /* skip first element in the wildcard buffer (i.e. username) */
               while (*cptr) cptr++;
               cptr++;
            }
            else
            if (*(unsigned short*)mrptr->TemplatePtr == '/~')
            {
               /*
                  e.g. exec /~@/cgi-bin/@ /0""::/web/user/@/cgi-bin/@
                  i.e. the user part must be prepended to the script part.
               */
               while (*sptr && *sptr != '*' && pptr < zptr)
               {
                  if (*sptr != '\"' || *(unsigned long*)sptr != '\"\"::')
                  {
                     *pptr++ = *sptr++;
                     continue;
                  }
                  /* mapping a DECnet username based script */
                  *pptr++ = *sptr++;
                  while (*cptr && pptr < zptr) *pptr++ = *cptr++;
                  /* we'll need this username again shortly */
                  cptr = WildBuffer;
               }
               while (*sptr == '*') sptr++;
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "!&Z", cptr);
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
            }

            /* now the leading script-location part of the result */
            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            while (*sptr == '*') sptr++;

            if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                WATCH_MODULE(WATCH_MOD__DETAIL))
               WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "!&Z", cptr);
            if (ExecFileType)
            {
               /* EXECing a file type here, script name (full buffer) */
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
               /* now get the script file type (e.g. ".cgi") */
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            }
            else
            {
               /* EXECing a directory here, script name */
               while (*cptr && *cptr != '/' && pptr < zptr) *pptr++ = *cptr++;
            }

            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr = '\0';

            /* generate the VMS script name */
            MapUrl_UrlToVms (Scratch, ScriptVmsPtr, SizeOfScriptVmsPtr,
                             RmsSubChar, MapEllipsis, PathOds);

            /* indicate CGIplus script via leading '+' instead of '/' */
            if (mrptr->IsCgiPlusScript && (!RteVmsPtr || !RteVmsPtr[0]))
               ScriptPtr[0] = '+';

            /* get all including and after the first slash as the new path */
            zptr = (pptr = DerivedPathBuffer) + sizeof(DerivedPathBuffer);
            while (*cptr && pptr < zptr) *pptr++ = *cptr++;
            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr++ = '\0';
            TotalLength = pptr - DerivedPathBuffer;

            if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                WATCH_MODULE(WATCH_MOD__DETAIL))
                WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "E/UXEC !&Z !&Z !&Z",
                           ScriptPtr, ScriptVmsPtr, DerivedPathBuffer);

            if (!rqptr->rqPathSet.MapEmpty && !DerivedPathBuffer[0])
               return ("\0\0");

            if (rqptr->rqPathSet.MapOnce)
            {
               /* convert the URL-style to a VMS-style specification */
               MapUrl_UrlToVms (DerivedPathBuffer, VmsPtr, SizeOfVmsPtr,
                                RmsSubChar, MapEllipsis, PathOds);
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                             "E/UXEC !&Z !&Z", DerivedPathBuffer, VmsPtr);
               return (DerivedPathBuffer);
            }

            /* restarting at first rule map the path derived from the script */
            MapPassCount++;
            *VmsPtr = '\0';
            PathPtr = DerivedPathBuffer;
            RmsSubChar = MAPURL_RMS_SUBSTITUTION_DEFAULT;
            MapEllipsis = rqptr->rqPathSet.MapEllipsis = false;
            MetaConParseReset (MetaGlobalMappingPtr, true);
            continue;

         case MAPURL_RULE_FAIL :

            /*************/
            /* FAIL rule */
            /*************/

            if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                WATCH_MODULE(WATCH_MOD__DETAIL))
            WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "FAIL !&Z", PathPtr);
            return (MsgFor(rqptr,MSG_MAPPING_DENIED_RULE)-1);

         case MAPURL_RULE_MAP :

            /************/
            /* MAP rule */
            /************/

            cptr = WildBuffer;
            /* place this in a buffer so not to overwrite original path */
            zptr = (pptr = PathPtr = PathBuffer+1) + sizeof(PathBuffer)-1;
            sptr = mrptr->ResultPtr;
            /* scan through the result string */
            while (*sptr)
            {
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
               if (!*sptr) break;
               /* a wildcard asterisk, substitute from original path */
               while (*sptr == '*') sptr++;
               if (*sptr == '\'')
               {
                  /* specific wildcard component, e.g. "*'1" */
                  sptr++;
                  idx = *sptr++ - '1';
                  if (!(cptr = WildBufferArray[idx])) cptr = "";
               }
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
            }
            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr = '\0';

            /* continue with the substituted URL mapping */
            if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                WATCH_MODULE(WATCH_MOD__DETAIL))
            WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "MAP !&Z", PathPtr);
            continue;

         case MAPURL_RULE_PASS :
         case MAPURL_RULE_REDIRECT :
         case MAPURL_RULE_USER :

            /****************************/
            /* PASS/REDIRECT/USER rules */
            /****************************/

            cptr = WildBuffer;
            sptr = mrptr->ResultPtr; 
            zptr = (pptr = PathBuffer+1) + sizeof(PathBuffer)-1;

            /* if there is no result to map just pass back the original path */
            if (!*sptr) 
            {
               /* convert the URL-style to a VMS-style specification */
               MapUrl_UrlToVms (PathPtr, VmsPtr, SizeOfVmsPtr,
                                RmsSubChar, MapEllipsis, PathOds);
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                              "PASS !&Z !&Z", PathPtr, VmsPtr);
               return (PathPtr);
            }

            /* if redirect rule leave room for the redirection indicator */
            if (mrptr->RuleType == MAPURL_RULE_REDIRECT) pptr++;

            if (mrptr->RuleType == MAPURL_RULE_USER)
            {
               /* begin with the mapped user instead of the /~username */
               if (PathOds)
                  uptr = MapUrl_VmsUserName (rqptr, WildBuffer, NULL);
               else
               {
                  uptr = MapUrl_VmsUserName (rqptr, WildBuffer, &PathOds);
                  rqptr->rqPathSet.OdsNaming = PathOds;
                  if (PathOdsPtr) *PathOdsPtr = PathOds;
               }
               if (!uptr[0]) return (uptr);
#ifdef ODS_EXTENDED
               PathOdsExtended = (OdsExtended && PathOds == MAPURL_PATH_ODS_5);
#endif /* ODS_EXTENDED */
               /* copy in the-now mapped username */
               while (*uptr && pptr < zptr) *pptr++ = *uptr++;
               /* skip over the equivalent in the result (i.e. "/~*") */
               while (*sptr && *sptr != '*') sptr++;
               if (*sptr) sptr++;
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "!&Z", cptr);
               /* skip first element in the wildcard buffer (i.e. username) */
               while (*cptr) cptr++;
               cptr++;
            }

            /* scan through the result string */
            if (mrptr->RuleType == MAPURL_RULE_PASS &&
                isdigit(sptr[0]) && isdigit(sptr[1]) &&
                isdigit(sptr[2]) && (!sptr[3] || sptr[3] == ' '))
            {
               StatusCodeMapping = true;
               while (*sptr && pptr < zptr) *pptr++ = *sptr++;
            }
            else
            { 
               StatusCodeMapping = false;
               /* map wildcard asterisks */
               while (*sptr)
               {
                  while (*sptr && *sptr != '*' && pptr < zptr)
                     *pptr++ = *sptr++;
                  if (!*sptr) break;
                  /* wildcard asterisk, substitute from the wildcard buffer */
                  while (*sptr == '*') sptr++;
                  if (*sptr == '\'')
                  {
                     /* specific wildcard component, e.g. "*'1" */
                     sptr++;
                     idx = *sptr++ - '1';
                     if (!(cptr = WildBufferArray[idx])) cptr = "";
                  }
                  while (*cptr && pptr < zptr) *pptr++ = *cptr++;
                  cptr++;
               }
            }
            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr = '\0';

            if (mrptr->RuleType == MAPURL_RULE_PASS ||
                mrptr->RuleType == MAPURL_RULE_USER)
            {
               /* if status code mapping return leading null char */
               if (StatusCodeMapping)
               {
                  PathBuffer[0] = '\0';
                  return (PathBuffer);
               }

               /* convert the URL-style to a VMS-style specification */
               MapUrl_UrlToVms (PathBuffer+1, VmsPtr, SizeOfVmsPtr,
                                RmsSubChar, MapEllipsis, PathOds);
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                             "PASS !&Z !&Z", PathPtr, VmsPtr);
               return (PathBuffer+1);
            }
            else
            {
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                             "REDIRECT !&Z !&Z", PathPtr, VmsPtr);
               /* indicate it's a redirect */
               PathBuffer[1] = '\1';
               return (PathBuffer+1);
            }

         case MAPURL_RULE_SCRIPT :

            /***************/
            /* SCRIPT rule */
            /***************/

            sptr = mrptr->TemplatePtr;
            zptr = (pptr = ScriptPtr) + SizeOfScriptPtr;
            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            if (pptr > ScriptPtr && pptr[-1] == '/') pptr--;
            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr = '\0';

            sptr = mrptr->ResultPtr;
            zptr = (pptr = Scratch) + sizeof(Scratch);
            while (*sptr &&
                   *sptr != '*' &&
                   *(unsigned short*)sptr != '/*' &&
                   pptr < zptr) *pptr++ = *sptr++;
            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            *pptr = '\0';

            /* convert the URL-format script to a VMS-format specification */
            MapUrl_UrlToVms (Scratch, ScriptVmsPtr, SizeOfScriptVmsPtr,
                             RmsSubChar, MapEllipsis, PathOds);

            /* indicate CGIplus script via leading '+' instead of '/' */
            if (mrptr->IsCgiPlusScript) ScriptPtr[0] = '+';

            /* get wildcard matched second section of path as new path */
            zptr = (pptr = DerivedPathBuffer) + sizeof(DerivedPathBuffer);
            while (*sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            cptr = WildBuffer;
            while (*cptr && pptr < zptr) *pptr++ = *cptr++;
            if (pptr < zptr) *pptr++ = '\0';
            if (pptr >= zptr)
               return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
            TotalLength = pptr - DerivedPathBuffer;

            if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                WATCH_MODULE(WATCH_MOD__DETAIL))
               WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                          "SCRIPT !&Z !&Z !&Z",
                          ScriptPtr, ScriptVmsPtr, DerivedPathBuffer);

            if (!rqptr->rqPathSet.MapEmpty && !DerivedPathBuffer[0])
               return ("\0\0");

            if (rqptr->rqPathSet.MapOnce)
            {
               /* convert the URL-style to a VMS-style specification */
               MapUrl_UrlToVms (DerivedPathBuffer, VmsPtr, SizeOfVmsPtr,
                                RmsSubChar, MapEllipsis, PathOds);
               if (WATCH_MODULE(WATCH_MOD_MAPURL) &&
                   WATCH_MODULE(WATCH_MOD__DETAIL))
                  WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                             "SCRIPT !&Z !&Z", DerivedPathBuffer, VmsPtr);
               return (DerivedPathBuffer);
            }

            /* restarting at first rule map the path derived from the script */
            MapPassCount++;
            *VmsPtr = '\0';
            PathPtr = DerivedPathBuffer;
            RmsSubChar = MAPURL_RMS_SUBSTITUTION_DEFAULT;
            MapEllipsis = rqptr->rqPathSet.MapEllipsis = false;
            MetaConParseReset (MetaGlobalMappingPtr, true);
            continue;

         case MAPURL_RULE_SET :

            /************/
            /* SET rule */
            /************/

            /* path SETTings have already been applied */
            continue;
         }
      }
      else
      {
         /***************************/
         /* mapping from VMS to URL */
         /***************************/

         /* REVERSE maps a VMS "result" to a "template" :^) */

         cptr = WildBuffer;
         zptr = (pptr = PathPtr) + SizeOfPathPtr;
         sptr = mrptr->TemplatePtr;
         /* scan through the template string */
         while (*sptr)
         {
            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            if (!*sptr) break;
            /* a wildcard asterisk, substitute from result path */
            while (*sptr == '*') sptr++;
            if (*cptr)
            {
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               cptr++;
            }
         }
         if (pptr >= zptr)
            return (MsgFor(rqptr,MSG_MAPPING_DENIED_INTERNAL)-1);
         *pptr = '\0';

         if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
            WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                          "PASS !&Z", PathPtr);
         return (PathPtr);
      }
   }

   /***********************/
   /* a mapping not found */
   /***********************/

   if (*PathPtr != '/')
   {
      /* a "normal" request must begin with an absolute path, it's a proxy! */
      return (PathPtr);
   }

   if (MapPathToVms)
      return (MsgFor(rqptr,MSG_MAPPING_DENIED_DEFAULT)-1);
   else
      return ("NO:[REVERSE.MAPPING.FOR.THIS]FILE.PATH");
}

/*****************************************************************************/
/*
Load mapping rules into meta-config structure.
*/ 
 
int MapUrl_ConfigLoad (META_CONFIG **MetaConPtrPtr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_ConfigLoad() !AZ", CONFIG_MAP_FILE_NAME);

   status = MetaConLoad (MetaConPtrPtr, CONFIG_MAP_FILE_NAME,
                         &MapUrl_ConfigLoadCallBack, true, true);
   if (*MetaConPtrPtr == MetaGlobalMappingPtr)
   {
      /* loading the server with run-time configuration */
      if (MetaGlobalMappingPtr->LoadReport.Count)
         WriteFaoStdout ("%!AZ-W-MAP, !UL report!%s at mapping load\n!AZ",
                         Utility, MetaGlobalMappingPtr->LoadReport.Count,
                         MetaGlobalMappingPtr->LoadReport.TextPtr);
      if (VMSnok (status)) exit (status);
   }
   return (status);
}

/*****************************************************************************/
/*
Called by MetaConUnload() to free resources allocated during mapping rule
configuration.
*/ 
 
MapUrl_ConfigUnload (META_CONFIG *mcptr)

{
   int  status;
   MAPPING_META  *mmptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "MapUrl_ConfigUnload()");

   if (mcptr->MappingMetaPtr)
   {
      if (mcptr->MappingMetaPtr == MappingMetaPtr)
      {
         memset (MappingMetaPtr, 0, sizeof(MAPPING_META));
         MappingMetaPtr = NULL;
      }
      else
         VmFree (mcptr->MappingMetaPtr, FI_LI);
      mcptr->MappingMetaPtr = NULL;
   }
}

/*****************************************************************************/
/*
For each non-meta-config directive line read by MetaConLoad() this function is
called to parse the line text's contents and to configure the private data
structure associated with each rule.
*/ 

BOOL MapUrl_ConfigLoadCallBack (META_CONFIG *mcptr)

{
   static char  Ods5Disabled [] = "ODS-5 processing CLI disabled",
                Ods5Enabled [] = "ODS-5 processing enabled",
                Ods5NotSupported [] = "ODS-5 not supported",
                ProblemAcceptLang [] =
                   "Set path \'accept=lang\' parameter problem",
                ProblemAllSpecifiedWildcard [] =
                   "All or none must be \'specified\' wildcards",
                ProblemConditional [] = "Conditional problem",
                ProblemConfused [] = "Can't make head nor tail of this one",
                ProblemHostNameLookup [] = "Cannot resolve host name !AZ, !&m",
                ProblemNameValue [] = "Set path \'name=value\' pair problem",
                ProblemNoSpecifiedWildcard [] =
                   "Cannot use \'specified\' wildcards with rule",
                ProblemOverflow [] = "Storage overflow",
                ProblemResultPath [] = "Problem with \'result\' path",
                ProblemResultRequired [] = "Requires a \'result\'",
                ProblemResultStatus [] = "Rule \'result\' cannot be a status",
                ProblemRmsSubstitution [] = "RMS substitution character",
                ProblemSetPath [] = "Set path problem",
                ProblemSpecifiedWildcard3 [] =
                   "Value of \'specified\' wildcard out-of-range",
                ProblemTemplateRequired [] = "Requires a \'template\'",
                ProblemThrottleValues [] =
                   "Successive dependent throttle values must be larger",
                ProblemWildcardMapping [] = "Wildcard mapping problem",
                ProblemTooManyWildcard [] = "Too many wildcards";

   BOOL  IsCgiPlusScript,
         PathSet,
         VirtualServerUnknown;
   int  status,
        ConditionalOffset,
        ConfigFileIndex,
        RuleType,
        SetPathNoAcceptLang,
        SetPathAlert,
        SetPathNoAlert,
        SetPathAuthMapped,
        SetPathNoAuthMapped,
        SetPathAuthOnce,
        SetPathNoAuthOnce,
        SetPathAuthRevalidateTimeout,
        SetPathAcceptLangTypeVariant,
        SetPathAcceptLangOffset,
        SetPathAcceptLangWildcard,
        SetPathAuthSysUafPwdExpUrlOffset,
        SetPathCache,
        SetPathNoCache,
        SetPathDirAccess,
        SetPathDirNoAccess,
        SetPathDirAccessSelective,
        SetPathDirImpliedWildcard,
        SetPathDirNoImpliedWildcard,
        SetPathDirWildcard,
        SetPathDirNoWildcard,
        SetPathCgiPrefixOffset,
        SetPathCharsetOffset,
        SetPathContentTypeOffset,
        SetPathHttpAcceptCharsetOffset,
        SetPathHttpAcceptLangOffset,
        SetPathDirCharsetOffset,
        SetPathIndexOffset,
        SetPathExpired,
        SetPathNoExpired,
        SetPathLog,
        SetPathNoLog,
        SetPathMapEllipsis,
        SetPathNoMapEllipsis,
        SetPathMapEmpty,
        SetPathMapNonEmpty,
        SetPathMapOnce,
        SetPathNoMapOnce,
        SetPathOds,
        SetPathPrivSsi,
        SetPathNoPrivSsi,
        SetPathProfile,
        SetPathNoProfile,
        SetPathProxyBindIpAddress,
        SetPathProxyBindIpAddressOffset,
        SetPathProxyChainHostPortOffset,
        SetPathProxyChainIpAddress,
        SetPathProxyChainPort,
        SetPathReportBasic,
        SetPathReportDetailed,
        SetPathReport400as,
        SetPathReport403as,
        SetPathReport404as,
        SetPathScriptAsOffset,
        SetPathScriptFind,
        SetPathScriptNoFind,
        SetPathScriptBitBucketTimeout,
        SetPathScriptCpuMax,
        SetPathScriptParamsOffset,
        SetPathThrottleBusy,
        SetPathThrottleFrom,
        SetPathThrottleResume,
        SetPathThrottleTo,
        SetPathThrottleTimeoutBusy,
        SetPathThrottleTimeoutQueue,
        SetPathTimeoutKeepAlive,
        SetPathTimeoutNoProgress,
        SetPathTimeoutOutput,
        SetPathStmLF,
        SetPathNoStmLF,
        SetPathSSLCGIvar,
        TemplateWildcardCount,
        TotalLength,
        ResultOffset,
        ResultWildcardCount,
        ResultSpecifiedWildcardCount;
   unsigned short  Length;
   char  RmsSubChar,
         SetPathAcceptLangChar;
   char  *cptr, *sptr, *zptr,
         *ConditionalPtr,
         *EndPtr,
         *ResultPtr,
         *SetPathAcceptLangPtr,
         *SetPathAuthSysUafPwdExpUrlPtr,
         *SetPathCgiPrefixPtr,
         *SetPathCharsetPtr,
         *SetPathContentTypePtr,
         *SetPathDirCharsetPtr,
         *SetPathHttpAcceptCharsetPtr,
         *SetPathHttpAcceptLangPtr,
         *SetPathIndexPtr,
         *SetPathProxyBindIpAddressPtr,
         *SetPathProxyChainHostPortPtr,
         *SetPathScriptAsPtr,
         *SetPathScriptParamsPtr,
         *SetPathPtr,
         *StringPtr,
         *TemplatePtr;
   char  Name [256],
         RuleName [64],
         Scratch [512],
         StringBuffer [1024],
         Value [1024];
   MAPPING_META  *mmptr;
   MAP_RULE_META  *mrptr;
   METACON_LINE  *mclptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
   {
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_ConfigLoadCallBack() !&F !&X",
                 &MapUrl_ConfigLoadCallBack, mcptr);
      if (WATCH_MODULE(WATCH_MOD__DETAIL))
      {
         mclptr = mcptr->ParsePtr;
         WatchDataFormatted ("!&X !UL !UL !UL !UL !&X !&Z !&Z\n",
            mclptr, mclptr->Size, mclptr->Token, mclptr->Number,
            mclptr->Length, mclptr->LineDataPtr, mclptr->TextPtr,
            mclptr->InlineTextPtr);
      }
   }

   /* get a pointer to the current "line" */
   mclptr = mcptr->ParsePtr;

   /* if this is during server startup/reload set the global service pointer */
   if (mcptr == MetaGlobalMappingPtr)
      mmptr = mcptr->MappingMetaPtr = MappingMetaPtr = &MappingMeta;
   else
   /* if a report then conjure one up through quantum mechanics */
   if (!mcptr->MappingMetaPtr)
      mmptr = mcptr->MappingMetaPtr = VmGet (sizeof(MAPPING_META));
   else
      /* not the first time through */
      mmptr = mcptr->MappingMetaPtr;

   if (mclptr->Token == METACON_TOKEN_PRE)
   {
      /******************/
      /* pre-initialize */
      /******************/

      MapUrlPathOds5 = OdsExtended = false;

      return (true);
   }

   if (mclptr->Token == METACON_TOKEN_POST)
   {
      /****************/
      /* post-process */
      /****************/

      /* if global service pointer, during server startup or reload */
      if (mcptr == MetaGlobalMappingPtr)
      {
         /* if no ODS5 paths then EFS can be disabled */
#ifdef ODS_EXTENDED
         if (CliOdsExtendedDisabled)
         {
            OdsExtended = false;
            MetaConReport (mcptr, Ods5Disabled);
         }
         else
#endif /* ODS_EXTENDED */
         if (CliOdsExtendedEnabled || MapUrlPathOds5)
         {
#ifdef ODS_EXTENDED
            if (SysInfo.VersionInteger >= 72)
            {
               OdsExtended = true;
               MetaConReport (mcptr, Ods5Enabled);
            }
            else
#endif /* ODS_EXTENDED */
            {
               OdsExtended = false;
               MetaConReport (mcptr, Ods5NotSupported);
            }
         }
         else
            OdsExtended = false;

         /* reset the username cache */
         MapUrl_VmsUserNameCache (NULL, NULL, NULL, NULL);

         /* initialize the throttled paths */
         ThrottleInit ();
      }

      return (true);
   }

   /***********/
   /* process */
   /***********/

   /* if it's not text/inline then mapping is not interested in it */
   if (mclptr->Token != METACON_TOKEN_TEXT && !mclptr->InlineTextPtr)
      return (true);

   /* buffer the text associated with the current "line" */
   zptr = (sptr = StringBuffer) + sizeof(StringBuffer);
   if (mclptr->InlineTextPtr)
      cptr = mclptr->InlineTextPtr;
   else
      cptr = mclptr->TextPtr;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      MetaConReport (mcptr, ProblemOverflow);
      return (false);
   }
   *sptr = '\0';
   cptr = StringBuffer;

   IsCgiPlusScript = 
      SetPathNoAcceptLang =
      SetPathAcceptLangTypeVariant = SetPathAcceptLangWildcard = 
      SetPathNoAlert =
      SetPathAuthMapped = SetPathNoAuthMapped =
      SetPathAuthOnce = SetPathNoAuthOnce =
      SetPathCache = SetPathNoCache =
      SetPathDirAccess = SetPathDirNoAccess = SetPathDirAccessSelective = 
      SetPathDirImpliedWildcard = SetPathDirNoImpliedWildcard =
      SetPathDirWildcard = SetPathDirNoWildcard =
      SetPathExpired = SetPathNoExpired =
      SetPathLog = SetPathNoLog =
      SetPathMapEllipsis = SetPathNoMapEllipsis =
      SetPathMapEmpty = SetPathMapNonEmpty =
      SetPathMapOnce = SetPathNoMapOnce =
      SetPathPrivSsi = SetPathNoPrivSsi =
      SetPathProfile = SetPathNoProfile =
      SetPathReportBasic = SetPathReportDetailed =
      SetPathReport400as = SetPathReport403as = SetPathReport404as =
      SetPathScriptFind = SetPathScriptNoFind =
      SetPathStmLF = SetPathNoStmLF =
      PathSet = false;
   ResultOffset = SetPathAuthRevalidateTimeout =
      SetPathAlert = 
      SetPathAuthSysUafPwdExpUrlOffset =
      SetPathAcceptLangOffset = SetPathCgiPrefixOffset =
      SetPathCharsetOffset =  SetPathContentTypeOffset =
      SetPathOds = SetPathHttpAcceptCharsetOffset = 
      SetPathHttpAcceptLangOffset = SetPathDirCharsetOffset =
      SetPathIndexOffset =  SetPathProxyBindIpAddressOffset =
      SetPathProxyBindIpAddress = SetPathProxyChainIpAddress =
      SetPathProxyChainPort = SetPathProxyChainHostPortOffset =
      SetPathScriptAsOffset = SetPathScriptBitBucketTimeout =
      SetPathScriptCpuMax = SetPathScriptParamsOffset =
      SetPathSSLCGIvar = SetPathThrottleBusy =
      SetPathThrottleFrom = SetPathThrottleResume =
      SetPathThrottleTo = SetPathThrottleTimeoutBusy =
      SetPathThrottleTimeoutQueue = SetPathTimeoutKeepAlive =
      SetPathTimeoutNoProgress = SetPathTimeoutOutput = 0;
   SetPathAcceptLangPtr = SetPathAuthSysUafPwdExpUrlPtr =
      SetPathCgiPrefixPtr = SetPathCharsetPtr =
      SetPathContentTypePtr = SetPathHttpAcceptCharsetPtr =
      SetPathDirCharsetPtr = SetPathHttpAcceptLangPtr =
      SetPathIndexPtr = SetPathProxyBindIpAddressPtr =
      SetPathProxyChainHostPortPtr = SetPathScriptAsPtr =
      SetPathScriptParamsPtr = NULL;
   SetPathAcceptLangChar = RmsSubChar = '\0';

   /********/
   /* rule */
   /********/

   StringPtr = cptr;
   if (VMSnok (StringParseValue (&StringPtr, RuleName, sizeof(RuleName))))
   {
      MetaConReport (mcptr, ProblemConfused);
      return (true);
   }
   cptr = StringPtr;

   IsCgiPlusScript = false;
   if (strsame (RuleName, "EXEC", -1))
      RuleType = MAPURL_RULE_EXEC;
   else
   if (strsame (RuleName, "EXEC+", -1))
   {
      RuleType = MAPURL_RULE_EXEC;
      IsCgiPlusScript = true;
   }
   else
   if (strsame (RuleName, "FAIL", -1))
      RuleType = MAPURL_RULE_FAIL;
   else
   if (strsame (RuleName, "MAP", -1))
      RuleType = MAPURL_RULE_MAP;
   else
   if (strsame (RuleName, "PASS", -1))
      RuleType = MAPURL_RULE_PASS;
   else
   if (strsame (RuleName, "REDIRECT", -1))
      RuleType = MAPURL_RULE_REDIRECT;
   else
   if (strsame (RuleName, "SCRIPT", -1))
      RuleType = MAPURL_RULE_SCRIPT;
   else
   if (strsame (RuleName, "SCRIPT+", -1))
   {
      RuleType = MAPURL_RULE_SCRIPT;
      IsCgiPlusScript = true;
   }
   else
   if (strsame (RuleName, "SET", -1))
      RuleType = MAPURL_RULE_SET;
   else
   if (strsame (RuleName, "UXEC", -1) ||
       strsame (RuleName, "UEXEC", -1))
      RuleType = MAPURL_RULE_UXEC;
   else
   if (strsame (RuleName, "UXEC+", -1) ||
       strsame (RuleName, "UEXEC+", -1))
   {
      RuleType = MAPURL_RULE_UXEC;
      IsCgiPlusScript = true;
   }
   else
   if (strsame (RuleName, "USER", -1))
      RuleType = MAPURL_RULE_USER;
   else
   if (strsame (RuleName, "VIRTUAL", -1))
      RuleType = MAPURL_RULE_VIRTUAL;
   else
   {
      MetaConReport (mcptr, ProblemConfused);
      return (true);
   }

   /* ensure a conditional is not mistaken for a missing template */
   if (!*cptr || *cptr == '[' || *(unsigned short*)cptr == '![') 
   {
      /* there must be a "template" to map from */
      MetaConReport (mcptr, ProblemTemplateRequired);
      return (true);
   }

   /* extract template (the path the rule is applied to) */
   StringPtr = cptr;
   if (VMSnok (StringSpanValue (&StringPtr, &TemplatePtr, &EndPtr)))
   {
      MetaConReport (mcptr, ProblemTemplateRequired);
      return (true);
   }
   cptr = StringPtr;
   *EndPtr = '\0';
   /* note the position of the empty string in case neither of these! */
   ResultPtr = SetPathPtr = ConditionalPtr = EndPtr;
   ResultOffset = ConditionalOffset = EndPtr - TemplatePtr;

   if (RuleType != MAPURL_RULE_SET)
   {
      /*********************/
      /* path mapping rule */
      /*********************/

      /* if this doesn't look like a conditional */
      if (*cptr && *cptr != '[' && *(unsigned short*)cptr != '![')
      {
         StringPtr = cptr;
         if (VMSnok (StringSpanValue (&StringPtr, &ResultPtr, &EndPtr)))
         {
            MetaConReport (mcptr, ProblemConfused);
            return (true);
         }
         cptr = StringPtr;
         *EndPtr = '\0';
         ResultOffset = ResultPtr - TemplatePtr;
         if (Debug) fprintf (stdout, "ResultPtr |%s|\n", ResultPtr);
      }
   }

   /* if it doesn't look like a conditional then assume it's SET directives */
   if (*cptr && *cptr != '[' && *(unsigned short*)cptr != '![')
   {
      /*****************/
      /* path SETtings */
      /*****************/

      PathSet = true;
      while (*cptr)
      {
         /* break if this looks like a conditional */
         if (*cptr == '[' || *(unsigned short*)cptr == '![') break;

         StringPtr = cptr;
         if (VMSnok (StringSpanValue (&StringPtr, &SetPathPtr, &EndPtr)))
         {
            MetaConReport (mcptr, ProblemSetPath);
            return (true);
         }
         cptr = StringPtr;
         *EndPtr = '\0';
         if (Debug) fprintf (stdout, "SetPathPtr |%s|\n", SetPathPtr);
         if (!*SetPathPtr)
         {
            MetaConReport (mcptr, ProblemSetPath);
            return (true);
         }

         if (strsame (SetPathPtr, "ACCEPT=LANG", -1))
            SetPathAcceptLangChar = '_';
         else
         if (strsame (SetPathPtr, "NOACCEPT=LANG", -1) ||
             strsame (SetPathPtr, "ACCEPT=NOLANG", -1))
            SetPathNoAcceptLang = true;
         else
         if (strsame (SetPathPtr, "ACCEPT=LANG=", 12))
         {
            SetPathAcceptLangChar = '_';
            sptr = SetPathPtr + 12;
            if (*sptr == '(') sptr++;
            while (*sptr)
            {
               if (strsame (sptr, "CHAR=", 5))
               {
                  SetPathAcceptLangChar = sptr[5];
                  sptr += 6;
               }
               else
               if (strsame (sptr, "DEFAULT=", 8))
               {
                  sptr += 8;
                  SetPathAcceptLangPtr = sptr;
                  SetPathAcceptLangOffset = SetPathAcceptLangPtr - TemplatePtr;
                  while (*sptr && *sptr != ',' && *sptr != ')') sptr++;
                  if (*sptr) *sptr++ = '\0';
               }
               else
               if (strsame (sptr, "VARIANT=", 8))
               {
                  sptr += 8;
                  if (strsame (sptr, "NAME", 4))
                     SetPathAcceptLangTypeVariant = false;
                  else
                  if (strsame (sptr, "TYPE", 4))
                     SetPathAcceptLangTypeVariant = true;
                  else
                  {
                     MetaConReport (mcptr, ProblemAcceptLang);
                     return (true);
                  }
                  while (*sptr && *sptr != ',' && *sptr != ')') sptr++;
               }
               else
               {
                  MetaConReport (mcptr, ProblemAcceptLang);
                  return (true);
               }
               if (*sptr == ',' || *sptr == ')') sptr++;
            }
            if (SetPathAcceptLangChar != '_' &&
                SetPathAcceptLangChar != '-' &&
                SetPathAcceptLangChar != '.' &&
                SetPathAcceptLangChar != '$')
            {
               MetaConReport (mcptr, ProblemAcceptLang);
               return (true);
            }
            /* check if there is a wildcard file type */
            for (sptr = TemplatePtr; *sptr; sptr++);
            while (sptr > TemplatePtr && *sptr != '.' && *sptr != '/') sptr--;
            if (*sptr == '.') SetPathAcceptLangWildcard = true;
         }
         else
         if (strsame (SetPathPtr, "ALERT", -1))
            SetPathAlert = MAPURL_PATH_ALERT_END;
         else
         if (strsame (SetPathPtr, "ALERT=END", -1))
            SetPathAlert = MAPURL_PATH_ALERT_END;
         else
         if (strsame (SetPathPtr, "ALERT=AUTH", -1))
            SetPathAlert = MAPURL_PATH_ALERT_AUTH;
         else
         if (strsame (SetPathPtr, "ALERT=MAP", -1))
            SetPathAlert = MAPURL_PATH_ALERT_MAP;
         else
         if (strsame (SetPathPtr, "NOALERT", -1))
            SetPathNoAlert = true;
         else
         if (strsame (SetPathPtr, "AUTH=MAPPED", -1))
            SetPathAuthMapped = true;
         else
         if (strsame (SetPathPtr, "NOAUTH=MAPPED", -1) ||
             strsame (SetPathPtr, "AUTH=REQUEST", -1))
            SetPathNoAuthMapped = true;
         else
         if (strsame (SetPathPtr, "AUTH=ONCE", -1) ||
             strsame (SetPathPtr, "AUTHONCE", -1))
            SetPathAuthOnce = true;
         else
         if (strsame (SetPathPtr, "AUTH=NOONCE", -1) ||
             strsame (SetPathPtr, "NOAUTH=ONCE", -1) ||
             strsame (SetPathPtr, "NOAUTHONCE", -1))
            SetPathNoAuthOnce = true;
         else
         if (strsame (SetPathPtr, "AUTH=REVALIDATE=", 16))
            SetPathAuthRevalidateTimeout =
               MetaConSetSeconds (mcptr, SetPathPtr+16, 1);
         else
         if (strsame (SetPathPtr, "AUTH=SYSUAF=PWDEXPURL=", 22))
         {
            SetPathAuthSysUafPwdExpUrlPtr = SetPathPtr + 22;
            SetPathAuthSysUafPwdExpUrlOffset =
               SetPathAuthSysUafPwdExpUrlPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "CACHE", -1))
            SetPathCache = true;
         else
         if (strsame (SetPathPtr, "NOCACHE", -1))
            SetPathNoCache = true;
         else
         if (strsame (SetPathPtr, "CGIPREFIX=", 10))
         {
            SetPathCgiPrefixPtr = SetPathPtr + 10;
            SetPathCgiPrefixOffset = SetPathCgiPrefixPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "CHARSET=", 8))
         {
            SetPathCharsetPtr = SetPathPtr + 8;
            SetPathCharsetOffset = SetPathCharsetPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "CONTENT=", 8))
         {
            SetPathContentTypePtr = SetPathPtr + 8;
            SetPathContentTypeOffset = SetPathContentTypePtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "CONTENT-TYPE=", 13))
         {
            /* just a synonym for "content=", bet someone uses it! */
            SetPathContentTypePtr = SetPathPtr + 13;
            SetPathContentTypeOffset = SetPathContentTypePtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "DIR=ACCESS=SELECTIVE", -1))
            SetPathDirAccessSelective = true;
         else
         if (strsame (SetPathPtr, "DIR=ACCESS", -1))
            SetPathDirAccess = true;
         else
         if (strsame (SetPathPtr, "DIR=NOACCESS", -1))
            SetPathDirNoAccess = true;
         else
         if (strsame (SetPathPtr, "DIR=CHARSET=", 12))
         {
            SetPathDirCharsetPtr = SetPathPtr + 12;
            SetPathDirCharsetOffset = SetPathDirCharsetPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "DIR=LAYOUT=", 11))
         {
            SetPathIndexPtr = SetPathPtr + 11;
            SetPathIndexOffset = SetPathIndexPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "DIR=IMPLIEDWILDCARD", -1))
            SetPathDirImpliedWildcard = true;
         else
         if (strsame (SetPathPtr, "DIR=NOIMPLIEDWILDCARD", -1))
            SetPathDirNoImpliedWildcard = true;
         else
         if (strsame (SetPathPtr, "DIR=WILDCARD", -1))
            SetPathDirWildcard = true;
         else
         if (strsame (SetPathPtr, "DIR=NOWILDCARD", -1))
            SetPathDirNoWildcard = true;
         else
         if (strsame (SetPathPtr, "HTTP=ACCEPT-CHARSET=", 20))
         {
            SetPathHttpAcceptCharsetPtr = SetPathPtr + 20;
            SetPathHttpAcceptCharsetOffset =
               SetPathHttpAcceptCharsetPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "HTTP=ACCEPT-LANGUAGE=", 21))
         {
            SetPathHttpAcceptLangPtr = SetPathPtr + 21;
            SetPathHttpAcceptLangOffset =
               SetPathHttpAcceptLangPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "INDEX=", 6))
         {
            SetPathIndexPtr = SetPathPtr + 6;
            SetPathIndexOffset = SetPathIndexPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "EXPIRED", -1))
            SetPathExpired = true;
         else
         if (strsame (SetPathPtr, "NOEXPIRED", -1))
            SetPathNoExpired = true;
         else
         if (strsame (SetPathPtr, "LOG", -1))
            SetPathLog = true;
         else
         if (strsame (SetPathPtr, "NOLOG", -1))
            SetPathNoLog = true;
         else
         if (strsame (SetPathPtr, "MAP=ELLIPSIS", -1))
            SetPathMapEllipsis = true;
         else
         if (strsame (SetPathPtr, "NOMAP=ELLIPSIS", -1) ||
             strsame (SetPathPtr, "MAP=NOELLIPSIS", -1))
            SetPathNoMapEllipsis = false;
         else
         if (strsame (SetPathPtr, "MAPEMPTY", -1) ||
             strsame (SetPathPtr, "MAP=EMPTY", -1))
            SetPathMapEmpty = true;
         else
         if (strsame (SetPathPtr, "NOMAPEMPTY", -1) ||
             strsame (SetPathPtr, "NOMAP=EMPTY", -1) ||
             strsame (SetPathPtr, "MAP=NONEMPTY", -1))
            SetPathMapNonEmpty = true;
         else
         if (strsame (SetPathPtr, "MAP=ONCE", -1) ||
             strsame (SetPathPtr, "MAPONCE", -1))
            SetPathMapOnce = true;
         else
         if (strsame (SetPathPtr, "MAP=NOONCE", -1) ||
             strsame (SetPathPtr, "NOMAP=ONCE", -1) ||
             strsame (SetPathPtr, "NOMAPONCE", -1))
            SetPathNoMapOnce = true;
         else
         if (strsame (SetPathPtr, "ODS=0", -1) ||
             strsame (SetPathPtr, "NOODS", -1))
            SetPathOds = MAPURL_PATH_ODS_0;
         else
         if (strsame (SetPathPtr, "ODS=2", -1) ||
             strsame (SetPathPtr, "ODS-2", -1))
            SetPathOds = MAPURL_PATH_ODS_2;
         else
         if (strsame (SetPathPtr, "ODS=5", -1) ||
             strsame (SetPathPtr, "ODS-5", -1))
         {
            SetPathOds = MAPURL_PATH_ODS_5;
#ifdef ODS_EXTENDED
            MapUrlPathOds5 = true;
#endif /* ODS_EXTENDED */
         }
         else
         if (strsame (SetPathPtr, "ODS=ADS", -1))
            SetPathOds = MAPURL_PATH_ODS_ADS;
         else
         if (strsame (SetPathPtr, "ODS=PWK", -1))
            SetPathOds = MAPURL_PATH_ODS_PWK;
         else
         if (strsame (SetPathPtr, "ODS=SMB", -1))
            SetPathOds = MAPURL_PATH_ODS_SMB;
         else
         if (strsame (SetPathPtr, "ODS=SRI", -1))
            SetPathOds = MAPURL_PATH_ODS_SRI;
         else
         if (strsame (SetPathPtr, "PROFILE", -1))
            SetPathProfile = true;
         else
         if (strsame (SetPathPtr, "NOPROFILE", -1))
            SetPathNoProfile = true;
         else
         if (strsame (SetPathPtr, "PROXY=BIND=", 11))
         {
            SetPathProxyBindIpAddressPtr = SetPathPtr + 11;
            SetPathProxyBindIpAddressOffset =
               SetPathProxyBindIpAddressPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "PROXY=CHAIN=", 12))
         {
            SetPathProxyChainHostPortPtr = SetPathPtr + 12;
            SetPathProxyChainHostPortOffset =
               SetPathProxyChainHostPortPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "REPORT=BAS", 10))
            SetPathReportBasic = true;
         else
         if (strsame (SetPathPtr, "REPORT=DET", 10))
            SetPathReportDetailed = true;
         else
         if (strsame (SetPathPtr, "REPORT=400=", 11))
            SetPathReport400as = atoi(SetPathPtr+11);
         else
         if (strsame (SetPathPtr, "REPORT=403=", 11))
            SetPathReport403as = atoi(SetPathPtr+11);
         else
         if (strsame (SetPathPtr, "REPORT=404=", 11))
            SetPathReport404as = atoi(SetPathPtr+11);
         else
         if (strsame (SetPathPtr, "RMSCHAR=", 8))
            RmsSubChar = SetPathPtr[8];
         else
         if (strsame (SetPathPtr, "SCRIPT=AS=", 10))
         {
            /* while setting the required pointers force to upper-case */
            for (sptr = SetPathScriptAsPtr = SetPathPtr+10; *sptr; sptr++)
                *sptr = toupper(*sptr);
            SetPathScriptAsOffset = SetPathScriptAsPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "SCRIPT=BIT-BUCKET=", 18))
            SetPathScriptBitBucketTimeout =
               MetaConSetSeconds (mcptr, SetPathPtr+18, 1);
         else
         if (strsame (SetPathPtr, "SCRIPT=CPU=", 11))
            SetPathScriptCpuMax =
               MetaConSetSeconds (mcptr, SetPathPtr+11, 1);
         else
         if (strsame (SetPathPtr, "SCRIPT=FIND", 8))
            SetPathScriptFind = true;
         else
         if (strsame (SetPathPtr, "SCRIPT=NOFIND", -1) ||
             strsame (SetPathPtr, "NOSCRIPT=FIND", -1))
            SetPathScriptNoFind = true;
         else
         if (strsame (SetPathPtr, "SCRIPT=PARAM=", 13))
         {
            SetPathScriptParamsPtr = SetPathPtr + 13;
            SetPathScriptParamsOffset = SetPathScriptParamsPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "SCRIPT=PARAMS=", 14))
         {
            SetPathScriptParamsPtr = SetPathPtr + 14;
            SetPathScriptParamsOffset = SetPathScriptParamsPtr - TemplatePtr;
         }
         else
         if (strsame (SetPathPtr, "SSI=PRIV", 8))
            SetPathPrivSsi = true;
         else
         if (strsame (SetPathPtr, "SSI=NOPRIV", -1) ||
             strsame (SetPathPtr, "NOSSI=PRIV", -1))
            SetPathNoPrivSsi = true;
         else
         if (strsame (SetPathPtr, "NOSSLCGI", -1))
            SetPathSSLCGIvar = SESOLA_CGI_VAR_NONE;
         else
         if (strsame (SetPathPtr, "SSLCGI=NONE", -1))
            SetPathSSLCGIvar = SESOLA_CGI_VAR_NONE;
         else
         if (strsame (SetPathPtr, "SSLCGI=APACHE_MOD_SSL", -1))
            SetPathSSLCGIvar = SESOLA_CGI_VAR_APACHE_MOD_SSL;
         else
         if (strsame (SetPathPtr, "SSLCGI=PURVEYOR", -1))
            SetPathSSLCGIvar = SESOLA_CGI_VAR_PURVEYOR;
         else
         if (strsame (SetPathPtr, "STMLF", -1))
            SetPathStmLF = true;
         else
         if (strsame (SetPathPtr, "NOSTMLF", -1))
            SetPathNoStmLF = true;
         else
         if (strsame (SetPathPtr, "THROTTLE=BUSY=", 14))
            SetPathThrottleBusy = atoi(SetPathPtr+14);
         else
         if (strsame (SetPathPtr, "THROTTLE=FROM=", 14))
            SetPathThrottleFrom = atoi(SetPathPtr+14);
         else
         if (strsame (SetPathPtr, "THROTTLE=RESUME=", 16))
            SetPathThrottleResume = atoi(SetPathPtr+16);
         else
         if (strsame (SetPathPtr, "THROTTLE=TO=", 12))
            SetPathThrottleTo = atoi(SetPathPtr+12);
         else
         if (strsame (SetPathPtr, "THROTTLE=TIMEOUT=BUSY=", 22))
            SetPathThrottleTimeoutBusy =
               MetaConSetSeconds (mcptr, SetPathPtr+22, 1);
         else
         if (strsame (SetPathPtr, "THROTTLE=TIMEOUT=QUEUE=", 23))
            SetPathThrottleTimeoutQueue =
               MetaConSetSeconds (mcptr, SetPathPtr+23, 1);
         else
         /* get the order correct chucklehead!! */
         if (strsame (SetPathPtr, "THROTTLE=", 9))
         {
            /* throttle=from,to,resume,busy,t/o-queue,t/o-busy */
            sptr = SetPathPtr+9;
            if (*sptr == '(') sptr++;
            if (isdigit(*sptr)) SetPathThrottleFrom = atoi(sptr);
            while (*sptr && isdigit(*sptr)) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr)) SetPathThrottleTo = atoi(sptr);
            while (*sptr && isdigit(*sptr)) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr)) SetPathThrottleResume = atoi(sptr);
            while (*sptr && isdigit(*sptr)) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr)) SetPathThrottleBusy = atoi(sptr);
            while (*sptr && isdigit(*sptr)) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr))
               SetPathThrottleTimeoutQueue =
                  MetaConSetSeconds (mcptr, sptr, 1);
            while (*sptr && (isdigit(*sptr) || *sptr == ':')) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr))
               SetPathThrottleTimeoutBusy =
                  MetaConSetSeconds (mcptr, sptr, 1);
            while (*sptr && (isdigit(*sptr) || *sptr == ':')) *sptr++;
         }
         else
         if (strsame (SetPathPtr, "TIMEOUT=KEEPALIVE=", 18))
            SetPathTimeoutKeepAlive =
               MetaConSetSeconds (mcptr, SetPathPtr+18, 1);
         else
         if (strsame (SetPathPtr, "TIMEOUT=NOPROGRESS=", 19))
            SetPathTimeoutNoProgress =
               MetaConSetSeconds (mcptr, SetPathPtr+19, 1);
         else
         if (strsame (SetPathPtr, "TIMEOUT=OUTPUT=", 15))
            SetPathTimeoutOutput =
               MetaConSetSeconds (mcptr, SetPathPtr+15, 1);
         else
         if (strsame (SetPathPtr, "TIMEOUT=", 8))
         {
            sptr = SetPathPtr+8;
            if (isdigit(*sptr))
               SetPathTimeoutKeepAlive = MetaConSetSeconds (mcptr, sptr, 1);
            while (*sptr && (isdigit(*sptr) || *sptr == ':')) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr))
               SetPathTimeoutNoProgress = MetaConSetSeconds (mcptr, sptr, 1);
            while (*sptr && (isdigit(*sptr) || *sptr == ':')) *sptr++;
            if (*sptr == ',') *sptr++;
            if (isdigit(*sptr))
               SetPathTimeoutOutput = MetaConSetSeconds (mcptr, sptr, 1);
            while (*sptr && (isdigit(*sptr) || *sptr == ':')) *sptr++;
         }
         else
         {
            MetaConReport (mcptr, ProblemSetPath);
            return (true);
         }
      }
   }

   /* check for optional mapping conditional(s) */
   if (*cptr == '[' || *(unsigned short*)cptr == '![')
   {
      /***************/
      /* conditional */
      /***************/

      ConditionalPtr = cptr;
      ConditionalOffset = cptr - TemplatePtr;
      while (*cptr == '[' || *(unsigned short*)cptr == '![')
      {
         while (*cptr == '[' || *cptr == '!') cptr++;
         while (*cptr && *cptr != ']')
         {
            while (ISLWS(*cptr)) cptr++;
            sptr = cptr;
            while (*cptr && *cptr != ':' && !ISLWS(*cptr) && *cptr != ']')
            {
               if (*cptr == '\\') cptr++;
               *cptr = toupper(*cptr);
               if (*cptr) cptr++;
            }
            /* basic check of conditional rule */
            if (*sptr == '!') sptr++;
            if ((*(unsigned short*)sptr != 'AC' &&
                 *(unsigned short*)sptr != 'AL' &&
                 *(unsigned short*)sptr != 'AS' &&
                 *(unsigned short*)sptr != 'CK' &&
                 *(unsigned short*)sptr != 'EX' &&
                 *(unsigned short*)sptr != 'FO' &&
                 *(unsigned short*)sptr != 'HH' &&
                 *(unsigned short*)sptr != 'HM' &&
                 *(unsigned short*)sptr != 'HO' &&
                 *(unsigned short*)sptr != 'ME' &&
                 *(unsigned short*)sptr != 'PI' &&
                 *(unsigned short*)sptr != 'QS' &&
                 *(unsigned short*)sptr != 'RF' &&
                 *(unsigned short*)sptr != 'RU' &&
                 *(unsigned short*)sptr != 'SC' &&
                 *(unsigned short*)sptr != 'SN' &&
                 *(unsigned short*)sptr != 'SP' &&
                 *(unsigned short*)sptr != 'UA' &&
                 *(unsigned short*)sptr != 'UD' &&
                 *(unsigned short*)sptr != 'VS') ||
                sptr[2] != ':')
            {
               /***********************/
               /* conditional problem */
               /***********************/

               MetaConReport (mcptr, ProblemConditional);
               return (true);
            }

            while (*cptr && !ISLWS(*cptr) && *cptr != ']')
            {
               if (*cptr == '\\') cptr++;
               if (*cptr) cptr++;
            }
            while (ISLWS(*cptr)) cptr++;
         }
         if (*cptr == ']') cptr++;
         while (ISLWS(*cptr)) cptr++;
      }
      /* terminate in the white-space following the conditional(s) */
      if (*cptr) *cptr = '\0';
   }

   TotalLength = cptr - TemplatePtr + 1;

   /***************************/
   /* rule consistency checks */
   /***************************/

   if (RmsSubChar &&
       !isalnum(RmsSubChar) && RmsSubChar != '$' &&
       RmsSubChar != '-' && RmsSubChar != '_')
   {
      MetaConReport (mcptr, ProblemRmsSubstitution);
      return (true);
   }

   TemplateWildcardCount = 0;
   cptr = TemplatePtr;
   while (*cptr)
   {
      if (*cptr++ != '*') continue;
      TemplateWildcardCount++;
      /* contiguous count as one! */
      while (*cptr == '*') cptr++;
   }

   if (TemplateWildcardCount > MAPURL_MAX_WILDCARDS)
   {
      MetaConReport (mcptr, ProblemTooManyWildcard);
      return (true);
   }

   ResultWildcardCount = ResultSpecifiedWildcardCount = 0;
   cptr = ResultPtr;
   while (*cptr)
   {
      if (*cptr++ != '*') continue;
      ResultWildcardCount++;
      /* contiguous count as one! */
      while (*cptr == '*') cptr++;
      if (*cptr != '\'') continue;
      ResultSpecifiedWildcardCount++;
      cptr++;
      if (*cptr < '1' || *cptr > '9')
      {
         MetaConReport (mcptr, ProblemSpecifiedWildcard3);
         return (true);
      }
   }

   switch (RuleType)
   {
      case MAPURL_RULE_SCRIPT :

         if (!ResultPtr[0])
         {
            MetaConReport (mcptr, ProblemResultRequired);
            return (true);
         }
         if (isdigit(ResultPtr[0]))
         {
            MetaConReport (mcptr, ProblemResultStatus);
            return (true);
         }
         if (ResultSpecifiedWildcardCount)
         {
            MetaConReport (mcptr, ProblemNoSpecifiedWildcard);
            return (true);
         }
         if (ResultWildcardCount != TemplateWildcardCount)
         {
            MetaConReport (mcptr, ProblemWildcardMapping);
            return (true);
         }
         break;

      case MAPURL_RULE_EXEC :
      case MAPURL_RULE_UXEC :

         if (!ResultPtr[0])
         {
            MetaConReport (mcptr, ProblemResultRequired);
            return (true);
         }
         if (isdigit(ResultPtr[0]))
         {
            MetaConReport (mcptr, ProblemResultStatus);
            return (true);
         }
         if (ResultSpecifiedWildcardCount)
         {
            MetaConReport (mcptr, ProblemNoSpecifiedWildcard);
            return (true);
         }
         if (!ResultWildcardCount ||
             !TemplateWildcardCount ||
             ResultWildcardCount != TemplateWildcardCount)
         {
            MetaConReport (mcptr, ProblemWildcardMapping);
            return (true);
         }
         break;

      case MAPURL_RULE_FAIL :
         break;

      case MAPURL_RULE_PASS :

         if (ResultSpecifiedWildcardCount &&
             ResultSpecifiedWildcardCount != ResultWildcardCount)
         {
            MetaConReport (mcptr, ProblemAllSpecifiedWildcard);
            return (true);
         }
         if (!isdigit(ResultPtr[0]) &&
             ResultWildcardCount > TemplateWildcardCount)
         {
            MetaConReport (mcptr, ProblemWildcardMapping);
            return (true);
         }

         /* convert the path into the VMS mapping */
         Scratch[0] = '\0';
         if (ResultPtr[0])
            MapUrl_UrlToVms (ResultPtr, Scratch, sizeof(Scratch),
                             MAPURL_RMS_SUBSTITUTION_DEFAULT, false, false);
         else
            MapUrl_UrlToVms (TemplatePtr, Scratch, sizeof(Scratch),
                             MAPURL_RMS_SUBSTITUTION_DEFAULT, false, false);

         break;

      case MAPURL_RULE_MAP :
      case MAPURL_RULE_REDIRECT :
      case MAPURL_RULE_USER :

         if (!ResultPtr[0])
         {
            MetaConReport (mcptr, ProblemResultRequired);
            return (true);
         }
         if (isdigit(ResultPtr[0]))
         {
            MetaConReport (mcptr, ProblemResultStatus);
            return (true);
         }
         if (ResultSpecifiedWildcardCount &&
             ResultSpecifiedWildcardCount != ResultWildcardCount)
         {
            MetaConReport (mcptr, ProblemAllSpecifiedWildcard);
            return (true);
         }
         if (ResultWildcardCount > TemplateWildcardCount)
         {
            MetaConReport (mcptr, ProblemWildcardMapping);
            return (true);
         }
         break;

      case MAPURL_RULE_SET :
         break;

      default :
         MetaConReport (mcptr, ProblemConfused);
         return (true);
   }

   /*****************************/
   /* throttle integrity checks */
   /*****************************/

   if (SetPathThrottleFrom)
   {
      /* throttle=from,to,resume,busy,t/o-resume,t/o-busy */
      if ((SetPathThrottleTo &&
           SetPathThrottleTo <= SetPathThrottleFrom) ||
          (SetPathThrottleResume && !SetPathThrottleTo) ||
          (SetPathThrottleResume &&
           SetPathThrottleResume <= SetPathThrottleTo) ||
          (SetPathThrottleBusy && SetPathThrottleTo &&
           SetPathThrottleBusy <= SetPathThrottleTo) ||
          (SetPathThrottleBusy && SetPathThrottleResume &&
           SetPathThrottleBusy <= SetPathThrottleResume))
      {
         MetaConReport (mcptr, ProblemThrottleValues);
         return (true);
      }
   }

   /************************************/
   /* script=param=(name=value) checks */
   /************************************/

   if (SetPathScriptParamsPtr)
   {
      /* these will generally end up as DCL symbols for a script */
      StringPtr = SetPathScriptParamsPtr;
      for (;;)
      {
         status = StringParseNameValue (&StringPtr, true,
                                        Name, sizeof(Name),
                                        Value, sizeof(Value));
         if (VMSnok (status)) break;
      }
      if (status != SS$_ENDOFFILE)
      {
         MetaConReport (mcptr, ProblemNameValue);
         return (true);
      }
   }

   /*********************************/
   /* add the rule as the line data */
   /*********************************/

   /* allocate memory for the structure plus the null terminated strings */
   mrptr = (MAP_RULE_META*)VmGet (sizeof(MAP_RULE_META)+TotalLength);

   /* set the meta-config data pointer to this structure */
   mclptr->LineDataPtr = mrptr;

   mrptr->RuleType = RuleType;
   mrptr->MetaConNumber = mclptr->Number;
   mrptr->PathSet = PathSet;
   mrptr->IsCgiPlusScript = IsCgiPlusScript;

   mrptr->mpPathSet.AcceptLangChar = SetPathAcceptLangChar;
   mrptr->mpPathSet.AcceptLangTypeVariant = SetPathAcceptLangTypeVariant;
   mrptr->mpPathSet.AcceptLangWildcard = SetPathAcceptLangWildcard;
   mrptr->mpPathSet.NoAcceptLang = SetPathNoAcceptLang;
   mrptr->mpPathSet.Alert = SetPathAlert;
   mrptr->mpPathSet.NoAlert = SetPathNoAlert;
   mrptr->mpPathSet.AuthMapped = SetPathAuthMapped;
   mrptr->mpPathSet.NoAuthMapped = SetPathNoAuthMapped;
   mrptr->mpPathSet.AuthOnce = SetPathAuthOnce;
   mrptr->mpPathSet.NoAuthOnce = SetPathNoAuthOnce;
   mrptr->mpPathSet.AuthRevalidateTimeout = SetPathAuthRevalidateTimeout;
   mrptr->mpPathSet.Cache = SetPathCache;
   mrptr->mpPathSet.NoCache = SetPathNoCache;
   mrptr->mpPathSet.DirAccess = SetPathDirAccess;
   mrptr->mpPathSet.DirNoAccess = SetPathDirNoAccess;
   mrptr->mpPathSet.DirAccessSelective = SetPathDirAccessSelective;
   mrptr->mpPathSet.DirImpliedWildcard = SetPathDirImpliedWildcard;
   mrptr->mpPathSet.DirNoImpliedWildcard = SetPathDirNoImpliedWildcard;
   mrptr->mpPathSet.DirWildcard = SetPathDirWildcard;
   mrptr->mpPathSet.DirNoWildcard = SetPathDirNoWildcard;
   mrptr->mpPathSet.Expired = SetPathExpired;
   mrptr->mpPathSet.NoExpired = SetPathNoExpired;
   mrptr->mpPathSet.Log = SetPathLog;
   mrptr->mpPathSet.NoLog = SetPathNoLog;
   mrptr->mpPathSet.MapEllipsis = SetPathMapEllipsis;
   mrptr->mpPathSet.NoMapEllipsis = SetPathNoMapEllipsis;
   mrptr->mpPathSet.MapEmpty = SetPathMapEmpty;
   mrptr->mpPathSet.MapNonEmpty = SetPathMapNonEmpty;
   mrptr->mpPathSet.MapOnce = SetPathMapOnce;
   mrptr->mpPathSet.NoMapOnce = SetPathNoMapOnce;
   mrptr->mpPathSet.OdsNaming = SetPathOds;
   mrptr->mpPathSet.RmsSubChar = RmsSubChar;
   mrptr->mpPathSet.ScriptFind = SetPathScriptFind;
   mrptr->mpPathSet.ScriptNoFind = SetPathScriptNoFind;
   mrptr->mpPathSet.PrivSsi = SetPathPrivSsi;
   mrptr->mpPathSet.NoPrivSsi = SetPathNoPrivSsi;
   mrptr->mpPathSet.Profile = SetPathProfile;
   mrptr->mpPathSet.NoProfile = SetPathNoProfile;
   mrptr->mpPathSet.ReportBasic = SetPathReportBasic;
   mrptr->mpPathSet.ReportDetailed = SetPathReportDetailed;
   mrptr->mpPathSet.Report400as = SetPathReport400as;
   mrptr->mpPathSet.Report403as = SetPathReport403as;
   mrptr->mpPathSet.Report404as = SetPathReport404as;
   mrptr->mpPathSet.StmLF = SetPathStmLF;
   mrptr->mpPathSet.NoStmLF = SetPathNoStmLF;
   mrptr->mpPathSet.SSLCGIvar = SetPathSSLCGIvar;
   mrptr->mpPathSet.ScriptBitBucketTimeout = SetPathScriptBitBucketTimeout;
   mrptr->mpPathSet.ScriptCpuMax = SetPathScriptCpuMax;
   mrptr->mpPathSet.TimeoutKeepAlive = SetPathTimeoutKeepAlive;
   mrptr->mpPathSet.TimeoutNoProgress = SetPathTimeoutNoProgress;
   mrptr->mpPathSet.TimeoutOutput = SetPathTimeoutOutput;

   /* copy all the now null terminated strings as one block */
   cptr = mrptr->TemplatePtr = &mrptr->Storage;
   memcpy (cptr, TemplatePtr, TotalLength);

   /* now adjust the rule pointers to this block of null-terminated strings */
   if (ResultOffset) mrptr->ResultPtr = cptr + ResultOffset;
   if (ConditionalOffset) mrptr->ConditionalPtr = cptr + ConditionalOffset;
   if (SetPathAcceptLangOffset)
   {
      mrptr->mpPathSet.AcceptLangPtr = cptr + SetPathAcceptLangOffset;
      mrptr->mpPathSet.AcceptLangLength =
         strlen(mrptr->mpPathSet.AcceptLangPtr);
   }
   if (SetPathAuthSysUafPwdExpUrlOffset)
   {
      mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr =
         cptr + SetPathAuthSysUafPwdExpUrlOffset;
      mrptr->mpPathSet.AuthSysUafPwdExpUrlLength =
         strlen(mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr);
   }
   if (SetPathCgiPrefixOffset)
   {
      mrptr->mpPathSet.CgiPrefixPtr = cptr + SetPathCgiPrefixOffset;
      mrptr->mpPathSet.CgiPrefixLength = strlen(mrptr->mpPathSet.CgiPrefixPtr);
   }
   if (SetPathCharsetOffset)
   {
      mrptr->mpPathSet.CharsetPtr = cptr + SetPathCharsetOffset;
      mrptr->mpPathSet.CharsetLength = strlen(mrptr->mpPathSet.CharsetPtr);
   }
   if (SetPathContentTypeOffset)
   {
      mrptr->mpPathSet.ContentTypePtr = cptr + SetPathContentTypeOffset;
      mrptr->mpPathSet.ContentTypeLength =
         strlen(mrptr->mpPathSet.ContentTypePtr);
   }
   if (SetPathDirCharsetOffset)
   {
      mrptr->mpPathSet.DirCharsetPtr = cptr + SetPathDirCharsetOffset;
      mrptr->mpPathSet.DirCharsetLength =
         strlen(mrptr->mpPathSet.DirCharsetPtr);
   }
   if (SetPathHttpAcceptCharsetOffset)
   {
      mrptr->mpPathSet.HttpAcceptCharsetPtr =
         cptr + SetPathHttpAcceptCharsetOffset;
      mrptr->mpPathSet.HttpAcceptCharsetLength =
         strlen(mrptr->mpPathSet.HttpAcceptCharsetPtr);
   }
   if (SetPathHttpAcceptLangOffset)
   {
      mrptr->mpPathSet.HttpAcceptLangPtr =
         cptr + SetPathHttpAcceptLangOffset;
      mrptr->mpPathSet.HttpAcceptLangLength =
         strlen(mrptr->mpPathSet.HttpAcceptLangPtr);
   }
   if (SetPathIndexOffset)
   {
      mrptr->mpPathSet.IndexPtr = cptr + SetPathIndexOffset;
      mrptr->mpPathSet.IndexLength = strlen(mrptr->mpPathSet.IndexPtr);
   }
   if (SetPathProxyBindIpAddressOffset)
   {
      mrptr->mpPathSet.ProxyBindIpAddressPtr =
         cptr + SetPathProxyBindIpAddressOffset;
      mrptr->mpPathSet.ProxyBindIpAddressLength =
         strlen(mrptr->mpPathSet.ProxyBindIpAddressPtr);
      if ((mrptr->mpPathSet.ProxyBindIpAddress =
           inet_addr (mrptr->mpPathSet.ProxyBindIpAddressPtr)) == -1)
         MetaConReport (mcptr, "Could not get bind address for !AZ",
                        mrptr->mpPathSet.ProxyBindIpAddressPtr);
      /* the -1 will make any problematic address impossible to bind to! */
   }
   if (SetPathProxyChainHostPortOffset)
   {
      mrptr->mpPathSet.ProxyChainHostPortPtr =
         cptr + SetPathProxyChainHostPortOffset;
      mrptr->mpPathSet.ProxyChainHostPortLength =
         strlen(mrptr->mpPathSet.ProxyChainHostPortPtr);

      status = NetHostNameLookup (mcptr,
                                  mrptr->mpPathSet.ProxyChainHostPortPtr,
                                  0, NULL, NULL, NULL,
                                  &mrptr->mpPathSet.ProxyChainIpAddress,
                                  &mrptr->mpPathSet.ProxyChainPort);
      if (Debug) fprintf (stdout, "NetHostNameLookup() %%X%08.08X\n", status);
      if (!mrptr->mpPathSet.ProxyChainPort)
         mrptr->mpPathSet.ProxyChainPort = DEFAULT_HTTP_PROXY_PORT;
      if (VMSnok (status))
      {
         MetaConReport (mcptr, ProblemHostNameLookup, 
                        mrptr->mpPathSet.ProxyChainHostPortPtr, status);
         /* make sure the connection can't succeed */
         mrptr->mpPathSet.ProxyChainIpAddress = 0xffffffff;
         mrptr->mpPathSet.ProxyChainPort = 0;
      }
   }
   if (SetPathScriptAsOffset)
   {
      mrptr->mpPathSet.ScriptAsPtr = cptr + SetPathScriptAsOffset;
      mrptr->mpPathSet.ScriptAsLength = strlen(mrptr->mpPathSet.ScriptAsPtr);
   }
   if (SetPathScriptParamsOffset)
   {
      mrptr->mpPathSet.ScriptParamsPtr = cptr + SetPathScriptParamsOffset;
      mrptr->mpPathSet.ScriptParamsLength =
         strlen(mrptr->mpPathSet.ScriptParamsPtr);
   }

   if (mrptr->mpPathSet.ThrottleFrom = SetPathThrottleFrom)
   {
      mrptr->mpPathSet.ThrottleIndex = mmptr->ThrottleIndex++;
      if (mmptr->ThrottleIndex > mmptr->ThrottleTotal)
         mmptr->ThrottleTotal = mmptr->ThrottleIndex;
      mrptr->mpPathSet.ThrottleBusy = SetPathThrottleBusy;
      mrptr->mpPathSet.ThrottleResume = SetPathThrottleResume;
      mrptr->mpPathSet.ThrottleTo = SetPathThrottleTo;
      mrptr->mpPathSet.ThrottleTimeoutBusy = SetPathThrottleTimeoutBusy;
      mrptr->mpPathSet.ThrottleTimeoutQueue = SetPathThrottleTimeoutQueue;
   }

   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchDataFormatted ("!&X !UL !&Z !&Z !&B !&Z\n",
         mrptr, mrptr->RuleType, mrptr->TemplatePtr,
         mrptr->ResultPtr, mrptr->PathSet, mrptr->ConditionalPtr);

   return (true);
}

/*****************************************************************************/
/*
Convert a URL-style specification into a RMS-style specification, example:
"/disk/dir1/dir2/file.txt" into "disk:[dir1.dir2]file.txt".

Can process ODS-2, ODS-5 (EFS), SRI (MultiNet NFS), PATHWORKS (v4/5) and
Advanced Server (PATHWORKS V6) / Samba encodings.

Returns the length of the VMS specification.
*/ 
 
int MapUrl_UrlToVms
(
char *PathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
char RmsSubChar,
BOOL MapEllipsis,
int PathOdsNaming
)
{
   int  len;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_UrlToVms() !UL !UL !&B !UL !&Z",
                 PathOdsNaming, RmsSubChar, MapEllipsis, SizeOfVmsPtr, PathPtr);

   switch (PathOdsNaming)
   {
      case 0 :
      case MAPURL_PATH_ODS_2 :
         len = MapUrl_UrlToOds2Vms (PathPtr, VmsPtr, SizeOfVmsPtr,
                                    MapEllipsis, RmsSubChar);
         break;

#ifdef ODS_EXTENDED
      case MAPURL_PATH_ODS_5 :
         len = MapUrl_UrlToOds5Vms (PathPtr, VmsPtr, SizeOfVmsPtr, MapEllipsis);
         break;
#endif /* ODS_EXTENDED */

      case MAPURL_PATH_ODS_ADS :
      case MAPURL_PATH_ODS_SMB :
         len = MapUrl_UrlToAdsVms (PathPtr, VmsPtr, SizeOfVmsPtr, MapEllipsis);
         break;

      case MAPURL_PATH_ODS_PWK :
         len = MapUrl_UrlToPwkVms (PathPtr, VmsPtr, SizeOfVmsPtr, MapEllipsis);
         break;

      case MAPURL_PATH_ODS_SRI :
         len = MapUrl_UrlToSriVms (PathPtr, VmsPtr, SizeOfVmsPtr, MapEllipsis);
         break;

      default :
         memcpy (VmsPtr, "SANITY:[CHECK]MAPURL_URLTOVMS", len=30);
   }

   /* ensure it's not mistaken for an error message! */
   if (!VmsPtr[0]) VmsPtr[1] = '\0';

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, VmsPtr);

   return (len);
}

/*****************************************************************************/
/*
Convert a URL-style specification into an ODS-2 RMS-style specification,
example: "/disk/dir1/dir2/file.txt" into "disk:[dir1.dir2]file.txt".

Forbidden characters (non-alphanumberic and non-"$_-") are converted to the
'RmsSubChar' (by default a dollar symbol).  Where a specification has multiple
".", all but the final one of the file component is also converted to a
'RmsSubChar'.

Returns the length of the VMS specification.
*/ 
 
int MapUrl_UrlToOds2Vms
(
char *PathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
BOOL MapEllipsis,
char RmsSubChar
)
{
   int  ecnt, len;
   char  *cptr, *eptr, *pptr, *sptr, *zptr,
         *FinalPeriodPtr;
   char  PathComponents [ODS_MAX_FILE_NAME_LENGTH];

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_UrlToOds2Vms() !&B !UL !&Z !UL",
                 MapEllipsis, RmsSubChar, PathPtr, SizeOfVmsPtr);

   pptr = PathPtr;
   if (!pptr || !*pptr || *(unsigned short*)pptr == '/\0')
   {
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }

   ecnt = 0;
   FinalPeriodPtr = NULL;
   zptr = (sptr = PathComponents) + sizeof(PathComponents);

   /* each '/' component becomes a null-terminated string */
   while (*pptr && sptr < zptr)
   {
      /* multiple consecutive might as well be just one */
      while (*pptr == '/') pptr++;
      eptr = sptr;
      while (*pptr && *pptr != '/' && sptr < zptr)
      {
         if (isalnum (*pptr) ||
             *pptr == '$' || *pptr == '_' || *pptr == '-' || *pptr == '.' ||
             *pptr == '*' || *pptr == '%' || *pptr == ';')
         {
            /* ODS-2 legal character (or wildcard) */
            *sptr++ = toupper(*pptr++);
            continue;
         }
         if (!ecnt)
         {
            /* DECnet only valid in first component */
            if (*pptr == '\"' || *pptr == '=')
            {
               /* DECnet access string (allow otherwise forbidden chars) */
               if (*(unsigned short*)pptr == '\"~' ||
                   *(unsigned short*)pptr == '\"$')
                  *sptr++ = *pptr++;
               if (sptr < zptr) *sptr++ = *pptr++;
               continue;
            }
            if (*(unsigned short*)pptr == '::')
            {
               /* DECnet (with or without access string) */
               *sptr++ = *pptr++;
               if (sptr < zptr) *sptr++ = *pptr++;
               continue;
            }
         }
         /* any other character substitute a question mark */
         if (*pptr == '.') FinalPeriodPtr = sptr;
         if (sptr < zptr) *sptr++ = '?';
         pptr++;
      }
      if (sptr < zptr) *sptr++ = '\0';
      ecnt++;
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!UL !&Z\n", ecnt, eptr);
   }

   if (sptr >= zptr)
   {
      /* storage overflowed */
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }
   *sptr = '\0';

   if (FinalPeriodPtr) *FinalPeriodPtr = '.';

   /* convert these null-terminated elements into a VMS file specification */
   len = MapUrl_ElementsToVms (PathComponents, ecnt, VmsPtr, SizeOfVmsPtr,
                               MapEllipsis);

   /* convert the non-RMS characters into the RMS substitution character */
   if (!RmsSubChar) RmsSubChar = MAPURL_RMS_SUBSTITUTION_DEFAULT;
   for (cptr = VmsPtr; *cptr; cptr++)
   {
      if (*cptr != '?') continue;
      *cptr = RmsSubChar;
   }

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, VmsPtr);

   return (len);
}

/*****************************************************************************/
/*
Convert a URL-style specification into an ODS-5 RMS-style specification,
example: "/disk/dir1/dir2/file.tar.gz" into "disk:[dir1.dir2]file^.tar.gz".

For ODS-5 paths forbidden characters are ^ (character or hexadecimal) escaped
and the case not altered.  Refer to the EFS sections of the RMS documentaion.

Returns the length of the VMS specification.
*/ 
 
#ifdef ODS_EXTENDED

int MapUrl_UrlToOds5Vms
(
char *PathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
BOOL MapEllipsis
)
{
   static char  HexDigits [] = "0123456789ABCDEF";

   BOOL  DECnet;
   int  ecnt, len;
   char  *cptr, *eptr, *pptr, *sptr, *zptr,
         *FinalPeriodPtr,
         *FinalSemiColonPtr,
         *FinalSlashPtr;
   char  PathComponents [ODS_MAX_FILE_NAME_LENGTH];

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_UrlToOds5Vms() !&B !&Z !UL",
                 MapEllipsis, PathPtr, SizeOfVmsPtr);

   pptr = PathPtr;
   if (!pptr || !*pptr || *(unsigned short*)pptr == '/\0')
   {
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }

   ecnt = 0;
   FinalPeriodPtr = FinalSemiColonPtr = FinalSlashPtr = NULL;
   zptr = (sptr = PathComponents) + sizeof(PathComponents);

   /* each '/' component becomes a null-terminated string */
   while (*pptr && sptr < zptr)
   {
      /* multiple consecutive might as well be just one */
      while (*pptr == '/') pptr++;
      eptr = sptr;
      while (*pptr && *pptr != '/' && sptr < zptr)
      {
         switch (*pptr)
         {
            case '!' : case '#' : case '&' : case '\'' :
            case '`' : case '(' : case ')' : case '+' :
            case '@' : case '{' : case '}' : case ',' :
            case ';' : case '[' : case ']' : case '=' :
               if (*pptr == ';') FinalSemiColonPtr = sptr;
               *sptr++ = '^';
               if (sptr < zptr) *sptr++ = *pptr++;
               break;

            case ' ' :
               *sptr++ = '^';
               if (sptr < zptr) *sptr++ = '_';
               pptr++;
               break;

            case '.' :
               if (*pptr == '.' && *(unsigned short*)(pptr+1) == '..')
               {
                  /* ellipsis wildcard */
                  *sptr++ = *pptr++;
                  if (sptr < zptr) *sptr++ = *pptr++;
                  if (sptr < zptr) *sptr++ = *pptr++;
               }
               else
               {
                  FinalPeriodPtr = sptr;
                  *sptr++ = '^';
                  if (sptr < zptr) *sptr++ = *pptr++;
               }
               break;

            case '^' :
               *sptr++ = *pptr++;
               if (isxdigit(*pptr))
               {
                  if (sptr < zptr) *sptr++ = *pptr++;
                  if (sptr < zptr && *pptr) *sptr++ = *pptr++;
               }
               else
               if (sptr < zptr && *pptr)
                  *sptr++ = *pptr++;
               break;

            default :
               if (*pptr == 0x07f ||
                   *pptr == 0x0a0 ||
                   *pptr == 0x0ff ||
                   (*pptr >= 0x80 && *pptr <= 0x9f))
               {
                  if (sptr+3 < zptr)
                  {
                     *sptr++ = '^';
                     *sptr++ = HexDigits[(*pptr & 0xf0) >> 4];
                     *sptr++ = HexDigits[*pptr++ & 0x0f];
                  }
                  else
                     sptr = zptr;
               }
               else
                  *sptr++ = *pptr++;
         }
      }
      if (*pptr == '/') FinalSlashPtr = sptr;
      if (sptr < zptr) *sptr++ = '\0';
      ecnt++;
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!UL !&Z\n", ecnt, eptr);
   }

   if (sptr >= zptr)
   {
      /* storage overflowed */
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }
   *sptr = '\0';

   if (FinalSemiColonPtr &&
       (FinalSemiColonPtr > FinalSlashPtr || !FinalSlashPtr) &&
       sptr < zptr)
   {
      /* a final semicolon delimitting the version must not be escaped */
      cptr = (sptr = FinalSemiColonPtr) + 1;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }

   if (FinalPeriodPtr &&
       (FinalPeriodPtr > FinalSlashPtr || !FinalSlashPtr) &&
       sptr < zptr)
   {
      /* a final period delimitting the type must not be escaped */
      cptr = (sptr = FinalPeriodPtr) + 1;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }

   len = MapUrl_ElementsToVms (PathComponents, ecnt, VmsPtr, SizeOfVmsPtr,
                               MapEllipsis);

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, VmsPtr);

   return (len);
}

#endif /* ODS_EXTENDED */

/*****************************************************************************/
/*
Convert a URL-style specification into a RMS-style specification, example:
"/disk/dir1/dir2/file.tar.gz" into "disk:[dir1.dir2]file__27tar.gz" - with
Advanced Server / Samba character conversion. If a character is not RMS legal
(excluding the dollar which is the escape) then encode it as a
double-underscore-escaped hexadecimal number.

Something like this ...
1) If an ODS-2 forbidden character replace with '__xx' (i.e. two underscores
plus the hex representation of the character).
2) Successive underscores have the first encoded as '__xx'.
3) If the final period occurs within the first 39 characters then it's not
encoded, it becomes the ODS-2 period.  If it occurs after character 39 then
it's left encoded and a (required ODS-2 but name-extraneous) period is inserted
at character 40. 
4) Case change is indicated with a leading '__$' (begins as lower).

Seems cumbersome, less versatile and less space-conservative than SRI!!
For the life of me, why didn't they use SRI encoding?

Returns the length of the VMS specification.
*/ 
 
int MapUrl_UrlToAdsVms
(
char *PathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
BOOL MapEllipsis
)
{
   static char  HexDigits [] = "0123456789ABCDEF";

   BOOL  inDevice,
         loCase;
   int  ecnt, len;
   char  *cptr, *eptr, *pptr, *sptr, *zptr,
         *FinalPeriodPtr,
         *FinalSlashPtr;
   char  PathComponents [ODS_MAX_FILE_NAME_LENGTH];

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_UrlToAdsVms() !&B !UL !&Z",
                 MapEllipsis, SizeOfVmsPtr, PathPtr);

   pptr = PathPtr;
   if (!pptr || !*pptr || *(unsigned short*)pptr == '/\0')
   {
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }

   /* any 'device' will not have any encoded characters in it's name */
   for (cptr = pptr; *cptr == '/'; cptr++);
   while (*cptr && *cptr != '/') cptr++;
   if (*cptr) inDevice = true; else inDevice = false;

   FinalPeriodPtr = FinalSlashPtr = NULL;
   ecnt = 0;
   /* leave an extra char elbow room for inserting a period - see below */
   zptr = (sptr = PathComponents) + sizeof(PathComponents)-1;
   eptr = sptr;

   /* each component becomes a null-terminated string in a character array */
   while (*pptr && sptr < zptr)
   {
      /* multiple consecutive might as well be just one */
      while (*pptr == '/') pptr++;
      FinalPeriodPtr = NULL;
      loCase = true;
      eptr = sptr;
      while (*pptr && *pptr != '/' && sptr < zptr)
      {
         if (isalnum (*pptr) || *pptr == '-' || *pptr == '$' ||
             *pptr == '*' || *pptr == '%' || inDevice)
         {
            if (loCase && isupper(*pptr))
            {
               loCase = false;
               *sptr++ = '_';
               if (sptr < zptr) *sptr++ = '_';
               if (sptr < zptr) *sptr++ = '$';
               if (sptr < zptr) *sptr++ = toupper(*pptr++);
               continue;
            }
            if (!loCase && islower(*pptr))
            {
               loCase = true;
               *sptr++ = '_';
               if (sptr < zptr) *sptr++ = '_';
               if (sptr < zptr) *sptr++ = '$';
               if (sptr < zptr) *sptr++ = toupper(*pptr++);
               continue;
            }
            *sptr++ = toupper(*pptr++);
            continue;
         }
         if (*pptr == '_' && *(unsigned short*)pptr != '__')
         {
            /* single underscore */
            *sptr++ = *pptr++;
            continue;
         }
         if (*pptr == '.') FinalPeriodPtr = sptr;
         *sptr++ = '_';
         if (sptr < zptr) *sptr++ = '_';
         if (sptr < zptr) *sptr++ = HexDigits[(*pptr & 0xf0) >> 4];
         if (sptr < zptr) *sptr++ = HexDigits[*pptr & 0x0f];
         pptr++;
      }
      if (sptr < zptr) *sptr++ = '\0';
      ecnt++;
      inDevice = false;
      if (*pptr == '/') FinalSlashPtr = sptr;
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!UL !&Z\n", ecnt, eptr);
   }

   if (sptr >= zptr)
   {
      /* storage overflowed */
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }
   *sptr = '\0';

   if (FinalPeriodPtr && FinalPeriodPtr <= eptr + 39)
   {
      /* less than required ODS-2 period, replace with 'real' period */
      sptr = FinalPeriodPtr;
      *sptr++ = '.';
      cptr = sptr + 3;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }
   else
   if (sptr > eptr + 39)
   {
      /* greater than required ODS-2 period, insert an ODS-2 period */
      cptr = sptr++;
      while (cptr > eptr+39) *sptr-- = *cptr--;
      *sptr = *cptr;
      *cptr = '.';
      while (*sptr) sptr++;
   }

   len = MapUrl_ElementsToVms (PathComponents, ecnt, VmsPtr, SizeOfVmsPtr,
                               MapEllipsis);

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, VmsPtr);

   return (len);
}

/*****************************************************************************/
/*
Convert a URL-style specification into a RMS-style specification, example:
"/disk/dir1/dir2/file.tar.gz" into "disk:[dir1.dir2]file$27tar.gz" - with
PATHWORKS v4/5 character conversion.  This is much simpler than SRI.  If a
character  is not RMS legal (excluding the dollar which is the escape) then
encode it as a dollar-escaped hexadecimal number.

Returns the length of the VMS specification.
*/ 
 
int MapUrl_UrlToPwkVms
(
char *PathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
BOOL MapEllipsis
)
{
   static char  HexDigits [] = "0123456789ABCDEF";

   BOOL  inDevice;
   int  ecnt, len;
   char  *cptr, *eptr, *pptr, *sptr, *zptr,
         *FinalPeriodPtr,
         *FinalSlashPtr;
   char  PathComponents [ODS_MAX_FILE_NAME_LENGTH];

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_UrlToPwkVms() !&B !UL !&Z",
                 MapEllipsis, SizeOfVmsPtr, PathPtr);

   pptr = PathPtr;
   if (!pptr || !*pptr || *(unsigned short*)pptr == '/\0')
   {
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }

   /* any 'device' will not have any encoded characters in it's name */
   for (cptr = pptr; *cptr == '/'; cptr++);
   while (*cptr && *cptr != '/') cptr++;
   if (*cptr) inDevice = true; else inDevice = false;

   FinalPeriodPtr = FinalSlashPtr = NULL;
   ecnt = 0;
   zptr = (sptr = PathComponents) + sizeof(PathComponents);

   /* each component becomes a null-terminated string in a character array */
   while (*pptr && sptr < zptr)
   {
      /* multiple consecutive might as well be just one */
      while (*pptr == '/') pptr++;
      eptr = sptr;
      while (*pptr && *pptr != '/' && sptr < zptr)
      {
         if (isalnum (*pptr) || *pptr == '-' || *pptr == '_' ||
                                *pptr == '*' || *pptr == '%' || inDevice)
         {
            *sptr++ = *pptr++;
            continue;
         }
         if (*pptr == '.') FinalPeriodPtr = sptr;
         *sptr++ = '$';
         if (sptr < zptr) *sptr++ = HexDigits[(*pptr & 0xf0) >> 4];
         if (sptr < zptr) *sptr++ = HexDigits[*pptr & 0x0f];
         pptr++;
      }
      if (sptr < zptr) *sptr++ = '\0';
      ecnt++;
      inDevice = false;
      if (*pptr == '/') FinalSlashPtr = sptr;
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!UL !&Z\n", ecnt, eptr);
   }

   if (sptr >= zptr)
   {
      /* storage overflowed */
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }
   *sptr = '\0';

   /* turn the final period back into one */
   if (FinalPeriodPtr)
   {
      sptr = FinalPeriodPtr;
      *sptr++ = '.';
      for (cptr = sptr + 2; *cptr; *sptr++ = *cptr++);
      *sptr = '\0';
   }

   len = MapUrl_ElementsToVms (PathComponents, ecnt, VmsPtr, SizeOfVmsPtr,
                               MapEllipsis);

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, VmsPtr);

   return (len);
}

/*****************************************************************************/
/*
Convert a URL-style specification into a RMS-style specification, example:
"/disk/dir1/dir2/file.tar.gz" into "disk:[dir1.dir2]file.tar$27gz" - with SRI
character conversion.  The SRI scheme is quite complex.  Refer to online
documentation associated with Process Software's MultiNet NFS server. 
Alphabetic '$' case flags reset on a per-path-component basis.

Returns the length of the VMS specification.
*/ 
 
int MapUrl_UrlToSriVms
(
char *PathPtr,
char *VmsPtr,
int SizeOfVmsPtr,
BOOL MapEllipsis
)
{
   /* directory wildcard concessions; '%' should be "$5E" and '*' be "$5J" */
   static char  *SriConvTable [] = {
"$6A", "$4A", "$4B", "$4C", "$4D", "$4E", "$4F", "$4G",  /* 0x07 */
"$4H", "$4I", "$4J", "$4K", "$4L", "$4M", "$4N", "$4O",  /* 0x0f */
"$4P", "$4Q", "$4R", "$4S", "$4T", "$4U", "$4V", "$4W",  /* 0x17 */
"$4X", "$4Y", "$4Z", "$6B", "$6C", "$6D", "$6E", "$6F",  /* 0x1f */
"$7A", "$5A", "$5B", "$5C", "$$",  NULL,  "$5F", "$5G",  /* 0x27 */
"$5H", "$5I", NULL,  "$5K", "$5L", NULL,  "$5N", "$5O",  /* 0x2f */
NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x37 */
NULL,  NULL,  "$5Z", "$7B", "$7C", "$7D", "$7E", "$7F",  /* 0x3f */
"$8A", NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x47 */
NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x4f */
NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x57 */
NULL,  NULL,  NULL,  "$8B", "$8C", "$8D", "$8E", NULL,   /* 0x5f */
"$9A", NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x67 */
NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x6f */
NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL,   /* 0x77 */
NULL,  NULL,  NULL,  "$9B", "$9C", "$9D", "$9E", "$9F",  /* 0x7f */
"$200", "$201", "$202", "$203", "$204", "$205", "$206", "$207", /* 0x87 */
"$210", "$211", "$212", "$213", "$214", "$215", "$216", "$217", /* 0x8f */
"$220", "$221", "$222", "$223", "$224", "$225", "$226", "$227", /* 0x97 */
"$230", "$231", "$232", "$233", "$234", "$235", "$236", "$237", /* 0x9f */
"$240", "$241", "$242", "$243", "$244", "$245", "$246", "$247", /* 0xa7 */
"$250", "$251", "$252", "$253", "$254", "$255", "$256", "$257", /* 0xaf */
"$260", "$261", "$262", "$263", "$264", "$265", "$266", "$267", /* 0xb7 */
"$270", "$271", "$272", "$273", "$274", "$275", "$276", "$277", /* 0xbf */
"$300", "$301", "$302", "$303", "$304", "$305", "$306", "$307", /* 0xc7 */
"$310", "$311", "$312", "$313", "$314", "$315", "$316", "$317", /* 0xcf */
"$320", "$321", "$322", "$323", "$324", "$325", "$326", "$327", /* 0xd7 */
"$330", "$331", "$332", "$333", "$334", "$335", "$336", "$337", /* 0xdf */
"$340", "$341", "$342", "$343", "$344", "$345", "$346", "$347", /* 0xe7 */
"$350", "$351", "$352", "$353", "$354", "$355", "$356", "$357", /* 0xef */
"$360", "$361", "$362", "$363", "$364", "$365", "$366", "$367", /* 0xf7 */
"$370", "$371", "$372", "$373", "$374", "$375", "$376", "$377", /* 0xff */
};

   BOOL  inDevice,
         loCase,
         typeHit;
   int  ecnt, len, pcnt;
   char  *cptr, *eptr, *pptr, *sptr, *zptr,
         *FinalSlashPtr;
   char  PathComponents [ODS_MAX_FILE_NAME_LENGTH];

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_UrlToSriVms() !&B !UL !&Z",
                 MapEllipsis, SizeOfVmsPtr, PathPtr);

   pptr = PathPtr;
   if (!pptr || !*pptr || *(unsigned short*)pptr == '/\0')
   {
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }

   /* any 'device' will not have any encoded characters in it's name */
   for (cptr = pptr; *cptr == '/'; cptr++);
   while (*cptr && *cptr != '/') cptr++;
   if (*cptr) inDevice = true; else inDevice = false;

   ecnt = pcnt = 0;
   FinalSlashPtr = NULL;
   typeHit = false;
   zptr = (sptr = PathComponents) + sizeof(PathComponents);

   /* each component becomes a null-terminated string in a character array */
   while (*pptr && sptr < zptr)
   {
      /* multiple consecutive might as well be just one */
      while (*pptr == '/') pptr++;
      loCase = true;
      eptr = sptr;
      while (*pptr && *pptr != '/' && sptr < zptr)
      {
         if (*pptr == '.' && !typeHit)
         {
            /* is it the final period in the specification? */
            for (cptr = pptr; *cptr && *cptr != '/'; cptr++);
            if (!*cptr)
            {
               /* yes! */
               for (cptr = pptr; *cptr && *cptr != '.'; cptr++);
               if (*cptr)
               {
                  typeHit = true;
                  loCase = true;
               }
            }
         }
         if (!SriConvTable[*(unsigned char*)pptr] || inDevice)
         {
            /* unencoded alpha-numeric (or other legal) character */
            if (loCase && isupper(*pptr))
            {
               loCase = false;
               *sptr++ = '$';
               if (sptr < zptr) *sptr++ = toupper(*pptr++);
               continue;
            }
            if (!loCase && islower(*pptr))
            {
               loCase = true;
               *sptr++ = '$';
               if (sptr < zptr) *sptr++ = toupper(*pptr++);
               continue;
            }
            *sptr++ = toupper(*pptr++);
            continue;
         }
         /* dollar-encoded character */
         cptr = SriConvTable[*(unsigned char*)pptr];
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         pptr++;
      }
      if (sptr < zptr) *sptr++ = '\0';
      ecnt++;
      inDevice = false;
      if (*pptr == '/') FinalSlashPtr = sptr;
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!UL !&Z\n", ecnt, eptr);
   }

   if (sptr >= zptr)
   {
      /* storage overflowed */
      if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }
   *sptr = '\0';

   /* turn the first (if any) '$5N' of the file name back into a period */
   if (!(sptr = FinalSlashPtr)) sptr = PathComponents;
   while (*sptr)
   {
      if (*sptr++ != '$') continue;
      if (*(unsigned short*)sptr != '5N') continue;
      sptr[-1] = '.';
      for (cptr = sptr + 2; *cptr; *sptr++ = *cptr++);
      *sptr = '\0';
      break;
   }

   len = MapUrl_ElementsToVms (PathComponents, ecnt, VmsPtr, SizeOfVmsPtr,
                               MapEllipsis);

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, VmsPtr);

   return (len);
}

/*****************************************************************************/
/*
'ElementsPtr' contains a series of null-delimited strings (the '/' path
components) terminated by a empty string.  'ElementsCount' the number of
'/'-delimited parts of the string, including the "file" element - even if
empty.  Turn this into a VMS file specification in storage pointed to by
'VmsPtr'.

Returns the length of the VMS specification.
*/ 
 
int MapUrl_ElementsToVms
(
char *ElementsPtr,
int ElementsCount,
char *VmsPtr,
int SizeOfVmsPtr,
BOOL MapEllipsis
)
{
   BOOL  DECnet;
   int  ecnt;
   char  *cptr, *sptr, *tptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_ElementsToVms() !UL !&B", ElementsCount, MapEllipsis);

   DECnet = false;
   for (cptr = ElementsPtr; *cptr; cptr++)
   {
      if (*(unsigned short*)cptr != '::') continue;
      DECnet = true;
      if (WATCH_MODULE(WATCH_MOD_MAPURL))
         WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "DECNET");
      break;
   }

   cptr = ElementsPtr;
   zptr = (sptr = VmsPtr) + SizeOfVmsPtr;

   if (DECnet) ElementsCount--;
   if (ElementsCount == 1)
   {
      /********************/
      /* single component */
      /********************/

      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   else
   if (ElementsCount == 2)
   {
      /***********************/
      /* top-level directory */
      /***********************/

      if (DECnet)
      {
         /* copy first component as-is */
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         cptr++;
      }
      /* start with the first component, the "device" */
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      cptr++;
      /* append a Master File Directory component */
      for (tptr = ":[000000]"; *tptr && sptr < zptr; *sptr++ = *tptr++);
      /* add the second component, the "file" */
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   else
   if (ElementsCount >= 3)
   {
      /****************/
      /* subdirectory */
      /****************/

      if (DECnet)
      {
         /* copy first component as-is */
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         cptr++;
      }
      /* start with the first component, the "device" */
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      cptr++;
      if (sptr < zptr) *sptr++ = ':';
      if (sptr < zptr) *sptr++ = '[';
      if (cptr[0] == '.' && cptr[1] == '.' && cptr[2] == '.' && MapEllipsis)
         for (tptr = "000000"; *tptr && sptr < zptr; *sptr++ = *tptr++);
      for (ecnt = 2; ecnt < ElementsCount; ecnt++)
      {
         /* add the next component, a "directory" */
         if (cptr[0] == '-')
         {
            /* check for parent directory syntax, excise if it is! */
            for (tptr = cptr; *tptr == '-' || *tptr == '.'; tptr++);
            if (*tptr)
            {
               /* not parent directory syntax */
               if (*(sptr-1) != '[' && sptr < zptr) *sptr++ = '.';
               while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            }
            else
               cptr = tptr;
         }
         else
         {
            if (*(sptr-1) != '[' && *cptr != '.' && sptr < zptr) *sptr++ = '.';
            while (*cptr && sptr < zptr)
            {
               if (*cptr != '.')
               {
                  *sptr++ = *cptr++;
                  continue;
               }
               if (*(unsigned short*)(cptr+1) != '..' || MapEllipsis)
               {
                  *sptr++ = *cptr++;
                  continue;
               }
               /* excise the ellipsis wildcard */
               cptr += 3;
               if (*(sptr-1) == '.') sptr--;
            }
         }
         cptr++;
      }
      if (*(sptr-1) == '[')
      {
         /* the entire directory structure has been excised! */
         for (tptr = "000000"; *tptr && sptr < zptr; *sptr++ = *tptr++);
      }
      if (sptr < zptr) *sptr++ = ']';
      /* add the last component, the "file" */
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr < zptr) *sptr = '\0';

   if (sptr >= zptr)
   {
      /* storage overflowed */
      if (Debug) fprintf (stdout, "VmsPtr OVERFLOW\n");
      *(unsigned short*)VmsPtr = '\0\0';
      return (0);
   }

   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchDataFormatted ("!&B !UL !&Z\n", DECnet, sptr-VmsPtr, VmsPtr);

   return (sptr - VmsPtr);
}

/*****************************************************************************/
/*
Convert a VMS full file specification into a URL-style specification, example:
"DEVICE:[DIR1.DIR2]FILE.TXT" into "/device/dir1/dir2/file.txt", or just a file
name and type portion (ensuring it's correctly converted and escaped).  The URL
has URL forbidden characters URL-encoded.

Can process ODS-2, ODS-5 (EFS), SRI (MultiNet NFS), PATHWORKS (v4/5) and
Advanced Server (PATHWORKS V6) / Samba encodings.

Returns the length of the generated URL-style path.
*/ 
 
int MapUrl_VmsToUrl
(
char *PathPtr,
char *VmsPtr,
int SizeOfPathPtr,
BOOL AbsorbMfd,
int PathOdsNaming
)
{
   int  len;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_VmsToUrl() !UL !&B !&Z !UL",
                 PathOdsNaming, AbsorbMfd, VmsPtr, SizeOfPathPtr);

   switch (PathOdsNaming)
   {
      case 0 :
      case MAPURL_PATH_ODS_2 :
         len = MapUrl_Ods2VmsToUrl (PathPtr, VmsPtr, SizeOfPathPtr, AbsorbMfd);
         break;

#ifdef ODS_EXTENDED
      case MAPURL_PATH_ODS_5 :
         len = MapUrl_Ods5VmsToUrl (PathPtr, VmsPtr, SizeOfPathPtr, AbsorbMfd);
         break;
#endif /* ODS_EXTENDED */

      case MAPURL_PATH_ODS_ADS :
      case MAPURL_PATH_ODS_SMB :
         len = MapUrl_AdsVmsToUrl (PathPtr, VmsPtr, SizeOfPathPtr, AbsorbMfd);
         break;

      case MAPURL_PATH_ODS_PWK :
         len = MapUrl_PwkVmsToUrl (PathPtr, VmsPtr, SizeOfPathPtr, AbsorbMfd);
         break;

      case MAPURL_PATH_ODS_SRI :
         len = MapUrl_SriVmsToUrl (PathPtr, VmsPtr, SizeOfPathPtr, AbsorbMfd);
         break;

      default :
         memcpy (PathPtr, "/sanity/check/mapurl_urltovms", len=30);
   }

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!UL !&Z\n", len, PathPtr);

   return (len);
}

/*****************************************************************************/
/*
Convert a VMS full file specification into a URL-style specification, example:
"DEVICE:[DIR1.DIR2]FILE.TXT" into "/device/dir1/dir2/file.txt", or just a file
name and type portion (ensuring it's correctly converted and escaped).  Expects
file specification to be ODS-2 compliant and forces the URL to all lower-case.

Returns the length of the generated URL-style path.
*/ 
 
int MapUrl_Ods2VmsToUrl
(
char *PathPtr,
char *VmsPtr,
int SizeOfPathPtr,
BOOL AbsorbMfd
)
{
   char  *cptr, *eptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_Ods2VmsToUrl() !&B !&Z !UL",
                 AbsorbMfd, VmsPtr, SizeOfPathPtr);

   zptr = (sptr = PathPtr) + SizeOfPathPtr;

   /* look for indications it's a full file specification */
   cptr = VmsPtr;
   while (*cptr && *(unsigned short*)cptr != ':[' &&
          *cptr != ']' && *cptr != '/') cptr++;

   if (!*cptr || *cptr == '/')
   {
      /* no VMS device/directory components detected */
      cptr = VmsPtr;
   }
   else
   {
      /* copy device/directory components */
      cptr = VmsPtr;
      if (sptr < zptr) *sptr++ = '/';
      while (*cptr && sptr < zptr)
      {
         if (*(unsigned short*)cptr == '::')
         {
            /* DECnet */
            *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = '/';
            continue;
         }

         if (*(unsigned short*)cptr == ':[')
         {
            cptr++;
            cptr++;
            /* remove any reference to a Master File Directory */
            if (AbsorbMfd && strncmp (cptr, "000000", 6) == 0)
               cptr += 6;
            else
               *sptr++ = '/';
            continue;
         }

         if (*cptr == '.')
         {
            *sptr++ = '/';
            if (*cptr == '.' && *(unsigned short*)(cptr+1) == '..')
            {
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
            }
            else
               cptr++;
            continue;
         }

         if (*cptr == ']')
         {
            *sptr++ = '/';
            cptr++;
            break;
         }

         *sptr++ = tolower(*cptr++);
      }
   }

   /* copy the file component */
   while (*cptr && sptr < zptr) *sptr++ = tolower(*cptr++);
   if (sptr < zptr) *sptr++ = '\0';
   if (sptr < zptr) *sptr++ = '\0';

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!&Z\n", PathPtr);

   if (sptr < zptr) return ((sptr - PathPtr) - 2);

   /* storage overflowed */
   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
   *(unsigned short*)PathPtr = '\0\0';
   return (0);
}

/*****************************************************************************/
/*
Convert a VMS full file ODS-5 compliant specification into a URL-style
specification, example: "DEVICE:[dir1.DIR2]a^.file.txt" into
"/device/dir1/DIR2/a file.txt", or a URL-style string that contains ODS-5
escaped   characters, example "/device/device/dir1/DIR2/a^.file.txt" into
"/device/dir1/DIR2/a file.txt" (ensuring it's correctly converted and escaped). 
Will decode literal ^-escaped and hexadecimal ^-escaped characters and leaves
the case as-is.

Returns the length of the generated URL-style path.
*/ 
 
#ifdef ODS_EXTENDED

int MapUrl_Ods5VmsToUrl
(
char *PathPtr,
char *VmsPtr,
int SizeOfPathPtr,
BOOL AbsorbMfd
)
{
   char  ch;
   char  *cptr, *eptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_Ods5VmsToUrl() !&B !&Z !UL",
                 AbsorbMfd, VmsPtr, SizeOfPathPtr);

   zptr = (sptr = PathPtr) + SizeOfPathPtr;

   /* look for indications it's a full file specification */
   cptr = VmsPtr;
   while (*cptr && *(unsigned short*)cptr != ':[' &&
          *cptr != ']' && *cptr != '/')
   {
      if (cptr[0] == '^' && isxdigit(cptr[1]) && isxdigit(cptr[2]))
         cptr += 3;
      else
      if (cptr[0] == '^' && cptr[1] == 'U' &&
          isxdigit(cptr[2]) && isxdigit(cptr[3]) &&
          isxdigit(cptr[4]) && isxdigit(cptr[5]))
         cptr += 6;
      else
      if (cptr[0] == '^')
         cptr += 2;
      else
         cptr++;
   }

   if (!*cptr || *cptr == '/')
   {
      /* no VMS device/directory components detected */
      cptr = VmsPtr;
   }
   else
   {
      /* copy device/directory components */
      cptr = VmsPtr;
      if (sptr < zptr) *sptr++ = '/';
      while (*cptr && sptr < zptr)
      {
         if (*(unsigned short*)cptr == '::')
         {
            /* DECnet */
            *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = '/';
            continue;
         }

         if (*(unsigned short*)cptr == ':[')
         {
            cptr++;
            cptr++;
            /* remove any reference to a Master File Directory */
            if (AbsorbMfd && strncmp (cptr, "000000", 6) == 0)
               cptr += 6;
            else
               *sptr++ = '/';
            continue;
         }

         if (*cptr == '.')
         {
            *sptr++ = '/';
            if (*cptr == '.' && *(unsigned short*)(cptr+1) == '..')
            {
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
            }
            else
               cptr++;
            continue;
         }

         if (*cptr == ']')
         {
            *sptr++ = '/';
            cptr++;
            break;
         }

         if (*cptr == '^')
         {
            /* ODS-5 extended specification escape character */
            cptr++;
            switch (*cptr)
            {
               case '!' : case '#' : case '&' : case '\'' : case '`' :
               case '(' : case ')' : case '+' : case '@'  : case '{' :
               case '}' : case '.' : case ',' : case ';'  : case '[' :
               case ']' : case '%' : case '^' : case '='  : case ' ' :
                  *sptr++ = *cptr++;
                  break;
               case '_' :
                  *sptr++ = ' ';
                  cptr++;
                  break;
               default :
                  if (!isxdigit (*cptr))
                  {
                     *sptr++ = *cptr++;
                     break;
                  }
                  ch = 0;
                  if (*cptr >= '0' && *cptr <= '9')
                     ch = (*cptr - (int)'0') << 4;
                  else
                  if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
                     ch = (tolower(*cptr) - (int)'a' + 10) << 4;
                  if (*cptr) cptr++;
                  if (*cptr >= '0' && *cptr <= '9')
                     ch += (*cptr - (int)'0');
                  else
                  if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
                     ch += (tolower(*cptr) - (int)'a' + 10);
                  if (*cptr) cptr++;
                  *sptr++ = ch;
            }
   
            continue;
         }

         if (sptr < zptr) *sptr++ = *cptr++;
      }
   }

   /* copy the file component */
   while (*cptr && sptr < zptr)
   {
      if (*cptr != '^')
      {
         *sptr++ = *cptr++;
         continue;
      }

      /* ODS-5 extended specification escape character */
      cptr++;
      switch (*cptr)
      {
         case '!' : case '#' : case '&' : case '\'' : case '`' :
         case '(' : case ')' : case '+' : case '@'  : case '{' :
         case '}' : case '.' : case ',' : case ';'  : case '[' :
         case ']' : case '%' : case '^' : case '='  : case ' ' :
            *sptr++ = *cptr++;
            break;
         case '_' :
            *sptr++ = ' ';
            cptr++;
            break;
         default :
            if (!isxdigit (*cptr))
            {
               *sptr++ = *cptr++;
               break;
            }
            ch = 0;
            if (*cptr >= '0' && *cptr <= '9')
               ch = (*cptr - (int)'0') << 4;
            else
            if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
               ch = (tolower(*cptr) - (int)'a' + 10) << 4;
            if (*cptr) cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               ch += (*cptr - (int)'0');
            else
            if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
               ch += (tolower(*cptr) - (int)'a' + 10);
            if (*cptr) cptr++;
            *sptr++ = ch;
      }
   }
   if (sptr < zptr) *sptr++ = '\0';
   if (sptr < zptr) *sptr++ = '\0';

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!&Z\n", PathPtr);

   if (sptr < zptr) return ((sptr - PathPtr) - 2);

   /* storage overflowed */
   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
   *(unsigned short*)PathPtr = '\0\0';
   return (0);
}

#endif /* ODS_EXTENDED */

/*****************************************************************************/
/*
Convert a VMS full file specification into a URL-style specification, example:
"DEVICE:[DIR1.DIR2]__$A__2A__$FILE.TXT" into "/device/dir1/dir2/A file.txt", or
a  URL-style specification, example "/device/dir1/dir2/A__20file.txt" into
"/device/dir1/dir2/A file.txt" (ensuring it's correctly converted and escaped). 
Expects file specification to be Advanced Server compliant.  Alphabetic '$'
case flags reset on a per-path-component basis.

Returns the length of the generated URL-style path.
*/ 
 
int MapUrl_AdsVmsToUrl
(
char *PathPtr,
char *VmsPtr,
int SizeOfPathPtr,
BOOL AbsorbMfd
)
{
   BOOL  isDevice,
         loCase;
   char  ch;
   char  *cptr, *eptr, *sptr, *zptr,
         *NamePtr,
         *PeriodPtr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_AdsVmsToUrl() !&B !&Z !UL",
                 AbsorbMfd, VmsPtr, SizeOfPathPtr);

   zptr = (sptr = PathPtr) + SizeOfPathPtr;

   /* look for indications it's a full file specification */
   cptr = VmsPtr;
   while (*cptr && *(unsigned short*)cptr != ':[' &&
          *cptr != ']' && *cptr != '/') cptr++;

   if (!*cptr || *cptr == '/')
   {
      /* no VMS device/directory components detected */
      cptr = VmsPtr;
   }
   else
   {
      /* copy device/directory components */
      isDevice = loCase = true;
      cptr = VmsPtr;
      if (sptr < zptr) *sptr++ = '/';
      while (*cptr && sptr < zptr)
      {
         if (*(unsigned short*)cptr == '::')
         {
            /* DECnet */
            *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = '/';
            loCase = true;
            continue;
         }

         if (*(unsigned short*)cptr == ':[')
         {
            cptr++;
            cptr++;
            /* remove any reference to a Master File Directory */
            if (AbsorbMfd && strncmp (cptr, "000000", 6) == 0)
               cptr += 6;
            else
               *sptr++ = '/';
            isDevice = false;
            loCase = true;
            continue;
         }

         if (*cptr == '.')
         {
            *sptr++ = '/';
            if (*cptr == '.' && *(unsigned short*)(cptr+1) == '..')
            {
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
            }
            else
               cptr++;
            loCase = true;
            continue;
         }

         if (*cptr == ']')
         {
            cptr++;
            *sptr++ = '/';
            break;
         }

         /* don't need to be concerned with periods in directories */
         if (*(unsigned short*)cptr != '__' || isDevice)
         {
            if (loCase)
               *sptr++ = tolower(*cptr++);
            else
               *sptr++ = toupper(*cptr++);
            continue;
         }

         /* convert from Advanced Server encoding */
         cptr++;
         cptr++;
         if (*cptr == '$')
         {
            cptr++;
            loCase = !loCase;
            continue;
         }
         ch = 0;
         if (*cptr >= '0' && *cptr <= '9')
            ch = (*cptr - (int)'0') << 4;
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            ch = (tolower(*cptr) - (int)'a' + 10) << 4;
         else
            ch = '?';
         if (*cptr) cptr++;   
         if (*cptr >= '0' && *cptr <= '9')
            ch += (*cptr - (int)'0');
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            ch += (tolower(*cptr) - (int)'a' + 10);
         else
            ch = '?';
         if (*cptr) cptr++;
         if (loCase)
            *sptr++ = tolower(ch);
         else
            *sptr++ = toupper(ch);
      }
   }

   /* copy the file component */
   PeriodPtr = NULL;
   NamePtr = cptr;
   loCase = true;
   while (*cptr && sptr < zptr)
   {
      if (*cptr != '_')
      {
         if (*cptr != '.')
         {
            if (loCase)
               *sptr++ = tolower(*cptr++);
            else
               *sptr++ = toupper(*cptr++);
            continue;
         }
         PeriodPtr = sptr;
         *sptr++ = *cptr++;
         continue;
      }

      /* need to cater for possible '__xx', '_._xx', '__.xx' */
      if (*(unsigned short*)cptr != '__' && *(unsigned short*)cptr != '_.')
      {
         /* an isolated underscore */
         if (sptr < zptr) *sptr++ = *cptr++;
         continue;
      }
      cptr++;
      if (*(unsigned short*)cptr == '._')
      {
         if (cptr != NamePtr + 39)
         {
            /* not an ODS-2 required period */
            if (sptr < zptr) *sptr++ = *cptr++;
            continue;
         }
         /* step over the ODS-2 required period */
         cptr++;
      }
      else
      if (*(unsigned short*)cptr == '_.')
      {
         /* after two underscores can only be an ODS-2 required period */
         cptr++;
      }
      cptr++;

      /* convert from Advanced Server encoding */
      if (*cptr == '$')
      {
         cptr++;
         loCase = !loCase;
         continue;
      }
      ch = 0;
      if (*cptr >= '0' && *cptr <= '9')
         ch = (*cptr - (int)'0') << 4;
      else
      if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
         ch = (tolower(*cptr) - (int)'a' + 10) << 4;
      else
         ch = '?';
      if (*cptr) cptr++;   
      /* if it's an ODS-2 required period just ignore it (i.e. '__x.x') */
      if (*cptr == '.') cptr++;
      if (*cptr >= '0' && *cptr <= '9')
         ch += (*cptr - (int)'0');
      else
      if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
         ch += (tolower(*cptr) - (int)'a' + 10);
      else
         ch = '?';
      if (*cptr) cptr++;
      if (loCase)
         *sptr++ = tolower(ch);
      else
         *sptr++ = toupper(ch);
   }

   if (sptr < zptr) *sptr++ = '\0';
   if (sptr < zptr) *sptr++ = '\0';

   if (PeriodPtr)
   {
      /* remove the ODS-2 required period if there is a trailing period */
      for (cptr = PeriodPtr+1; *cptr && *cptr != '.'; cptr++);
      if (*cptr)
      {
         for (cptr = (sptr = PeriodPtr) + 1; *cptr; *sptr++ = *cptr++);
         if (sptr < zptr) *sptr++ = '\0';
         if (sptr < zptr) *sptr++ = '\0';
      }
   }

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!&Z\n", PathPtr);

   if (sptr < zptr) return ((sptr - PathPtr) - 2);

   /* storage overflowed */
   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
   *(unsigned short*)PathPtr = '\0\0';
   return (0);
}

/*****************************************************************************/
/*
The Advanced Server (PATHWORKS 6) / Samba encoding maps the period as described
in the description of MapUrl_UrlToAdsVms().  This function decodes the file
anem plus type then find the last period in it and uses that following as the
file type.
*/ 
 
char* MapUrl_AdsFileType (char *FileType)

{
   static char  AdsFileType [96];

   int  Length;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_AdsFileType() !&Z", FileType);

   Length = MapUrl_AdsVmsToUrl (AdsFileType, FileType,
                                sizeof(AdsFileType), true);
   cptr = AdsFileType + Length;
   while (cptr > AdsFileType && *cptr != '.') cptr--;
   if (WATCH_MODULE(WATCH_MOD_MAPURL)) WatchDataFormatted ("!&Z\n", cptr);

   return (cptr);
}

/*****************************************************************************/
/*
Convert a VMS full file specification into a URL-style specification, example:
"DEVICE:[DIR1.DIR2]A$2AFILE.TXT" into "/device/dir1/dir2/a file.txt", or a
URL-style specification, example "/device/dir1/dir2/a$20file.txt" into
"/device/dir1/dir2/a file.txt" (ensuring it's correctly converted and escaped). 
Expects file specification to be PATHWORKS v4/5 compliant and forces the URL to
all  lower-case.

Returns the length of the generated URL-style path.
*/ 
 
int MapUrl_PwkVmsToUrl
(
char *PathPtr,
char *VmsPtr,
int SizeOfPathPtr,
BOOL AbsorbMfd
)
{
   BOOL  isDevice;
   char  ch;
   char  *cptr, *eptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_PwkVmsToUrl() !&B !&Z !UL",
                 AbsorbMfd, VmsPtr, SizeOfPathPtr);

   zptr = (sptr = PathPtr) + SizeOfPathPtr;

   /* look for indications it's a full file specification */
   cptr = VmsPtr;
   while (*cptr && *(unsigned short*)cptr != ':[' &&
          *cptr != ']' && *cptr != '/') cptr++;

   if (!*cptr || *cptr == '/')
   {
      /* no VMS device/directory components detected */
      cptr = VmsPtr;
   }
   else
   {
      /* copy device/directory components */
      isDevice = true;
      cptr = VmsPtr;
      if (sptr < zptr) *sptr++ = '/';
      while (*cptr && sptr < zptr)
      {
         if (*(unsigned short*)cptr == '::')
         {
            /* DECnet */
            *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = '/';
            continue;
         }

         if (*(unsigned short*)cptr == ':[')
         {
            cptr++;
            cptr++;
            /* remove any reference to a Master File Directory */
            if (AbsorbMfd && strncmp (cptr, "000000", 6) == 0)
               cptr += 6;
            else
               *sptr++ = '/';
            isDevice = false;
            continue;
         }

         if (*cptr == '.')
         {
            *sptr++ = '/';
            if (*cptr == '.' && *(unsigned short*)(cptr+1) == '..')
            {
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
            }
            else
               cptr++;
            continue;
         }

         if (*cptr == ']')
         {
            cptr++;
            *sptr++ = '/';
            break;
         }

         if (*cptr != '$' || isDevice)
         {
            *sptr++ = tolower(*cptr++);
            continue;
         }

         /* convert from PATHWORKS encoding */
         cptr++;
         ch = 0;
         if (*cptr >= '0' && *cptr <= '9')
            ch = (*cptr - (int)'0') << 4;
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            ch = (tolower(*cptr) - (int)'a' + 10) << 4;
         else
            ch = '?';
         if (*cptr) cptr++;   
         if (*cptr >= '0' && *cptr <= '9')
            ch += (*cptr - (int)'0');
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            ch += (tolower(*cptr) - (int)'a' + 10);
         else
            ch = '?';
         if (*cptr) cptr++;
         *sptr++ = ch;
      }
   }

   /* copy the file component */
   while (*cptr && sptr < zptr)
   {
      if (*cptr != '$')
      {
         *sptr++ = tolower(*cptr++);
         continue;
      }

      /* convert from PATHWORKS encoding */
      cptr++;
      ch = 0;
      if (*cptr >= '0' && *cptr <= '9')
         ch = (*cptr - (int)'0') << 4;
      else
      if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
         ch = (tolower(*cptr) - (int)'a' + 10) << 4;
      else
         ch = '?';
      if (*cptr) cptr++;   
      if (*cptr >= '0' && *cptr <= '9')
         ch += (*cptr - (int)'0');
      else
      if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
         ch += (tolower(*cptr) - (int)'a' + 10);
      else
         ch = '?';
      if (*cptr) cptr++;
      *sptr++ = ch;
   }
   if (sptr < zptr) *sptr++ = '\0';
   if (sptr < zptr) *sptr++ = '\0';

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!&Z\n", PathPtr);

   if (sptr < zptr) return ((sptr - PathPtr) - 2);

   /* storage overflowed */
   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
   *(unsigned short*)PathPtr = '\0\0';
   return (0);
}

/*****************************************************************************/
/*
Convert a VMS full file SRI (e.g. MultiNet NFS) compliant specification into a
URL-style specification, example:  "DEVICE:[DIR1.DIR2]A$5NFILE.TXT" into
"/device/dir1/dir2/a file.txt", or a URL-style specification, example
"/device/dir1/dir2/a$5nfile.txt" into "/device/dir1/dir2/a file.txt" (ensuring
it's correctly converted and escaped). Expects file specification to be SRI
compliant.  The SRI scheme is quite complex.  Refer to online documentation
associated with Process Software's MultiNet NFS server.  Alphabetic '$' case
flags reset on a per-path-component basis.

Returns the length of the generated URL-style path.
*/ 
 
int MapUrl_SriVmsToUrl
(
char *PathPtr,
char *VmsPtr,
int SizeOfPathPtr,
BOOL AbsorbMfd
)
{
   BOOL  isDevice,
         loCase;
   char  ch;
   char  *cptr, *eptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_SriVmsToUrl() !&B !&Z !UL",
                 AbsorbMfd, VmsPtr, SizeOfPathPtr);

   zptr = (sptr = PathPtr) + SizeOfPathPtr;

   /* look for indications it's a full file specification */
   cptr = VmsPtr;
   while (*cptr && *(unsigned short*)cptr != ':[' &&
          *cptr != ']' && *cptr != '/') cptr++;

   if (!*cptr || *cptr == '/')
   {
      /* no VMS device/directory components detected */
      cptr = VmsPtr;
   }
   else
   {
      /* copy device/directory components */
      isDevice = loCase = true;
      cptr = VmsPtr;
      if (sptr < zptr) *sptr++ = '/';
      while (*cptr && sptr < zptr)
      {
         if (*(unsigned short*)cptr == '::')
         {
            /* DECnet */
            *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = '/';
            loCase = true;
            continue;
         }

         if (*(unsigned short*)cptr == ':[')
         {
            cptr++;
            cptr++;
            /* remove any reference to a Master File Directory */
            if (AbsorbMfd && strncmp (cptr, "000000", 6) == 0)
               cptr += 6;
            else
               *sptr++ = '/';
            isDevice = false;
            loCase = true;
            continue;
         }

         if (*cptr == '.')
         {
            *sptr++ = '/';
            if (*cptr == '.' && *(unsigned short*)(cptr+1) == '..')
            {
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
               if (sptr < zptr) *sptr++ = *cptr++;
            }
            else
               cptr++;
            loCase = true;
            continue;
         }

         if (*cptr == ']')
         {
            *sptr++ = '/';
            cptr++;
            break;
         }

         if (isDevice)
            *sptr++ = tolower(*cptr++);
         else
         if (*cptr == '$')
            *sptr++ = MapUrl_SriDecode (&cptr, &loCase);
         else
         if (loCase)
            *sptr++ = tolower(*cptr++);
         else
            *sptr++ = toupper(*cptr++);
      }
   }

   /* copy the file component */
   loCase = true;
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '.') loCase = true;
      if (*cptr == '$')
         *sptr++ = MapUrl_SriDecode (&cptr, &loCase);
      else
      if (loCase)
         *sptr++ = tolower(*cptr++);
      else
         *sptr++ = toupper(*cptr++);
   }
   if (sptr < zptr) *sptr++ = '\0';
   if (sptr < zptr) *sptr++ = '\0';

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchDataFormatted ("!&Z\n", PathPtr);

   if (sptr < zptr) return ((sptr - PathPtr) - 2);

   /* storage overflowed */
   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "OVERFLOW");
   *(unsigned short*)PathPtr = '\0\0';
   return (0);
}

/*****************************************************************************/
/*
Return a single character that has been decoded from the string pointed to by
the storage supplied by 'stringPtrPtr'.  If the decode detects an error in the
encoding then it returns a question mark (crude, but hey it should be obvious
to the end-user!)  The SRI scheme is quite complex.   Refer to online 
documentation associated with Process Software's MultiNet NFS server.
*/ 
 
char MapUrl_SriDecode
(
char **stringPtrPtr,
BOOL *loCasePtr
)
{
   static char  SriDecodeTable [6][26] =
{
     /* table $4A-Z */
   { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,  /* A-H */
     0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10,  /* I-P */
     0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,  /* Q-X */
     0x19, 0x1a },                                    /* Y-Z */

     /* table $5A-Z */
   { 0x21, 0x22, 0x23, '?',  0x25, 0x26, 0x27, '?', 
     0x29, 0x2a, 0x2b, 0x2c, '?',  0x2e, 0x2f, '?', 
     '?',  '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  0x3a },

     /* table $6A-Z */
   { 0x00, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, '?',  '?', 
     0x29, '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?' },

     /* table $7A-Z */
   { 0x20, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, '?',  '?', 
     0x29, '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?' },

     /* table $8A-Z */
   { 0x40, 0x5b, 0x5c, 0x5d, 0x5e, '?',  '?',  '?', 
     0x29, '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?' },

     /* table $9A-Z */
   { 0x60, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, '?',  '?', 
     0x29, '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?',  '?',  '?',  '?',  '?',  '?',  '?', 
     '?',  '?' },
};

   char  ch, idx;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_SriDecode() !&B !4AZ", *loCasePtr, *stringPtrPtr);

   cptr = *stringPtrPtr;
   if (*cptr == '$')
   {
      cptr++;
      if (*cptr == '2' || *cptr == '3')
      {
         ch = (*cptr++ - '0') << 6;
         if (*cptr >= '0' && *cptr <= '7')
         {
            ch |= (*cptr++ - '0') << 3;
            if (*cptr >= '0' && *cptr <= '7')
               ch |= (*cptr++ -'0');
            else
            {
               ch = '?';
               cptr++;
            }
         }
         else
         {
            ch = '?';
            cptr++;
         }
      }
      else
      if (*cptr >= '4' && *cptr <= '9')
      {
         idx = *cptr++ - '4';
         if (*cptr >= 'A' && *cptr <= 'Z')
            ch = SriDecodeTable[idx][*(unsigned char*)cptr-'A'];
         else
            ch = '?';
         cptr++;
      }
      else
      if (*cptr >= 'A' && *cptr <= 'Z')
      {
         *loCasePtr = !*loCasePtr;
         if (*loCasePtr)
            ch = tolower(*cptr++);
         else
            ch = toupper(*cptr++);
      }
      else
      if (*cptr == '$')
         ch = *cptr++;
      else
      {
         ch = '?';
         cptr++;
      }
   }
   else
   if (*loCasePtr)
      ch = tolower(*cptr++);
   else
      ch = toupper(*cptr++);

   if (WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
      WatchDataFormatted ("{!UL}!-!#AZ 0x!2XL !-!&C\n",
                          cptr - *stringPtrPtr, *stringPtrPtr, ch);

   *stringPtrPtr = cptr;

   return (ch);
}

/*****************************************************************************/
/*
The SRI encoding maps the first period in any multi-period file type as a
period and any thereafter as '$5N' sequences.  Find any final '$5N' sequence
and map this into a "real" file type returning a pointer to this static
storage.  If there is no '$5N' in the string then just return a pointer to the
original string.
*/ 
 
char* MapUrl_SriFileType (char *FileType)

{
   static char  SriFileType [96];

   int  Length;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_SriFileType() !&Z", FileType);

   Length = MapUrl_SriVmsToUrl (SriFileType, FileType,
                                sizeof(SriFileType), true);
   cptr = SriFileType + Length;
   while (cptr > SriFileType && *cptr != '.') cptr--;
   if (WATCH_MODULE(WATCH_MOD_MAPURL)) WatchDataFormatted ("!&Z\n", cptr);

   return (cptr);
}

/*****************************************************************************/
/*
Given the current document's URL path and the specified document's URL path, 
generate a resultant URL path based on whether the specified document's path 
is absolute or relative the the current document's path.  Insufficient space to
contain the result path will not crash the function and the result will be set
to an empty string.
*/ 

int MapUrl_VirtualPath
(
char *CurrentPath,
char *DocumentPath,
char *ResultPath,
int SizeOfResultPath
)
{
   char  *cptr, *dptr, *sptr, *tptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_VirtualPath() !&Z !&Z !UL",
                 CurrentPath, DocumentPath, SizeOfResultPath);

   if (*(dptr = DocumentPath) == '/' || strstr (DocumentPath, "//")) 
   {
      /*****************/
      /* absolute path */
      /*****************/

      strcpy (ResultPath, dptr);
      zptr = (sptr = ResultPath) + SizeOfResultPath;
      while (*dptr && sptr < zptr) *sptr++ = *dptr++;
      if (sptr >= zptr) return (-1);
      *sptr = '\0';
      if (Debug) fprintf (stdout, "ResultPath |%s|\n", ResultPath);
      return (sptr - ResultPath);
   }

   /* step over any "relative to this directory" syntax ("./") */
   if (dptr[0] == '.' && dptr[1] == '/') dptr += 2;
   if (*dptr != '.')
   {
      /*****************/
      /* inferior path */
      /*****************/

      zptr = (sptr = tptr = ResultPath) + SizeOfResultPath;
      cptr = CurrentPath;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '/') tptr = sptr+1;
         *sptr++ = *cptr++;
      }
      sptr = tptr;
      while (*dptr && sptr < zptr) *sptr++ = *dptr++;
      if (sptr >= zptr) return (-1);
      *sptr = '\0';
      if (Debug) fprintf (stdout, "ResultPath |%s|\n", ResultPath);
      return (sptr - ResultPath);
   }

   /*****************/
   /* relative path */
   /*****************/

   zptr = (sptr = tptr = ResultPath) + SizeOfResultPath;
   cptr = CurrentPath;
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '/') tptr = sptr;
      *sptr++ = *cptr++;
   }
   sptr = tptr;
   /* loop, stepping back one level for each "../" in the document path */
   while (dptr[0] == '.' && dptr[1] == '.' && dptr[2] == '/')
   {
      dptr += 3;
      if (sptr > ResultPath) sptr--;
      while (sptr > ResultPath && *sptr != '/') sptr--;
   }
   if (sptr > ResultPath && sptr < zptr)
      sptr++;
   else
   if (sptr < zptr)
      *sptr++ = '/';
   while (*dptr && sptr < zptr) *sptr++ = *dptr++;
   if (sptr >= zptr) return (-1);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "ResultPath |%s|\n", ResultPath);
   return (sptr - ResultPath);
}

/*****************************************************************************/
/*
A server administration report on the server's mapping rules. This function
just wraps the reporting function, loading a temporary database if necessary
for reporting from the rule file.
*/

MapUrl_Report
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction,
BOOL UseServerDatabase,
char *VirtualService
)
{
   int  status;
   META_CONFIG  *mcptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && (WATCH_MODULE(WATCH_MOD_MAPURL)))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_Report() !&X !UL !&Z",
                 NextTaskFunction, UseServerDatabase, VirtualService);

   if (UseServerDatabase)
   {
      /* use mapping rules already loaded into the server */
      MapUrl_ReportNow (rqptr, true, MetaGlobalMappingPtr, VirtualService);
   }
   else
   {
      /* load temporary set of rules from mapping file */
      status = MapUrl_ConfigLoad (&mcptr);
      if (VMSnok (status))
      {
         /* severe error reported */
         rqptr->rqResponse.HttpStatus = 403;
         ErrorGeneral (rqptr, mcptr->LoadReport.TextPtr, FI_LI);
      }
      else
         MapUrl_ReportNow (rqptr, false, mcptr, VirtualService);
      MetaConUnload (&mcptr, NULL);
   }

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Return a report on the HTTPd server's mapping rules ... now!
*/

MapUrl_ReportNow
(
REQUEST_STRUCT *rqptr,
BOOL UseServerDatabase,
META_CONFIG *mcptr,
char *VirtualService
)
{
   static char  BeginPage [] =
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Mapping Rules</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=5 CELLSPACING=0 BORDER=0>\n\
<TR><TD><PRE>";

   static char  RuleFao [] = "!3ZL  !#* <B>!AZ</B>  !&;AZ!&@!&@!&@\n";
   static char  NonRuleFao [] = "!3ZL  !#* <B>!#AZ</B>  !&;AZ\n";
   static char  QuestionFao [] = "!3ZL  !#* <B>?</B>  <I>!&;AZ</I>\n";
   static char  RuleVirtualFao [] = "!3ZL  !#* <B>[[!&;AZ]]</B>\n";

   static char  UserMappingFao [] =
"!AZ\
</PRE></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>User Mapping Cache &nbsp;&nbsp\
<FONT SIZE=-1>(maximum !UL)</FONT></TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TH></TH>\
<TH ALIGN=left><U>Name</U></TH>\
<TH ALIGN=left><U>Path</U></TH>\
<TH ALIGN=left><U>ODS</U></TH>\
<TH ALIGN=right><U>Hits</U>&nbsp;</TH>\
<TH ALIGN=left><U>Last</U></TH>\
<TH ALIGN=left><U>Reuse</U></TH>\
</TR>\n";

   static char  EntryFao [] =
"<TR>\
<TH ALIGN=right>!UL.&nbsp;</TH>\
<TD ALIGN=left>!AZ&nbsp;</TD>\
<TD ALIGN=left>!AZ/&nbsp;</TD>\
<TD ALIGN=right>!AZ</TD>\
<TD ALIGN=right>!UL&nbsp;</TD>\
<TD ALIGN=left><FONT SIZE=-1>!20&W</FONT>&nbsp;</TD>\
<TD ALIGN=right>!UL</TD>\
</TR>\n";

   static char  EndPageFao [] =
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n";

   BOOL  ThisVirtualService;
   int  status,
        EntryCount;
   unsigned long  FaoVector [32];
   unsigned long  *vecptr;
   unsigned short  Length;
   char  *cptr;
   char  Buffer [2048],
         RmsSubChar [2],
         TimeString [32];
   LIST_ENTRY  *leptr;
   MAP_RULE_META  *mrptr;
   MAP_URL_USER_ENTRY  *ucptr;
   METACON_LINE  *mclptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL, "MapUrl_ReportNow() !&B !&Z",
                 UseServerDatabase, VirtualService);

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_MAPPING_RULES;
   RESPONSE_HEADER_200_HTML (rqptr);
   AdminPageTitle (rqptr, "Path Mapping");
   AdminMetaConReport (rqptr, mcptr, MetaGlobalMappingPtr);
   AdminMetaConSource (rqptr, mcptr, MetaGlobalMappingPtr, mcptr->IncludeFile);
   AdminVirtualServiceForm (rqptr, ADMIN_REPORT_MAPPING,
                            VirtualService, UseServerDatabase);

   /*****************/
   /* mapping rules */
   /*****************/

   status = NetWriteFaol (rqptr, BeginPage, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   status = SS$_NORMAL;
   EntryCount = 0;
   ThisVirtualService = true;

   MetaConParseReset (mcptr, true);
   while (mclptr = MetaConParseRaw (mcptr))
   {
      if (mclptr->Token == METACON_TOKEN_SERVICE)
      {
         /* if filtering on a virtual service */
         if (VirtualService[0] &&
             *(unsigned long*)mclptr->TextPtr != '*:*\0' &&
             !StringMatch (mclptr->TextPtr, VirtualService, false))
         {
            ThisVirtualService = false;
            continue;
         }
         ThisVirtualService = true;
         vecptr = FaoVector;
         *vecptr++ = mclptr->Number;
         *vecptr++ = mclptr->FlowControlLevel * 2;
         *vecptr++ = mclptr->TextPtr;
         status = NetWriteFaol (rqptr, RuleVirtualFao, &FaoVector);
         if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
         continue;
      }

      if (!ThisVirtualService) continue;
      EntryCount++;

      if (mclptr->Token != METACON_TOKEN_TEXT)
      {
         for (cptr = mclptr->TextPtr; *cptr && isalnum(*cptr); cptr++);
         vecptr = FaoVector;
         *vecptr++ = mclptr->Number;
         switch (mclptr->Token)
         {
            case METACON_TOKEN_ELIF :
            case METACON_TOKEN_ELSE :
            case METACON_TOKEN_UNIF :
            case METACON_TOKEN_IFIF :
               if (mclptr->FlowControlLevel)
                  *vecptr++ = (mclptr->FlowControlLevel - 1) * 2;
               else
                  *vecptr++ = mclptr->FlowControlLevel;
               break;
            default:
               *vecptr++ = mclptr->FlowControlLevel * 2;
         }
         *vecptr++ = cptr - mclptr->TextPtr;
         *vecptr++ = mclptr->TextPtr;
         while (*cptr && ISLWS(*cptr)) cptr++;
         *vecptr++ = cptr;
         status = NetWriteFaol (rqptr, NonRuleFao, &FaoVector);
         if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
         continue;
      }

      vecptr = FaoVector;
      *vecptr++ = mclptr->Number;
      *vecptr++ = mclptr->FlowControlLevel * 2;

      if (!(mrptr = mclptr->LineDataPtr))
      {
         *vecptr++ = mclptr->TextPtr;
         status = NetWriteFaol (rqptr, QuestionFao, &FaoVector);
         if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
         continue;
      }

      if (WATCHING(rqptr) &&
          WATCH_MODULE(WATCH_MOD_MAPURL) && WATCH_MODULE(WATCH_MOD__DETAIL))
         WatchDataFormatted ("!&X !UL !&Z !&Z !&Z\n",
            mrptr, mrptr->RuleType, mrptr->TemplatePtr,
            mrptr->ResultPtr, mrptr->ConditionalPtr);

      if (mrptr->RuleType == MAPURL_RULE_EXEC)
      {
         if (mrptr->IsCgiPlusScript)
            *vecptr++ = "exec+";
         else
            *vecptr++ = "exec";
      }
      else
      if (mrptr->RuleType == MAPURL_RULE_FAIL)
         *vecptr++ = "fail";
      else
      if (mrptr->RuleType == MAPURL_RULE_MAP)
         *vecptr++ = "map";
      else
      if (mrptr->RuleType == MAPURL_RULE_PASS)
         *vecptr++ = "pass";
      else
      if (mrptr->RuleType == MAPURL_RULE_REDIRECT)
         *vecptr++ = "redirect";
      else
      if (mrptr->RuleType == MAPURL_RULE_SCRIPT)
      {
         if (mrptr->IsCgiPlusScript)
            *vecptr++ = "script+";
         else
            *vecptr++ = "script";
      }
      else
      if (mrptr->RuleType == MAPURL_RULE_SET)
         *vecptr++ = "set";
      else
      if (mrptr->RuleType == MAPURL_RULE_UXEC)
         *vecptr++ = "uxec";
      else
      if (mrptr->RuleType == MAPURL_RULE_USER)
         *vecptr++ = "user";
      else
         *vecptr++ = "*ERROR*";

      *vecptr++ = mrptr->TemplatePtr;
      if (isdigit(*mrptr->ResultPtr))
      {
         *vecptr++ = "  &quot;!&;AZ&quot;";
         *vecptr++ = mrptr->ResultPtr;
      }
      else
      if (mrptr->ResultPtr[0])
      {
         *vecptr++ = "  !&;AZ";
         *vecptr++ = mrptr->ResultPtr;
      }
      else
         *vecptr++ = "";

      if (mrptr->PathSet)
      {
         /* just one space, ExplainPathSet() adds it's own leading space */
         *vecptr++ = " !&;AZ";
         *vecptr++ = MapUrl_ExplainPathSet (mrptr);
      }
      else
         *vecptr++ = "";

      if (mrptr->ConditionalPtr[0])
      {
         *vecptr++ = "  !&;AZ";
         *vecptr++ = mrptr->ConditionalPtr;
      }
      else
         *vecptr++ = "";
      status = NetWriteFaol (rqptr, RuleFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   vecptr = FaoVector;
   if (EntryCount)
      *vecptr++ = "";
   else
      *vecptr++ = "<I>(none)</I>\n";
   *vecptr++ = MapUrlUserNameCacheEntries;

   status = NetWriteFaol (rqptr, UserMappingFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   /*****************/
   /* user mappings */
   /*****************/

   EntryCount = 0;

   /* process the cache entry list from most to least recent */
   for (leptr = MapUrlUserNameCacheList.HeadPtr;
        leptr;
        leptr = leptr->NextPtr)
   {
      ucptr = (MAP_URL_USER_ENTRY*)leptr;

      /* if empty name, must have been reset, no point in going any further */
      if (!ucptr->UserName[0]) break;

      vecptr = FaoVector;
      *vecptr++ = ++EntryCount;
      *vecptr++ = ucptr->UserName;
      *vecptr++ = ucptr->UserPath;

      if (ucptr->PathOds == MAPURL_PATH_ODS_2)
         *vecptr++ = "ODS-2";
      else
      if (ucptr->PathOds == MAPURL_PATH_ODS_5)
         *vecptr++ = "ODS-5";
      else
         *vecptr++ = "ods-2";

      *vecptr++ = ucptr->HitCount;
      *vecptr++ = &ucptr->LastBinaryTime;
      *vecptr++ = ucptr->ReuseCount;

      status = NetWriteFaol (rqptr, EntryFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   vecptr = FaoVector;
   if (!EntryCount)
   {
      static char  NotNone [] =
"<TR><TD ALIGN=center COLSPAN=6><I>(none)</I></TD></TR>\n";

      status = NetWriteFaol (rqptr, NotNone, NULL);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   status = NetWriteFaol (rqptr, EndPageFao, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
}

/*****************************************************************************/
/*
Report an individual rule during processing of a mapping rule check.
*/

MapUrl_WatchRule
(
REQUEST_STRUCT *rqptr,
MAP_RULE_META  *mrptr,
char *PathPtr,
int PathOds,
int Match
)
{
   static char  RuleFao [] = "!3ZL !AZ!AZ  !AZ  !8AZ  !AZ  !&@!&@!&@\n",
                RuleVirtualFao [] = "!3ZL !AZ!AZ  !AZ  !8AZ  [[!AZ]]\n";

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   unsigned long  *vecptr;
   char  Buffer [2048],
         RmsSubChar [2];

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && (WATCH_MODULE(WATCH_MOD_MAPURL)))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_WatchRule() !&Z !UL !UL", PathPtr, PathOds, Match);

   vecptr = FaoVector;
   *vecptr++ = mrptr->MetaConNumber;
   *vecptr++ = PathPtr;

   if (PathOds == MAPURL_PATH_ODS_0)
      *vecptr++ = "  NOODS";
   else
   if (PathOds == MAPURL_PATH_ODS_2)
      *vecptr++ = "  ODS=2";
   else
   if (PathOds == MAPURL_PATH_ODS_5)
      *vecptr++ = "  ODS=5";
   else
   if (PathOds == MAPURL_PATH_ODS_PWK)
      *vecptr++ = "  ODS=PWK";
   else
   if (PathOds == MAPURL_PATH_ODS_SMB)
      *vecptr++ = "  ODS=SMB";
   else
   if (PathOds == MAPURL_PATH_ODS_SRI)
      *vecptr++ = "  ODS=SRI";
   else
      *vecptr++ = "";
 
   if (Match == MAPURL_REPORT_MATCH_NOT)
      *vecptr++ = "..";
   else
   if (Match == MAPURL_REPORT_MATCH_RULE)
      *vecptr++ = "YN";
   else
   if (Match == MAPURL_REPORT_MATCH_RULECOND)
      *vecptr++ = "YY";
   else
   if (Match ==  MAPURL_REPORT_MATCH_RULENOCOND)
      *vecptr++ = "Y-";
   else
      *vecptr++ = "??";

   switch (mrptr->RuleType)
   {
      case MAPURL_RULE_EXEC :
         if (mrptr->IsCgiPlusScript)
            *vecptr++ = "EXEC+";
         else
            *vecptr++ = "EXEC";
         break;
      case MAPURL_RULE_FAIL :
         *vecptr++ = "FAIL";
         break;
      case MAPURL_RULE_MAP :
         *vecptr++ = "MAP";
         break;
      case MAPURL_RULE_PASS :     
         *vecptr++ = "PASS";
         break;
      case MAPURL_RULE_REDIRECT :
         *vecptr++ = "REDIRECT";
         break;
      case MAPURL_RULE_SCRIPT :
         if (mrptr->IsCgiPlusScript)
            *vecptr++ = "SCRIPT+";
         else
            *vecptr++ = "SCRIPT";
         break;
      case MAPURL_RULE_SET :
         *vecptr++ = "SET";
         break;
      case MAPURL_RULE_USER :
         *vecptr++ = "USER";
         break;
      case MAPURL_RULE_UXEC :
         if (mrptr->IsCgiPlusScript)
            *vecptr++ = "UXEC+";
         else
            *vecptr++ = "UXEC";
         break;
      default :
         *vecptr++ = "*ERROR*";
   }

   *vecptr++ = mrptr->TemplatePtr;

   if (isdigit(*mrptr->ResultPtr))
   {
      *vecptr++ = "\"!AZ\"";
      *vecptr++ = mrptr->ResultPtr;
   }
   else
   {
      *vecptr++ = "!AZ";
      *vecptr++ = mrptr->ResultPtr;
   }

   if (mrptr->PathSet)
   {
      *vecptr++ = " !AZ";
      *vecptr++ = MapUrl_ExplainPathSet (mrptr);
   }
   else
      *vecptr++ = "";

   if (mrptr->ConditionalPtr[0])
   {
      *vecptr++ = " !AZ";
      *vecptr++ = mrptr->ConditionalPtr;
   }
   else
      *vecptr++ = "";

   status = WriteFaol (Buffer, sizeof(Buffer), &Length, RuleFao, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorNoticed (status, "WriteFaol()", FI_LI);

   if (WATCHING(rqptr) && (WATCH_CATEGORY(WATCH_MAPPING)))
      WatchData (Buffer, Length); 
}

/*****************************************************************************/
/*
Display mapping rules used to resolve the path in the query string.  Called
from the Request.c modules.  For HTTP report, uses blocking network write.
*/

char* MapUrl_ExplainPathSet (MAP_RULE_META *mrptr)

{
   static char  SetFao [] =
"!&@!AZ!AZ!AZ!&@!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ\
!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@!&@";
   static char  Buffer [256];

   int  status;
   unsigned long  FaoVector [128];
   unsigned long  *vecptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL, "MapUrl_ExplainPathSet()");

   vecptr = FaoVector;

   if (mrptr->mpPathSet.NoAcceptLang)
      *vecptr++ = " NOaccept=LANG";
   else
   if (mrptr->mpPathSet.AcceptLangChar)
   {
      *vecptr++ =
" accept=LANG=(default=!AZ,char=!&C,!&?type\rname\r)";
      if (mrptr->mpPathSet.AcceptLangPtr)
         *vecptr++ = mrptr->mpPathSet.AcceptLangPtr;
      else
         *vecptr++ = "";
      *vecptr++ = mrptr->mpPathSet.AcceptLangChar;
      *vecptr++ = mrptr->mpPathSet.AcceptLangTypeVariant;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoAlert)
      *vecptr++ = " NOALERT";
   else
   if (mrptr->mpPathSet.Alert == MAPURL_PATH_ALERT_END)
      *vecptr++ = " ALERT=end";
   else
   if (mrptr->mpPathSet.Alert == MAPURL_PATH_ALERT_AUTH)
      *vecptr++ = " ALERT=auth";
   else
   if (mrptr->mpPathSet.Alert == MAPURL_PATH_ALERT_MAP)
      *vecptr++ = " ALERT=map";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoAuthMapped)
      *vecptr++ = " NOauth=MAPPED";
   else
   if (mrptr->mpPathSet.AuthMapped)
      *vecptr++ = " auth=MAPPED";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoAuthOnce)
      *vecptr++ = " NOauth=ONCE";
   else
   if (mrptr->mpPathSet.AuthOnce)
      *vecptr++ = " auth=ONCE";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.AuthRevalidateTimeout)
   {
      *vecptr++ = " auth=REVALIDATE=!2ZL:!2ZL:!2ZL";
      *vecptr++ = mrptr->mpPathSet.AuthRevalidateTimeout / 3600;
      *vecptr++ = (mrptr->mpPathSet.AuthRevalidateTimeout % 3600) / 60;
      *vecptr++ = (mrptr->mpPathSet.AuthRevalidateTimeout % 3600) % 60;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoCache)
      *vecptr++ = " NOcache";
   else
   if (mrptr->mpPathSet.Cache)
      *vecptr++ = " cache";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.DirNoAccess)
      *vecptr++ = " dir=NOaccess";
   else
   if (mrptr->mpPathSet.DirAccess)
      *vecptr++ = " dir=access";
   else
   if (mrptr->mpPathSet.DirAccessSelective)
      *vecptr++ = " dir=access=selective";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.DirNoImpliedWildcard)
      *vecptr++ = " dir=NOimpliedwildcard";
   else
   if (mrptr->mpPathSet.DirImpliedWildcard)
      *vecptr++ = " dir=impliedwildcard";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.DirNoWildcard)
      *vecptr++ = " dir=NOwildcard";
   else
   if (mrptr->mpPathSet.DirWildcard)
      *vecptr++ = " dir=wildcard";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoExpired)
      *vecptr++ = " NOexpired";
   else
   if (mrptr->mpPathSet.Expired)
      *vecptr++ = " expired";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoLog)
      *vecptr++ = " NOlog";
   else
   if (mrptr->mpPathSet.Log)
      *vecptr++ = " log";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoMapEllipsis)
      *vecptr++ = " map=NOELLIPSIS";
   else
   if (mrptr->mpPathSet.MapEllipsis)
      *vecptr++ = " map=ELLIPSIS";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.MapNonEmpty)
      *vecptr++ = " map=NONEMPTY";
   else
   if (mrptr->mpPathSet.MapEmpty)
      *vecptr++ = " map=EMPTY";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoMapOnce)
      *vecptr++ = " NOmap=ONCE";
   else
   if (mrptr->mpPathSet.MapOnce)
      *vecptr++ = " map=ONCE";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_0)
      *vecptr++ = " NOods";
   else
   if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_2)
      *vecptr++ = " ods=2";
   else
   if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_5)
      *vecptr++ = " ods=5";
   else
   if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_PWK)
      *vecptr++ = " ods=PWK";
   else
   if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_SMB)
      *vecptr++ = " ods=SMB";
   else
   if (mrptr->mpPathSet.OdsNaming == MAPURL_PATH_ODS_SRI)
      *vecptr++ = " ods=SRI";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.ScriptNoFind)
      *vecptr++ = " NOscript=FIND";
   else
   if (mrptr->mpPathSet.ScriptFind)
      *vecptr++ = " script=FIND";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoPrivSsi)
      *vecptr++ = " NOssi=PRIV";
   else
   if (mrptr->mpPathSet.PrivSsi)
      *vecptr++ = " ssi=PRIV";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoProfile)
      *vecptr++ = " NOprofile";
   else
   if (mrptr->mpPathSet.Profile)
      *vecptr++ = " profile";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.SSLCGIvar == SESOLA_CGI_VAR_APACHE_MOD_SSL)
      *vecptr++ = " SSLCGI=Apache_mod_SSL";
   else
   if (mrptr->mpPathSet.SSLCGIvar == SESOLA_CGI_VAR_PURVEYOR)
      *vecptr++ = " SSLCGI=Purveyor";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.NoStmLF)
      *vecptr++ = " NOstmLF";
   else
   if (mrptr->mpPathSet.StmLF)
      *vecptr++ = " stmLF";
   else
      *vecptr++ = "";

   if (!mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " auth=SYSUAF=pwdexpURL=!AZ";
      *vecptr++ = mrptr->mpPathSet.AuthSysUafPwdExpUrlPtr;
   }

   if (!mrptr->mpPathSet.CgiPrefixPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " CGIprefix=!AZ";
      *vecptr++ = mrptr->mpPathSet.CgiPrefixPtr;
   }

   if (!mrptr->mpPathSet.CharsetPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " charset=!AZ";
      *vecptr++ = mrptr->mpPathSet.CharsetPtr;
   }

   if (!mrptr->mpPathSet.DirCharsetPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " dir=charset=!AZ";
      *vecptr++ = mrptr->mpPathSet.DirCharsetPtr;
   }

   if (!mrptr->mpPathSet.ContentTypePtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " content=!AZ";
      *vecptr++ = mrptr->mpPathSet.ContentTypePtr;
   }

   if (!mrptr->mpPathSet.HttpAcceptCharsetPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " http=Accept-Charset=!AZ";
      *vecptr++ = mrptr->mpPathSet.HttpAcceptCharsetPtr;
   }

   if (!mrptr->mpPathSet.HttpAcceptLangPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " http=Accept-Language=!AZ";
      *vecptr++ = mrptr->mpPathSet.HttpAcceptLangPtr;
   }

   if (!mrptr->mpPathSet.IndexPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " index=!AZ";
      *vecptr++ = mrptr->mpPathSet.IndexPtr;
   }

   if (mrptr->mpPathSet.ReportBasic)
      *vecptr++ = " report=BASIC";
   else
   if (mrptr->mpPathSet.ReportDetailed)
      *vecptr++ = " report=DETAILED";
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.Report400as)
   {
      *vecptr++ = " report=400=!UL";
      *vecptr++ = mrptr->mpPathSet.Report400as;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.Report403as)
   {
      *vecptr++ = " report=403=!UL";
      *vecptr++ = mrptr->mpPathSet.Report403as;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.Report404as)
   {
      *vecptr++ = " report=404=!UL";
      *vecptr++ = mrptr->mpPathSet.Report404as;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.TimeoutKeepAlive ||
       mrptr->mpPathSet.TimeoutNoProgress ||
       mrptr->mpPathSet.TimeoutOutput)
   {
      *vecptr++ =
" timeout=!&@(k),!2ZL:!2ZL:!2ZL(np),!2ZL:!2ZL:!2ZL(o)";
      if (mrptr->mpPathSet.TimeoutKeepAlive < 0)
         *vecptr++ = "none";
      else
      {
         *vecptr++ = "!2ZL:!2ZL:!2ZL";
         *vecptr++ = mrptr->mpPathSet.TimeoutKeepAlive / 3600;
         *vecptr++ = (mrptr->mpPathSet.TimeoutKeepAlive % 3600) / 60;
         *vecptr++ = (mrptr->mpPathSet.TimeoutKeepAlive % 3600) % 60;
      }
      *vecptr++ = mrptr->mpPathSet.TimeoutNoProgress / 3600;
      *vecptr++ = (mrptr->mpPathSet.TimeoutNoProgress % 3600) / 60;
      *vecptr++ = (mrptr->mpPathSet.TimeoutNoProgress % 3600) % 60;
      *vecptr++ = mrptr->mpPathSet.TimeoutOutput / 3600;
      *vecptr++ = (mrptr->mpPathSet.TimeoutOutput % 3600) / 60;
      *vecptr++ = (mrptr->mpPathSet.TimeoutOutput % 3600) % 60;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.ThrottleFrom)
   {
      *vecptr++ =
" throttle=!UL,!UL,!UL,!UL,!2ZL:!2ZL:!2ZL,!2ZL:!2ZL:!2ZL";
      *vecptr++ = mrptr->mpPathSet.ThrottleFrom;
      *vecptr++ = mrptr->mpPathSet.ThrottleTo;
      *vecptr++ = mrptr->mpPathSet.ThrottleResume;
      *vecptr++ = mrptr->mpPathSet.ThrottleBusy;
      *vecptr++ = mrptr->mpPathSet.ThrottleTimeoutQueue / 3600;
      *vecptr++ = (mrptr->mpPathSet.ThrottleTimeoutQueue % 3600) / 60;
      *vecptr++ = (mrptr->mpPathSet.ThrottleTimeoutQueue % 3600) % 60;
      *vecptr++ = mrptr->mpPathSet.ThrottleTimeoutBusy / 3600;
      *vecptr++ = (mrptr->mpPathSet.ThrottleTimeoutBusy % 3600) / 60;
      *vecptr++ = (mrptr->mpPathSet.ThrottleTimeoutBusy % 3600) % 60;
   }
   else
      *vecptr++ = "";

   if (!mrptr->mpPathSet.ProxyBindIpAddressPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " proxy=BIND=!AZ";
      *vecptr++ = mrptr->mpPathSet.ProxyBindIpAddressPtr;
   }

   if (!mrptr->mpPathSet.ProxyChainHostPortPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " proxy=CHAIN=!AZ";
      *vecptr++ = mrptr->mpPathSet.ProxyChainHostPortPtr;
   }

   if (!mrptr->mpPathSet.ScriptAsPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " script=AS=!AZ";
      *vecptr++ = mrptr->mpPathSet.ScriptAsPtr;
   }

   if (mrptr->mpPathSet.ScriptBitBucketTimeout)
   {
      *vecptr++ = " script=BIT-BUCKET=!2ZL:!2ZL:!2ZL";
      *vecptr++ = mrptr->mpPathSet.ScriptBitBucketTimeout / 3600;
      *vecptr++ = (mrptr->mpPathSet.ScriptBitBucketTimeout % 3600) / 60;
      *vecptr++ = (mrptr->mpPathSet.ScriptBitBucketTimeout % 3600) % 60;
   }
   else
      *vecptr++ = "";

   if (mrptr->mpPathSet.ScriptCpuMax)
   {
      *vecptr++ = " script=CPU=!UL";
      *vecptr++ = mrptr->mpPathSet.ScriptCpuMax;
   }
   else
      *vecptr++ = "";

   if (!mrptr->mpPathSet.ScriptParamsPtr)
      *vecptr++ = "";
   else
   {
      *vecptr++ = " script=PARAMS=!AZ";
      *vecptr++ = mrptr->mpPathSet.ScriptParamsPtr;
   }

   if (mrptr->mpPathSet.RmsSubChar)
   {
      *vecptr++ = " RMSchar=!&C";
      *vecptr++ = mrptr->mpPathSet.RmsSubChar;
   }
   else
      *vecptr++ = "";

   status = WriteFaol (Buffer, sizeof(Buffer), NULL, SetFao, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      ErrorNoticed (status, "WriteFaol()", FI_LI);
      return ("*ERROR*");
   }
   return (Buffer);
}

/*****************************************************************************/
/*
Supplied with an integer in 'ThrottleIndex' representing the element number of
the 'ThrottleArray' path, scan through all the paths in the rule list
noting all SET THROTTLE= paths until the count equals that of the integer
supplied.  This is the rule representing that index.  Return a pointer to the
rule, or NULL.
*/ 
 
MAP_RULE_META* MapUrl_ThrottleRule (int ThrottleIndex)

{
   int  RuleCount;
   MAP_RULE_META  *mrptr;
   METACON_LINE  *mclptr;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_ThrottlePath() !UL", ThrottleIndex);

   RuleCount = 0;
   MetaConParseReset (MetaGlobalMappingPtr, true);
   while (mclptr = MetaConParseRaw (MetaGlobalMappingPtr))
   {
      if (!(mrptr = mclptr->LineDataPtr)) continue;
      if (!mrptr->mpPathSet.ThrottleFrom) continue;
      if (ThrottleIndex != RuleCount++) continue;
      return (mrptr);
   }
   return (NULL);
}

/*****************************************************************************/
/*
Reload the mapping rules.
*/

MapUrl_ControlReload
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction
)
{
   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && (WATCH_MODULE(WATCH_MOD_MAPURL)))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_ControlReload() !&X", NextTaskFunction);

   MetaConUnload (&MetaGlobalMappingPtr, NULL);

   MapUrl_ConfigLoad (&MetaGlobalMappingPtr);

   /* purge DCL module script task list and script name cache */
   DclLoadedMappingRules();

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;

   if (MetaGlobalMappingPtr->LoadReport.Count)
   {
      ReportSuccess (rqptr,
"$Server !AZ mapping rules reloaded.\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!%s During Load</FONT></TH></TR>\n\
<TR><TD><PRE>!&;AZ</PRE></TD></TR>\n\
</TABLE>\n",
         ServerHostPort, MetaGlobalMappingPtr->LoadReport.Count,
         MetaGlobalMappingPtr->LoadReport.TextPtr);
   }
   else
      ReportSuccess (rqptr, "$Server !AZ mapping rules reloaded.",
                     ServerHostPort);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Scan the conditional string evaluating the conditions!  Return true or false.
Anything it cannot understand it ignores!
*/

BOOL MapUrl_Conditional
(
REQUEST_STRUCT *rqptr,
MAP_RULE_META *mrptr
)
{
   BOOL  NegateThisCondition,
         NegateEntireConditional,
         Result,
         SoFarSoGood,
         WatchThisOne;
   int  ConditionalCount;
   char  *cptr, *sptr, *zptr,
         *CurrentPtr;
   char  Scratch [2048];
   NET_MASK  NetMask;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && (WATCH_MODULE(WATCH_MOD_MAPURL)))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_Conditional() !&Z", mrptr->ConditionalPtr);

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_MAPPING))
      WatchThisOne = true;
   else
      WatchThisOne = false;

   CurrentPtr = NULL;
   ConditionalCount = 0;
   NegateEntireConditional = NegateThisCondition = SoFarSoGood = false;
   cptr = mrptr->ConditionalPtr;
   if (WATCH_CAT && WatchThisOne)
      WatchDataFormatted ("conditional !AZ\n", cptr);
   while (*cptr)
   {
      while (ISLWS(*cptr)) cptr++;
      if (!*cptr) break;

      if (*cptr == '[' || *(unsigned short*)cptr == '![')
      {
         if (*cptr == '!')
         {
            NegateEntireConditional = true;
            cptr++;
         }
         else
            NegateEntireConditional = false;
         cptr++;
         ConditionalCount = 0;
         SoFarSoGood = false;
         continue;
      }
      if (*cptr == ']')
      {
         cptr++;
         if (NegateEntireConditional)
         {
            SoFarSoGood = !SoFarSoGood;
            NegateEntireConditional = false;
         }
         if (ConditionalCount && !SoFarSoGood)
         {
            cptr = "";
            break;
         }
         continue;
      }
      if (SoFarSoGood)
      {
         if (NegateEntireConditional)
         {
            SoFarSoGood = !SoFarSoGood;
            NegateEntireConditional = false;
         }
         /* at least one OK, skip to the end of the conditional */
         while (*cptr && *cptr != ']') cptr++;
         /* remember, there may be more than one per line */
         if (*cptr) continue;
         break;
      }

      CurrentPtr = cptr;
      NegateThisCondition = Result = false;
      zptr = (sptr = Scratch) + sizeof(Scratch)-1;

      if (*cptr == '!')
      {
         cptr++;
         NegateThisCondition = true;
      }

      switch (*(unsigned short*)cptr)
      {
         case 'AC' :

            /*************/
            /* "Accept:" */
            /*************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            if (rqptr->rqHeader.AcceptPtr)
               Result = MapUrl_ConditionalList (rqptr, Scratch,
                                                rqptr->rqHeader.AcceptPtr);
            break;
             
         case 'AL' :

            /**********************/
            /* "Accept-Language:" */
            /**********************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = MapUrl_ConditionalList (rqptr, Scratch,
                                             rqptr->rqHeader.AcceptLangPtr);
            break;

         case 'AS' :

            /*********************/
            /* "Accept-Charset:" */
            /*********************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = MapUrl_ConditionalList (rqptr, Scratch,
                                             rqptr->rqHeader.AcceptCharsetPtr);
            break;

         case 'CK' :

            /**********/
            /* cookie */
            /**********/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.CookiePtr, Scratch, true);
            break;

         case 'EX' :

            /**********************/
            /* extended file path */
            /**********************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']') cptr++;
            Result = rqptr->PathOdsExtended;
            break;

         case 'HM' :

            /****************************/
            /* client host network mask */
            /****************************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            /* if there's a problem with the mask then just fail it */
            if (!ParseNetMask (Scratch, &NetMask)) return (false); 
            Result = (htonl(NetMask.BitsNetwork) & NetMask.BitsMask) ==
                     (ntohl(rqptr->rqClient.IpAddress) & NetMask.BitsMask);
            break;

         case 'HO' :

            /****************************/
            /* client host name/address */
            /****************************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            for (sptr = Scratch; *sptr && !isalnum(*sptr); sptr++);
            if (isdigit(*sptr))
               sptr = rqptr->rqClient.IpAddressString;
            else
               sptr = rqptr->rqClient.HostName;
            Result = StringMatch (sptr, Scratch, true);
            break;

         case 'FO' :

            /*******************/
            /* proxy forwarded */
            /*******************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            /* it's unlikely we'll have any of this header field */
            if (rqptr->rqHeader.ForwardedPtr)
               Result = MapUrl_ConditionalList (rqptr, Scratch,
                                                rqptr->rqHeader.ForwardedPtr);
            break;

         case 'ME' :

            /***************/
            /* HTTP method */
            /***************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.MethodName, Scratch, true);
            break;

         case 'PI' :

            /*************/
            /* path-info */
            /*************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.PathInfoPtr, Scratch, true);
            break;

         case 'QS' :

            /****************/
            /* query string */
            /****************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.QueryStringPtr, Scratch, true);
            break;

         case 'RF' :

            /****************/
            /* refering URL */
            /****************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.RefererPtr, Scratch, true);
            break;

         case 'RU' :

            /***************/
            /* request URI */
            /***************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.RequestUriPtr, Scratch, true);
            break;

         case 'SC' :

            /****************************************/
            /* request scheme ("http:" or "https:") */
            /****************************************/

            ConditionalCount++;
            cptr += 3;
            /* also ignore any trailing colon on the scheme string */
            while (*cptr && !ISLWS(*cptr) && *cptr != ':' && *cptr != ']' &&
                   sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            if (*cptr == ':') cptr++;
            /* any trailing colon has been stripped */
            if (strsame (Scratch, "https", -1) &&
                rqptr->ServicePtr->RequestScheme == SCHEME_HTTPS)
               Result = true;
            else
            if (strsame (Scratch, "http", -1) &&
                rqptr->ServicePtr->RequestScheme == SCHEME_HTTP)
               Result = true;
            else
               Result = false;

            break;

         case 'SN' :

            /***************/
            /* server name */
            /***************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->ServicePtr->ServerHostName, Scratch, true);
            break;

         case 'SP' :

            /***************/
            /* server port */
            /***************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->ServicePtr->ServerPortString, Scratch, true);
            break;

         case 'UA' :

            /*****************/
            /* "User-Agent:" */
            /*****************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.UserAgentPtr, Scratch, true);
            break;

         case 'UD' :

            /****************/
            /* user details */
            /****************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqAuth.UserDetailsPtr, Scratch, true);
            break;

         case 'VS' :
         case 'HH' :  /* backward compatibility */

            /******************************************/
            /* virtual service ("Host:") name/address */
            /******************************************/

            ConditionalCount++;
            cptr += 3;
            while (*cptr && !ISLWS(*cptr) && *cptr != ']' && sptr < zptr)
               *sptr++ = *cptr++;
            *sptr = '\0';
            Result = StringMatch (rqptr->rqHeader.HostPtr, Scratch, true);
            break;

         default :

            /*******************/
            /* unknown, ignore */
            /*******************/

            /* should never occur due to basic check in MapUrl_ConfigLoad()! */
            if (WATCH_CAT && WatchThisOne)
               WatchDataFormatted ("IGNORE !AZ\n", CurrentPtr);
            while (*cptr && !ISLWS(*cptr) && *cptr != ']') cptr++;
            continue;
      }

      if (NegateThisCondition)
         SoFarSoGood = SoFarSoGood || !Result;
      else
         SoFarSoGood = SoFarSoGood || Result;

      if (WATCH_CAT && WatchThisOne)
         WatchDataFormatted ("!AZ !AZ\n",
                             SoFarSoGood ? "MATCH" : "NOMATCH", CurrentPtr);
   }

   if (Debug)
      fprintf (stdout, "%d %d %d\n",
               ConditionalCount, NegateEntireConditional, SoFarSoGood);

   if (!ConditionalCount)
      SoFarSoGood = true;
   else
   if (NegateEntireConditional)
      SoFarSoGood = !SoFarSoGood;

   if (WATCH_CAT && WatchThisOne)
      WatchDataFormatted ("!AZ conditional\n",
                          SoFarSoGood ? "PASSED" : "FAILED");

   return (SoFarSoGood);
}

/*****************************************************************************/
/*
Compare single "??:" conditional string to each of a comma-separated list in
the form "item" or "item1, item2, item3", etc. String may contain '*' and '%'
wildcards.  We can munge the list string like this because we're operating at
AST-delivery level and cannot be preempted!
*/

BOOL MapUrl_ConditionalList
(
REQUEST_STRUCT *rqptr,
char *String,
char *List
)
{
   char  ch;
   char  *cptr, *lptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && (WATCH_MODULE(WATCH_MOD_MAPURL)))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_ConditionalList() !&Z !&Z", String, List);

   if (!(lptr = List)) return (false);
   if (!String) return (false);
   while (*lptr)
   {
      /* spaces and commas are element separators, step over */
      while (ISLWS(*lptr) || *lptr == ',') lptr++;
      /* if end of list (or empty) then finished */
      if (!*lptr) return (false);
      /* find end of this element, save next character, terminate element */
      for (cptr = lptr; *lptr && !ISLWS(*lptr) && *lptr != ','; lptr++);
      ch = *lptr;
      *lptr = '\0';
      /* look for what we want */
      if (StringMatch (cptr, String, true))
      {
         /* restore list */
         *lptr = ch;
         return (true);
      }
      /* restore list */
      *lptr = ch;
   }
   /* ran out of list elements, so it can't be a match! */
   return (false);
}

/*****************************************************************************/
/*
Get the path representing the username default device/directory.  This can be
retrieved from a cache, or retrieved from directly from the SYSUAF and the
cache subsequently updated.  The 'MAPURL_USER_RULE_FORBIDDEN_MSG' used to
return a user-access-denied to the calling routine, is designed to mask the
actual reason for denial.  This is done to help prevent the leakage of user
account information by being able to differentiate between accounts that exist
and those that don't, by looking at the message.  RequestExecute() checks for
this message and converts it into a "directory not found" message, the same as
that generated if a user does not have a [.WWW] subdirectory in the home area.
*/ 

char* MapUrl_VmsUserName
(
REQUEST_STRUCT* rqptr,
char *UserNamePtr,
int *PathOdsPtr
)
{
   /* UAI flags that disallow SYSUAF-controlled account access */
   static unsigned long  DisallowVmsFlags =
          UAI$M_DISACNT | UAI$M_PWD_EXPIRED | UAI$M_PWD2_EXPIRED |
          UAI$M_CAPTIVE | UAI$M_RESTRICTED;

   static unsigned long  Context = -1;

   static unsigned long  UaiFlags;
   static unsigned long  UaiPriv [2];
   static char  UaiDefDev [MAPURL_USER_DEFDEV_SIZE+1],
                UaiDefDir [MAPURL_USER_DEFDIR_SIZE+1],
                UserName [MAPURL_USER_NAME_SIZE+1],
                UserPath [MAPURL_USER_PATH_SIZE+1];
   static $DESCRIPTOR (UserNameDsc, UserName);

   static struct {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } UaiItems [] = 
   {
      { sizeof(UaiFlags), UAI$_FLAGS, &UaiFlags, 0 },
      { sizeof(UaiPriv), UAI$_PRIV, &UaiPriv, 0 },
      { sizeof(UaiDefDev), UAI$_DEFDEV, &UaiDefDev, 0 },
      { sizeof(UaiDefDir), UAI$_DEFDIR, &UaiDefDir, 0 },
      { 0,0,0,0 }
   };

   int  status,
        PathOds;
   char  *cptr, *sptr, *zptr;
   char  UserDefault [MAPURL_USER_DEFDEV_SIZE+MAPURL_USER_DEFDIR_SIZE+1];

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && (WATCH_MODULE(WATCH_MOD_MAPURL)))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_VmsUserName() !&Z", UserNamePtr);

   zptr = (sptr = UserName) + sizeof(UserName);
   for (cptr = UserNamePtr; *cptr && sptr < zptr; *sptr++ = toupper(*cptr++));
   if (sptr >= zptr) return (MAPURL_USER_RULE_FORBIDDEN_MSG);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "UserName |%s|\n", UserName);
   UserNameDsc.dsc$w_length = sptr - UserName;

   /* look for it, and if found, return it from the cache */
   if ((cptr = MapUrl_VmsUserNameCache (rqptr, UserName, NULL, PathOdsPtr)))
      return (cptr);

   /***************************************/
   /* get the information from the SYSUAF */
   /***************************************/

   /* turn on SYSPRV to allow access to SYSUAF records */
   sys$setprv (1, &SysPrvMask, 0, 0);

   status = sys$getuai (0, &Context, &UserNameDsc, &UaiItems, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$getuai() %%X%08.08X\n", status);

   sys$setprv (0, &SysPrvMask, 0, 0);

   if (VMSnok (status)) 
   {
      if (status == RMS$_RNF)
      {
         if (WATCHING(rqptr) &&
             WATCH_CATEGORY(WATCH_MAPPING))
            WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                       "USER fail !AZ unknown", UserName);
         return (MAPURL_USER_RULE_FORBIDDEN_MSG);
      }
      if (WATCHING(rqptr) &&
          WATCH_CATEGORY(WATCH_MAPPING))
         WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                    "USER sys$getuai() !&S %!&M", status, status);
      return (MAPURL_USER_RULE_FORBIDDEN_MSG);
   }

   /* automatically disallow if any of these flags are set! */
   if (UaiFlags & DisallowVmsFlags)
   {
      if (Debug) fprintf (stdout, "failed on flags\n");
      if (WATCHING(rqptr) &&
          WATCH_CATEGORY(WATCH_MAPPING))
         WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                    "USER !AZ fail SYSUAF flags", UserName);
      return (MAPURL_USER_RULE_FORBIDDEN_MSG);
   }

   if (!AuthPolicySysUafRelaxed && (UaiPriv[0] & PRV$M_SYSPRV))
   {
      /* not allowing all accounts, exclude those with extended privileges */
      if (Debug) fprintf (stdout, "failed on privileges\n");
      if (WATCHING(rqptr) &&
          WATCH_CATEGORY(WATCH_MAPPING))
         WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                    "USER !AZ fail SYSUAF privileges", UserName);
      return (MAPURL_USER_RULE_FORBIDDEN_MSG);
   }

   zptr = (sptr = UserDefault) + sizeof(UserDefault);
   for (cptr = UaiDefDev+1; UaiDefDev[0] && sptr < zptr; UaiDefDev[0]--)
      *sptr++ = *cptr++;
   for (cptr = UaiDefDir+1; UaiDefDir[0] && sptr < zptr; UaiDefDir[0]--)
      *sptr++ = *cptr++;
   if (sptr >= zptr)
      return (MAPURL_USER_RULE_FORBIDDEN_MSG);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "UserDefault |%s|\n", UserDefault);

#ifdef ODS_EXTENDED

   if (OdsExtended)
   {
      /* on-disk-structure of user area */
      PathOds = MapUrl_VolumeOds (UserDefault);
      if (PathOds && (PathOds != MAPURL_PATH_ODS_2 &&
                      PathOds != MAPURL_PATH_ODS_5))
      {
         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_MAPPING))
            WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                       "USER !AZ !AZ ODS-? %X8XL %!&M",
                       UserName, UserDefault, PathOds, PathOds);
         return (MAPURL_USER_RULE_FORBIDDEN_MSG);
      }
   }
   else

#endif /* ODS_EXTENDED */

      PathOds = 0;

   /* generate a URL-style version of the VMS specification */
   MapUrl_VmsToUrl (UserPath, UserDefault, sizeof(UserPath), true, PathOds);

   if (WATCHING(rqptr) &&
       WATCH_CATEGORY(WATCH_MAPPING))
      WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                 "USER !AZ !AZ !AZ ODS-!UL",
                 UserName, UserDefault, UserPath, PathOds);

   /* trim trailing slash */
   for (cptr = UserPath; *cptr && *(unsigned short*)cptr != '/\0'; cptr++);
   *cptr = '\0';

   /* update it in the cache */
   MapUrl_VmsUserNameCache (rqptr, UserName, UserPath, &PathOds);

   /* if required return the path's on-disk-structure */
   if (PathOdsPtr) *PathOdsPtr = PathOds;

   return (UserPath);
}

/*****************************************************************************/
/*
Keep a linked-list of cache entries.  If 'UserName' and 'UserPath' are NULL
then the list is reset (this happen on mapping rule reload).  If 'UserName' is
non-NULL and 'UserPath' is NULL the list is searched for a matching username
and the associated path string returned.  If 'UserName' and 'UserPath' are
non-NULL add/update a cache entry.  If the list has reached maximum size reuse
the last entry, otherwise create a new entry.  Move/add the entry to the head
of the list.  It's therefore a first-in/first-out queue.  Cache contents remain
current until demands on space (due to new entries) cycles through the maximum
available entries.  To explicitly flush the contents reload the rules.
*/ 

char* MapUrl_VmsUserNameCache
(
REQUEST_STRUCT* rqptr,
char *UserName,
char *UserPath,
int *PathOdsPtr
)
{
   LIST_ENTRY  *leptr;
   MAP_URL_USER_ENTRY  *ucptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (rqptr, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_VmsUserNameCache() !&Z !&Z !&X",
                 UserName, UserPath, PathOdsPtr);

   if (!UserName && !UserPath)
   {
      /***************/
      /* reset cache */
      /***************/

      MapUrlUserNameCacheEntries = Config.cfMisc.MapUserNameCacheEntries;
      if (!MapUrlUserNameCacheEntries)
         MapUrlUserNameCacheEntries = MAPURL_DEFAULT_USER_CACHE_SIZE;
      MapUrlUserNameCacheCount = 0;

      /* empty the list */
      leptr = MapUrlUserNameCacheList.HeadPtr;
      MapUrlUserNameCacheList.HeadPtr = MapUrlUserNameCacheList.TailPtr = NULL;
      MapUrlUserNameCacheList.EntryCount = 0;
      while (leptr) 
      {
         ucptr = (MAP_URL_USER_ENTRY*)leptr;
         leptr = leptr->NextPtr;
         VmFree (ucptr, FI_LI);
      }
      return (NULL);
   }

   if (!UserPath)
   {
      /****************/
      /* search cache */
      /****************/

      /* process the cache entry list from most to least recent */
      for (leptr = MapUrlUserNameCacheList.HeadPtr;
           leptr;
           leptr = leptr->NextPtr)
      {
         ucptr = (MAP_URL_USER_ENTRY*)leptr;

         if (Debug)
            fprintf (stdout, "|%s|%s|\n", ucptr->UserName, ucptr->UserPath);

         /* if this one has been reset there won't be any more down the list */
         if (!ucptr->UserName[0]) break;

         /* compare the first two characters (at least one and a null) */
         if (*(unsigned short*)ucptr->UserName != *(unsigned short*)UserName)
            continue;
         /* full string comparison */
         if (strcmp (ucptr->UserName, UserName)) continue;

         /*************/
         /* cache hit */
         /*************/

         if ((void*)MapUrlUserNameCacheList.HeadPtr != (void*)ucptr)
         {
            /* move it to the head of the list */
            ListRemove (&MapUrlUserNameCacheList, ucptr);
            ListAddHead (&MapUrlUserNameCacheList, ucptr);
         }

         ucptr->HitCount++;
         memcpy (&ucptr->LastBinaryTime, &rqptr->rqTime.Vms64bit, 8);

         if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_MAPPING))
            WatchThis (rqptr, FI_LI, WATCH_MAPPING,
                       "USER !AZ cache !AZ/ ODS-!UL !UL hits",
                       ucptr->UserName, ucptr->UserPath,
                       ucptr->PathOds, ucptr->HitCount);

         if (PathOdsPtr) *PathOdsPtr = ucptr->PathOds;
         return (ucptr->UserPath);
      }

      /* not found */
      return (NULL);
   }

   /****************/
   /* update cache */
   /****************/

   if (MapUrlUserNameCacheCount < MapUrlUserNameCacheEntries)
   {
      /* allocate memory for a new entry */
      ucptr = VmGet (sizeof (MAP_URL_USER_ENTRY));
      MapUrlUserNameCacheCount++;
   }
   else
   {
      /* reuse the tail entry (least recent) */
      ucptr = MapUrlUserNameCacheList.TailPtr;
      ucptr->ReuseCount++;
      ListRemove (&MapUrlUserNameCacheList, ucptr);
   }

   /* add entry to the head of the user cache list (most recent) */
   ListAddHead (&MapUrlUserNameCacheList, ucptr);

   strncpy (ucptr->UserName, UserName, sizeof(ucptr->UserName));
   strncpy (ucptr->UserPath, UserPath, sizeof(ucptr->UserPath));
   ucptr->UserName[sizeof(ucptr->UserName)-1] =
      ucptr->UserPath[sizeof(ucptr->UserPath)-1] = '\0';
   ucptr->HitCount = 1;
   memcpy (&ucptr->LastBinaryTime, &rqptr->rqTime.Vms64bit, 8);
#ifdef ODS_EXTENDED
   if (PathOdsPtr) ucptr->PathOds = *PathOdsPtr;
#else /* ODS_EXTENDED */
   if (PathOdsPtr) ucptr->PathOds = 0;
#endif /* ODS_EXTENDED */

   return (UserPath);
}

/****************************************************************************/
/*
Get a volume's underlying file system (ODS-2 or ODS-5), by sys$getdvi() the
ACPTYPE for the device.  Return 2 for an ODS-2 volume and 5 for an ODS-5
volume.  Any other value returned is a VMS status value (error)!
*/ 

#ifdef ODS_EXTENDED

/* if pre-7.2 Alpha then define this */
#ifndef DVI$C_ACP_F11V5
#define DVI$C_ACP_F11V5 11
#endif

int MapUrl_VolumeOds (char *DeviceName)

{
   static int  DevOds;
   static unsigned long  DevAcpType,
                         DevChar;
   static $DESCRIPTOR (DevNameDsc, "");
   static struct {
      short  buf_len;
      short  item;
      char   *buf_addr;
      short  *ret_len;
   }
   AcpTypeItemList [] = 
   {
      { sizeof(DevAcpType), DVI$_ACPTYPE, &DevAcpType, 0 },
      { sizeof(DevChar), DVI$_DEVCHAR, &DevChar, 0 },
      { 0, 0, 0, 0 }
   };

   int  status;
   char  *cptr;
   struct {
      unsigned short  Status;
      unsigned short  Count;
      unsigned long  Unused;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_MAPURL))
      WatchThis (NULL, FI_LI, WATCH_MOD_MAPURL,
                 "MapUrl_VolumeOds() !&Z", DeviceName);

   if (!OdsExtended) return (0);

   for (cptr = DeviceName; *cptr && *cptr != ':'; cptr++);
   /* if we can't find a device in it then assume it's ODS-2 */
   if (!*cptr) return (0);
   DevNameDsc.dsc$a_pointer = DeviceName;
   DevNameDsc.dsc$w_length = cptr - DeviceName;
   if (Debug)
      fprintf (stdout, "|%*.*s|\n",
               DevNameDsc.dsc$w_length, DevNameDsc.dsc$w_length,
               DevNameDsc.dsc$a_pointer);

   status = sys$getdviw (EfnWait, 0, &DevNameDsc, &AcpTypeItemList,
                         &IOsb, 0, 0, 0);
   if (VMSok (status)) status = IOsb.Status;
   if (!(DevChar & DEV$M_MNT)) status = RMS$_DNR;
   if (VMSnok(status) || VMSnok(status = IOsb.Status))
   {
      /* volume not mounted or error */
      return (status);
   }

   if (DevAcpType == DVI$C_ACP_F11V2)
   { 
      if (Debug) fputs ("ODS:2\n", stdout); 
      return (DevOds = MAPURL_PATH_ODS_2);
   }

   if (DevAcpType == DVI$C_ACP_F11V5)
   {
      if (Debug) fputs ("ODS:5\n", stdout); 
      return (DevOds = MAPURL_PATH_ODS_5);
   }

   /* slightly bogus message */
   return (SS$_NOTFILEDEV);
}

#endif /* ODS_EXTENDED */

/*****************************************************************************/

